@@ -59,17 +59,6 @@
 #include "isochronous.hpp"
 #include "Mutex.h"
 
-// Define fatal and nonfatal write errors
-#ifdef WIN32
-#define FATALTCPWRITERR(errno)  ((errno = WSAGetLastError()) != WSAETIMEDOUT)
-#define NONFATALTCPWRITERR(errno) ((errno = WSAGetLastError()) == WSAETIMEDOUT)
-#define FATALUDPWRITERR(errno)  (((errno = WSAGetLastError()) != WSAETIMEDOUT) && (errno != WSAECONNREFUSED))
-#else
-#define FATALTCPWRITERR(errno)  (errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
-#define NONFATALTCPWRITERR(errno)  (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)
-#define FATALUDPWRITERR(errno) 	((errno != EAGAIN) && (errno != EWOULDBLOCK) && (errno != EINTR) && (errno != ECONNREFUSED) && (errno != ENOBUFS))
-#endif
-
 /* ------------------------------------------------------------------- */
 class Client {
 public:
@@ -93,6 +82,7 @@ public:
     bool isConnected(void) const;
     int SendFirstPayload(void);
     int BarrierClient(struct BarrierMutex *);
+    void RunBounceBackTCP(void);
     struct ReportHeader *myJob;
 
 private:
@@ -107,6 +97,8 @@ private:
     bool InProgress(void);
     void PostNullEvent(void);
     void AwaitServerCloseEvent(void);
+    inline void tcp_shutdown(void);
+    inline void tcp_drain(void);
     bool connected;
     ReportStruct scratchpad;
     ReportStruct *reportstruct;
@@ -115,16 +107,17 @@ private:
     bool one_report;
     bool apply_first_udppkt_delay;
     int udp_payload_minimum;
-    void myReportPacket (void);
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-    struct tcp_info my_tcpi_stats;
-    bool myReportPacket (bool sample_tcpi);
-#endif
+    void myReportPacket(void);
+
     // TCP plain
     void RunTCP(void);
     // TCP version which supports rate limiting per -b
     void RunRateLimitedTCP(void);
     void RunNearCongestionTCP(void);
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    bool AwaitWriteSelectEventTCP(void);
+    void RunWriteEventsTCP(void);
+#endif
     // UDP traffic with isochronous and vbr support
     void RunUDPIsochronous(void);
     // UDP plain
@@ -138,7 +131,6 @@ private:
     int mySocket;
 #endif
     struct ReporterData *myReport;
-    char* mBuf;
     Timestamp mEndTime;
     Timestamp lastPacketTime;
     Timestamp now;
@@ -147,6 +139,9 @@ private:
     Isochronous::FrameCounter *framecounter;
     bool isburst;
     bool peerclose;
+    Timestamp drain_start;
+    Timestamp drain_end;
+    struct tcp_init_conditions my_init_cond;
 }; // end class Client
 
 #endif // CLIENT_H
