@@ -96,6 +96,8 @@ extern const char client_report_epoch_start_current[];
 
 extern const char server_pid_port[];
 
+extern const char server_pid_portrange[];
+
 extern const char client_pid_port[];
 
 extern const char client_pid_port_dev[];
@@ -104,6 +106,8 @@ extern const char bind_address[];
 
 extern const char bind_address_iface[];
 
+extern const char bind_address_iface_taptun[];
+
 extern const char multicast_ttl[];
 
 extern const char join_multicast[];
@@ -132,6 +136,12 @@ extern const char wait_server_threads[];
 
 extern const char client_isochronous[];
 
+extern const char client_burstperiod[];
+
+extern const char client_bounceback[];
+
+extern const char server_burstperiod[];
+
 extern const char client_fq_pacing[];
 /* -------------------------------------------------------------------
  * Legacy reports
@@ -139,12 +149,14 @@ extern const char client_fq_pacing[];
 
 extern const char report_bw_header[];
 
-extern const char report_bw_sumcnt_header[];
+extern const char report_sumcnt_bw_header[];
 
 extern const char report_bw_format[];
 
 extern const char report_sum_bw_format[];
+
 extern const char report_sumcnt_bw_format[];
+
 extern const char report_bw_read_format[];
 
 extern const char report_bw_jitter_loss_header[];
@@ -153,6 +165,8 @@ extern const char report_bw_jitter_loss_format[];
 
 extern const char report_sum_bw_jitter_loss_format[];
 
+extern const char report_sumcnt_bw_jitter_loss_header[];
+
 extern const char report_sumcnt_bw_jitter_loss_format[];
 
 /* -------------------------------------------------------------------
@@ -165,6 +179,12 @@ extern const char server_read_size[];
 
 extern const char report_bw_enhanced_format[];
 
+extern const char report_write_enhanced_isoch_format[];
+
+extern const char report_write_enhanced_isoch_header[];
+
+extern const char report_write_enhanced_isoch_nocwnd_format[];
+
 extern const char report_sum_bw_enhanced_format[];
 
 extern const char report_bw_read_enhanced_header[];
@@ -177,20 +197,36 @@ extern const char report_bw_read_enhanced_format[];
 
 extern const char report_sum_bw_read_enhanced_format[];
 
+extern const char report_sumcnt_bw_read_enhanced_header[];
+
 extern const char report_sumcnt_bw_read_enhanced_format[];
 
+extern const char report_sumcnt_bw_pps_enhanced_header[];
+
+extern const char report_sumcnt_bw_pps_enhanced_format[];
+
 extern const char report_triptime_enhanced_format[];
 
 extern const char report_bw_write_enhanced_header[];
 
-extern const char report_bw_write_sumcnt_enhanced_header[];
+extern const char report_bw_write_enhanced_header[];
+
+extern const char report_sumcnt_write_enhanced_drain_header[];
 
 extern const char report_bw_write_enhanced_format[];
 
+extern const char report_write_enhanced_drain_header[];
+
+extern const char report_write_enhanced_drain_format[];
+
+extern const char report_write_enhanced_nocwnd_drain_format[];
+
 extern const char report_bw_write_enhanced_nocwnd_format[];
 
 extern const char report_sum_bw_write_enhanced_format[];
 
+extern const char report_sumcnt_bw_write_enhanced_header[];
+
 extern const char report_sumcnt_bw_write_enhanced_format[];
 
 extern const char report_bw_pps_enhanced_header[];
@@ -215,6 +251,14 @@ extern const char report_bw_jitter_loss_enhanced_isoch_header[];
 
 extern const char report_bw_jitter_loss_enhanced_isoch_format[];
 
+extern const char report_bw_jitter_loss_enhanced_triptime_header[];
+
+extern const char report_bw_jitter_loss_enhanced_triptime_format[];
+
+extern const char report_bw_jitter_loss_enhanced_isoch_triptime_header[];
+
+extern const char report_bw_jitter_loss_enhanced_isoch_triptime_format[];
+
 extern const char report_bw_jitter_loss_suppress_enhanced_format[];
 
 extern const char report_sum_bw_jitter_loss_enhanced_format[];
@@ -233,6 +277,12 @@ extern const char report_frame_jitter_loss_suppress_enhanced_format[];
 
 extern const char report_frame_tcp_enhanced_header[];
 
+extern const char report_burst_read_tcp_header[];
+
+extern const char report_burst_read_tcp_format[];
+
+extern const char report_burst_read_tcp_final_format[];
+
 extern const char report_udp_fullduplex_header[];
 
 extern const char report_udp_fullduplex_format[];
@@ -241,6 +291,9 @@ extern const char report_udp_fullduplex_sum_format[];
 
 extern const char report_udp_fullduplex_enhanced_format[];
 
+extern const char report_sumcnt_udp_fullduplex_header[];
+
+extern const char report_sumcnt_udp_fullduplex_format[];
 
 /* -------------------------------------------------------------------
  * Misc reports
@@ -248,6 +301,8 @@ extern const char report_udp_fullduplex_enhanced_format[];
 
 extern const char report_outoforder[];
 
+extern const char report_sumcnt_outoforder[];
+
 extern const char report_l2statistics[];
 
 extern const char report_sum_outoforder[];
@@ -256,6 +311,8 @@ extern const char report_peer[];
 
 extern const char report_peer_dev[];
 
+extern const char report_peer_fail[];
+
 extern const char report_mss_unsupported[];
 
 extern const char report_mss[];
@@ -325,6 +382,13 @@ extern const char warn_len_too_small_peer_exchange[];
 extern const char warn_compat_and_peer_exchange[];
 
 extern const char warn_seqno_wrap[];
+
+extern const char warn_start_before_now[];
+
+extern const char error_starttime_exceeds[];
+
+extern const char error_delaytime_exceeds[];
+
 #ifdef __cplusplus
 } /* end extern "C" */
 #endif
