@@ -132,17 +132,22 @@ struct ReadStats {
 struct WriteStats {
     int WriteCnt;
     int WriteErr;
-    int TCPretry;
     int totWriteCnt;
     int totWriteErr;
+#if (HAVE_TCP_STATS)
+    int TCPretry;
     int totTCPretry;
-    int lastTCPretry;
     int cwnd;
     int rtt;
-    double meanrtt;
-    int up_to_date;
+    int rttvar;
+#endif
 };
 
+struct tcp_init_conditions {
+    int cwnd;
+    int rtt;
+    double connecttime;
+};
 /*
  * This struct contains all important information from the sending or
  * recieving thread.
@@ -169,6 +174,10 @@ struct L2Stats {
     intmax_t tot_lengtherr;
 };
 
+struct DrainStats {
+    struct MeanMinMaxStats current;
+    struct MeanMinMaxStats total;
+};
 /*
  * The type field of ReporterData is a bitmask
  * with one or more of the following
@@ -215,17 +224,26 @@ struct ReportCommon {
     int BufLen;
     int MSS;
     int TCPWin;
+    int TOS;
+    int RTOS;
+#if HAVE_DECL_TCP_WINDOW_CLAMP
+    int ClampSize;
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    int WritePrefetch;
+#endif
     int winsize_requested;
     unsigned int FQPacingRate;
-    int RXbins;
-    int RXbinsize;
-    int RXunits;
+    int HistBins;
+    int HistBinsize;
+    int HistUnits;
     double pktIPG;
     iperf_sockaddr peer;
     Socklen_t size_peer;
     iperf_sockaddr local;
     Socklen_t size_local;
     char* Host;                   // -c
+    char* HideHost;
     char* Localhost;              // -B
     char* Ifrname;
     char* Ifrnametx;
@@ -250,13 +268,13 @@ struct ReportCommon {
 struct ConnectionInfo {
     struct ReportCommon *common;
     struct timeval connect_timestamp;
-    double connecttime;
     struct timeval txholdbacktime;
     struct timeval epochStartTime;
     int winsize;
     char peerversion[PEERVERBUFSIZE];
     struct MeanMinMaxStats connect_times;
     int MSS;
+    struct tcp_init_conditions init_cond;
 };
 
 struct ShiftIntCounter {
@@ -324,6 +342,7 @@ struct ReportTimeStamps {
     struct timeval nextTime;
     struct timeval intervalTime;
     struct timeval IPGstart;
+    struct timeval nextTCPStampleTime;
 };
 
 struct TransferInfo {
@@ -332,7 +351,7 @@ struct TransferInfo {
     void (*output_handler) (struct TransferInfo *stats);
     int groupID;
     int threadcnt;
-    int filter_this_sample_output;
+    bool isMaskOutput;
     uintmax_t cntBytes;
     intmax_t cntError;
     intmax_t cntOutofOrder;
@@ -356,6 +375,11 @@ struct TransferInfo {
     char csv_peer[CSVPEERLIMIT];
     bool final;
     bool burstid_transition;
+    bool isEnableTcpInfo;
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    struct DrainStats drain_mmm;
+    struct histogram *drain_histogram;
+#endif
 };
 
 struct SumReport {
@@ -369,10 +393,10 @@ struct SumReport {
 
 struct ReporterData {
     // function pointer for per packet processing
-    void (*packet_handler) (struct ReporterData *data, struct ReportStruct *packet);
+    void (*packet_handler_pre_report) (struct ReporterData *data, struct ReportStruct *packet);
+    void (*packet_handler_post_report) (struct ReporterData *data, struct ReportStruct *packet);
     void (*transfer_protocol_handler) (struct ReporterData *data, int final);
     int (*transfer_interval_handler) (struct ReporterData *data, struct ReportStruct *packet);
-    bool burst_boundary;
 
     struct PacketRing *packetring;
     int reporter_thread_suspends; // used to detect CPU bound systems
@@ -406,16 +430,12 @@ typedef void (* report_serverstatistics)( struct ConnectionInfo *, struct Transf
 void SetSumHandlers (struct thread_Settings *inSettings, struct SumReport* sumreport);
 struct SumReport* InitSumReport(struct thread_Settings *inSettings, int inID, int fullduplex);
 struct ReportHeader* InitIndividualReport(struct thread_Settings *inSettings);
-struct ReportHeader* InitConnectionReport(struct thread_Settings *inSettings, double ct);
+struct ReportHeader* InitConnectionReport(struct thread_Settings *inSettings, struct tcp_init_conditions *init_cond);
 struct ConnectionInfo* InitConnectOnlyReport(struct thread_Settings *thread);
 struct ReportHeader *InitSettingsReport(struct thread_Settings *inSettings);
 struct ReportHeader* InitServerRelayUDPReport(struct thread_Settings *inSettings, struct server_hdr *server);
 void PostReport(struct ReportHeader *reporthdr);
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-bool ReportPacket (struct ReporterData* data, struct ReportStruct *packet, struct tcp_info *tcp_stats);
-#else
-void ReportPacket (struct ReporterData* data, struct ReportStruct *packet);
-#endif
+bool ReportPacket (struct ReporterData* data, struct ReportStruct *packet);
 int EndJob(struct ReportHeader *reporthdr,  struct ReportStruct *packet);
 void FreeReport(struct ReportHeader *reporthdr);
 void FreeSumReport (struct SumReport *sumreport);
@@ -495,6 +515,9 @@ void tcp_output_write_enhanced (struct TransferInfo *stats);
 void tcp_output_write_enhanced_isoch (struct TransferInfo *stats);
 void tcp_output_sum_write_enhanced (struct TransferInfo *stats);
 void tcp_output_sumcnt_write_enhanced (struct TransferInfo *stats);
+#if (HAVE_DECL_TCP_NOTSENT_LOWAT)
+void tcp_output_write_enhanced_drain (struct TransferInfo *stats);
+#endif
 // TCP fullduplex
 void tcp_output_fullduplex(struct TransferInfo *stats);
 void tcp_output_fullduplex_enhanced(struct TransferInfo *stats);
@@ -538,17 +561,13 @@ void PrintMSS(struct ReporterData *data);
 void reporter_default_heading_flags(int);
 void reporter_connect_printf_tcp_final(struct ConnectionInfo *report);
 
-void write_UDP_AckFIN(struct TransferInfo *stats);
+void write_UDP_AckFIN(struct TransferInfo *stats, int len);
 
 int reporter_process_transfer_report (struct ReporterData *this_ireport);
 int reporter_process_report (struct ReportHeader *reporthdr);
 
 void setTransferID(struct thread_Settings *inSettings, int role_reversal);
 
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-void gettcpistats(struct ReporterData *data, int final, struct tcp_info *tcp_stats);
-#endif
-
 #ifdef __cplusplus
 } /* end extern "C" */
 #endif
