@@ -59,13 +59,6 @@
 #include "util.h"
 #include "Timestamp.hpp"
 
-#ifdef WIN32
-#define NONFATALTCPREADERR(errno) (WSAGetLastError() != WSAEWOULDBLOCK)
-#else
-#define NONFATALTCPREADERR(errno) (errno != EAGAIN && errno != EWOULDBLOCK)
-#endif
-
-
 /* ------------------------------------------------------------------- */
 class Server {
 public:
@@ -78,12 +71,11 @@ public:
     // accepts connection and receives data
     void RunUDP(void);
     void RunTCP(void);
+    void RunTcpBounceBack(void);
     static void Sig_Int(int inSigno);
 
 private:
     thread_Settings *mSettings;
-    char* mBuf;
-    int mBufLen;
     Timestamp mEndTime;
     Timestamp now;
     ReportStruct scratchpad;
@@ -100,6 +92,7 @@ private:
     void udp_isoch_processing(int);
     bool InProgress(void);
     int SkipFirstPayload(void);
+    void ClientReverseFirstRead(void);
     Timestamp connect_done;
     bool peerclose;
     bool isburst;
