@@ -88,10 +88,17 @@ extern "C" {
 #define MAXDIFFTXSTART 3600
 // maximum difference in seconds to bound --txdelay-time,
 // if this is too large and w/o keep-alives the connect may drop
-#define MAXDIFFTXDELAY 60
+#define MAXDIFFTXDELAY 3600
 // maximum inter packet gap (or write delay) for UDP packets
 #define MAXIPGSECS 60
 #define CSVPEERLIMIT ((REPORT_ADDRLEN * 2) + 40)
+#define NEARCONGEST_DEFAULT 0.5
+#define DEFAULT_PERMITKEY_LIFE 20.0 // units is seconds
+#define TESTEXCHANGETIMEOUT (4 * 1000000) // 4 secs, units is microseconds
+#ifndef MAXTTL
+#define MAXTTL 255
+#endif
+#define DEFAULT_BOUNCEBACK_BYTES 100
 
 // server/client mode
 enum ThreadMode {
@@ -123,7 +130,7 @@ enum TestMode {
 enum IntervalMode {
     kInterval_None = 0,
     kInterval_Time,
-    kInterval_Frames,
+    kInterval_Frames
 };
 
 // rate request units
@@ -133,6 +140,7 @@ enum RateUnits {
 };
 
 #include "Reporter.h"
+#include "payloads.h"
 
 /*
  * The thread_Settings is a structure that holds all
@@ -149,14 +157,18 @@ struct thread_Settings {
     // Pointers
     char*  mFileName;               // -F
     char*  mHost;                   // -c
+    char*  mHideHost;
     char*  mLocalhost;              // -B
     char*  mOutputFileName;         // -o
     char*  mIfrname;                // %<device> name (for rx)
     char*  mIfrnametx;              // %<device> name (for tx)
     char*  mSSMMulticastStr;        // --ssm-host
     char*  mIsochronousStr;         // --isochronous
-    char*  mRxHistogramStr;         // --histograms (packets)
+    char*  mHistogramStr;         // --histograms (packets)
+    char*  mTransferIDStr;          //
+    char*  mBuf;
     FILE*  Extractor_file;
+
     struct ReportHeader* reporthdr;
     struct SumReport* mSumReport;
     struct SumReport* mFullDuplexReport;
@@ -165,6 +177,9 @@ struct thread_Settings {
     // int's
     int mThreads;                   // -P
     int mTOS;                       // -S
+    int mRTOS;                      // reflected TOS
+    int mTransferID;
+    int mConnectRetries;
 #if WIN32
     SOCKET mSock;
 #else
@@ -202,20 +217,22 @@ struct thread_Settings {
         bool   mSinlgeClient;           // -1 */
     int flags;
     int flags_extend;
+    int flags_extend2;
     // enums (which should be special int's)
     enum ThreadMode mThreadMode;         // -s or -c
     enum ReportMode mReportMode;
     enum TestMode mMode;              // -r or -d
     bool clientListener;              // set to True if client mode listener per -r or -d
     // Hopefully int64_t's
-    uintmax_t mUDPRate;            // -b or -u
-    char mUDPRateUnits;            // -b is either bw or pps
-    uintmax_t mAmount;             // -n or -t
+    uintmax_t mAppRate;            // -b or -u
+    char mAppRateUnits;            // -b is either bw or pps
+    uintmax_t mAmount;             // -n or -t time unit is 10 ms
     unsigned int mInterval;               // -i integer time units is usec
     enum IntervalMode mIntervalMode;
     // shorts
     unsigned short mListenPort;     // -L
     unsigned short mPort;           // -p
+    unsigned short mPortLast;       // -p last port, e.g. when -p 6001-6010
     unsigned short mBindPort;      // -B
     // chars
     char   mFormat;                 // -f
@@ -228,18 +245,21 @@ struct thread_Settings {
     Socklen_t size_local;
     nthread_t mTID;
     int incrdstip;
+    int incrsrcip;
+    int incrsrcport;
     int connectonly_count;
     char* mCongestion;
-    int mRXbins;
-    int mRXbinsize;
-    int mRXunits;
-    double mRXci_lower;
-    double mRXci_upper;
+    int mHistBins;
+    int mHistBinsize;
+    int mHistUnits;
+    double mHistci_lower;
+    double mHistci_upper;
 #if defined(HAVE_WIN32_THREAD)
     HANDLE mHandle;
 #endif
     double mFPS; //frames per second
     double mMean; //variable bit rate mean
+    uint32_t mBurstSize; //number of bytes in a burst
     int mJitterBufSize; //Server jitter buffer size, units is frames
     double mBurstIPG; //Interpacket gap
     int l4offset; // used in l2 mode to offset the raw packet
@@ -250,7 +270,7 @@ struct thread_Settings {
     struct timeval txholdback_timer;
     struct timeval txstart_epoch;
     struct timeval accept_time;
-    struct timeval triptime_start;
+    struct timeval sent_time;
     struct Condition awake_me;
     struct PacketRing *ackring;
     struct BarrierMutex *connects_done;
@@ -258,6 +278,20 @@ struct thread_Settings {
     int32_t peer_version_u;
     int32_t peer_version_l;
     double connecttime;
+    double rtt_nearcongest_weight_factor;
+    char mPermitKey[MAX_PERMITKEY_LEN + 1]; //add some space for timestamp
+    struct timeval mPermitKeyTime;
+    bool mKeyCheck;
+    double mListenerTimeout;
+    int tuntapdev;
+    int firstreadbytes;
+    int mBounceBack;
+#if HAVE_DECL_TCP_WINDOW_CLAMP
+    int mClampSize;
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    int mWritePrefetch;
+#endif
 };
 
 /*
@@ -276,7 +310,7 @@ struct thread_Settings {
 #define FLAG_BUFLENSET      0x00000001
 #define FLAG_COMPAT         0x00000002
 #define FLAG_DAEMON         0x00000004
-#define FLAG_DOMAIN         0x00000008
+#define FLAG_DOMAINV6       0x00000008
 #define FLAG_FILEINPUT      0x00000010
 #define FLAG_NODELAY        0x00000020
 #define FLAG_PRINTMSS       0x00000040
@@ -309,7 +343,7 @@ struct thread_Settings {
 #define FLAG_REVERSE        0x00000004
 #define FLAG_ISOCHRONOUS    0x00000008
 #define FLAG_UDPUNUSED      0x00000010
-#define FLAG_RXHISTOGRAM    0x00000020
+#define FLAG_HISTOGRAM      0x00000020
 #define FLAG_L2LENGTHCHECK  0x00000100
 #define FLAG_TXSTARTTIME    0x00000200
 #define FLAG_INCRDSTIP      0x00000400
@@ -321,18 +355,42 @@ struct thread_Settings {
 #define FLAG_MODEINFINITE   0x00010000
 #define FLAG_CONNECTONLY    0x00020000
 #define FLAG_SERVERREVERSE  0x00040000
-#define FLAG_FULLDUPLEX          0x00080000
+#define FLAG_FULLDUPLEX     0x00080000
 #define FLAG_WRITEACK       0x00100000
 #define FLAG_NOUDPFIN       0x00200000
 #define FLAG_NOCONNECTSYNC  0x00400000
 #define FLAG_SUMONLY        0x00800000
 #define FLAG_FRAMEINTERVAL  0x01000000
 #define FLAG_IPG            0x02000000
+#define FLAG_DONTROUTE      0x04000000
+#define FLAG_NEARCONGEST    0x08000000
+#define FLAG_PERMITKEY      0x10000000
+#define FLAG_SETTCPMSS      0x20000000
+#define FLAG_INCRDSTPORT    0x40000000
+#define FLAG_INCRSRCIP      0x80000000
+
+/*
+ * More extended flags
+ */
+#define FLAG_PERIODICBURST  0x00000001
+#define FLAG_SUMDSTIP       0x00000002
+#define FLAG_SMALLTRIPTIME  0x00000004
+#define FLAG_RXCLAMP        0x00000008
+#define FLAG_WRITEPREFETCH  0x00000010
+#define FLAG_TUNDEV         0x00000020
+#define FLAG_TAPDEV         0x00000040
+#define FLAG_HIDEIPS        0x00000080
+#define FLAG_BOUNCEBACK     0x00000100
+#define FLAG_TCPDRAIN       0x00000200
+#define FLAG_INCRSRCPORT    0x00000400
+#define FLAG_OVERRIDETOS    0x00000800
+#define FLAG_DOMAINV4       0x00001000
 
 #define isBuflenSet(settings)      ((settings->flags & FLAG_BUFLENSET) != 0)
 #define isCompat(settings)         ((settings->flags & FLAG_COMPAT) != 0)
 #define isDaemon(settings)         ((settings->flags & FLAG_DAEMON) != 0)
-#define isIPV6(settings)           ((settings->flags & FLAG_DOMAIN) != 0)
+#define isIPV6(settings)           ((settings->flags & FLAG_DOMAINV6) != 0)
+#define isIPV4(settings)           ((settings->flags_extend2 & FLAG_DOMAINV4) != 0)
 #define isFileInput(settings)      ((settings->flags & FLAG_FILEINPUT) != 0)
 #define isNoDelay(settings)        ((settings->flags & FLAG_NODELAY) != 0)
 #define isPrintMSS(settings)       ((settings->flags & FLAG_PRINTMSS) != 0)
@@ -365,14 +423,18 @@ struct thread_Settings {
 #define isFullDuplex(settings)          ((settings->flags_extend & FLAG_FULLDUPLEX) != 0)
 #define isServerReverse(settings)  ((settings->flags_extend & FLAG_SERVERREVERSE) != 0)
 #define isIsochronous(settings)    ((settings->flags_extend & FLAG_ISOCHRONOUS) != 0)
-#define isRxHistogram(settings)    ((settings->flags_extend & FLAG_RXHISTOGRAM) != 0)
+#define isHistogram(settings)    ((settings->flags_extend & FLAG_HISTOGRAM) != 0)
 #define isL2LengthCheck(settings)  ((settings->flags_extend & FLAG_L2LENGTHCHECK) != 0)
 #define isIncrDstIP(settings)      ((settings->flags_extend & FLAG_INCRDSTIP) != 0)
+#define isIncrSrcIP(settings)      ((settings->flags_extend & FLAG_INCRSRCIP) != 0)
+#define isIncrDstPort(settings)    ((settings->flags_extend & FLAG_INCRDSTPORT) != 0)
+#define isIncrSrcPort(settings)    ((settings->flags_extend2 & FLAG_INCRSRCPORT) != 0)
 #define isTxStartTime(settings)    ((settings->flags_extend & FLAG_TXSTARTTIME) != 0)
 #define isTxHoldback(settings)     ((settings->flags_extend & FLAG_TXHOLDBACK) != 0)
 #define isVaryLoad(settings)       ((settings->flags_extend & FLAG_VARYLOAD) != 0)
 #define isFQPacing(settings)       ((settings->flags_extend & FLAG_FQPACING) != 0)
 #define isTripTime(settings)       ((settings->flags_extend & FLAG_TRIPTIME) != 0)
+#define isSmallTripTime(settings)  ((settings->flags_extend2 & FLAG_SMALLTRIPTIME) != 0)
 #define isModeInfinite(settings)   ((settings->flags_extend & FLAG_MODEINFINITE) != 0)
 #define isModeAmount(settings)     (!isModeTime(settings) && !isModeInfinite(settings))
 #define isConnectOnly(settings)    ((settings->flags_extend & FLAG_CONNECTONLY) != 0)
@@ -382,11 +444,26 @@ struct thread_Settings {
 #define isSumOnly(settings)        ((settings->flags_extend & FLAG_SUMONLY) != 0)
 #define isFrameInterval(settings)  ((settings->flags_extend & FLAG_FRAMEINTERVAL) != 0)
 #define isIPG(settings)  ((settings->flags_extend & FLAG_IPG) != 0)
+#define isDontRoute(settings)      ((settings->flags_extend & FLAG_DONTROUTE) != 0)
+#define isNearCongest(settings)    ((settings->flags_extend & FLAG_NEARCONGEST) != 0)
+#define isPermitKey(settings)  ((settings->flags_extend & FLAG_PERMITKEY) != 0)
+#define isTCPMSS(settings)         ((settings->flags_extend & FLAG_SETTCPMSS) != 0)
+#define isPeriodicBurst(settings)  ((settings->flags_extend2 & FLAG_PERIODICBURST) != 0)
+#define isSumServerDstIP(settings) ((settings->flags_extend2 & FLAG_SUMDSTIP) != 0)
+#define isRxClamp(settings)        ((settings->flags_extend2 & FLAG_RXCLAMP) != 0)
+#define isWritePrefetch(settings) ((settings->flags_extend2 & FLAG_WRITEPREFETCH) != 0)
+#define isTapDev(settings)         ((settings->flags_extend2 & FLAG_TAPDEV) != 0)
+#define isTunDev(settings)         ((settings->flags_extend2 & FLAG_TUNDEV) != 0)
+#define isHideIPs(settings)        ((settings->flags_extend2 & FLAG_HIDEIPS) != 0)
+#define isBounceBack(settings)     ((settings->flags_extend2 & FLAG_BOUNCEBACK) != 0)
+#define isTcpDrain(settings)       ((settings->flags_extend2 & FLAG_TCPDRAIN) != 0)
+#define isOverrideTOS(settings)    ((settings->flags_extend2 & FLAG_OVERRIDETOS) != 0)
 
 #define setBuflenSet(settings)     settings->flags |= FLAG_BUFLENSET
 #define setCompat(settings)        settings->flags |= FLAG_COMPAT
 #define setDaemon(settings)        settings->flags |= FLAG_DAEMON
-#define setIPV6(settings)          settings->flags |= FLAG_DOMAIN
+#define setIPV6(settings)          settings->flags |= FLAG_DOMAINV6
+#define setIPV4(settings)          settings->flags_extend2 |= FLAG_DOMAINV4
 #define setFileInput(settings)     settings->flags |= FLAG_FILEINPUT
 #define setNoDelay(settings)       settings->flags |= FLAG_NODELAY
 #define setPrintMSS(settings)      settings->flags |= FLAG_PRINTMSS
@@ -410,21 +487,25 @@ struct thread_Settings {
 #define setRealtime(settings)      settings->flags |= FLAG_REALTIME
 #define setBWSet(settings)         settings->flags |= FLAG_BWSET
 #define setEnhanced(settings)      settings->flags |= FLAG_ENHANCEDREPORT
-#define setServerModeTime(settings)      settings->flags |= FLAG_SERVERMODETIME
+#define setServerModeTime(settings)    settings->flags |= FLAG_SERVERMODETIME
 #define setPeerVerDetect(settings) settings->flags_extend |= FLAG_PEERVER
 #define setSeqNo64b(settings)      settings->flags_extend |= FLAG_SEQNO64
 #define setReverse(settings)       settings->flags_extend |= FLAG_REVERSE
-#define setFullDuplex(settings)         settings->flags_extend |= FLAG_FULLDUPLEX
+#define setFullDuplex(settings)    settings->flags_extend |= FLAG_FULLDUPLEX
 #define setServerReverse(settings) settings->flags_extend |= FLAG_SERVERREVERSE
 #define setIsochronous(settings)   settings->flags_extend |= FLAG_ISOCHRONOUS
-#define setRxHistogram(settings)   settings->flags_extend |= FLAG_RXHISTOGRAM
+#define setHistogram(settings)   settings->flags_extend |= FLAG_HISTOGRAM
 #define setL2LengthCheck(settings) settings->flags_extend |= FLAG_L2LENGTHCHECK
 #define setIncrDstIP(settings)     settings->flags_extend |= FLAG_INCRDSTIP
+#define setIncrSrcIP(settings)     settings->flags_extend |= FLAG_INCRSRCIP
+#define setIncrDstPort(settings)   settings->flags_extend |= FLAG_INCRDSTPORT
+#define setIncrSrcPort(settings)   settings->flags_extend2 |= FLAG_INCRSRCPORT
 #define setTxStartTime(settings)   settings->flags_extend |= FLAG_TXSTARTTIME
 #define setTxHoldback(settings)    settings->flags_extend |= FLAG_TXHOLDBACK
 #define setVaryLoad(settings)      settings->flags_extend |= FLAG_VARYLOAD
 #define setFQPacing(settings)      settings->flags_extend |= FLAG_FQPACING
 #define setTripTime(settings)      settings->flags_extend |= FLAG_TRIPTIME
+#define setSmallTripTime(settings) settings->flags_extend2 |= FLAG_SMALLTRIPTIME
 #define setModeInfinite(settings)  settings->flags_extend |= FLAG_MODEINFINITE
 #define setConnectOnly(settings)   settings->flags_extend |= FLAG_CONNECTONLY
 #define setWriteAck(settings)      settings->flags_extend |= FLAG_WRITEACK
@@ -432,12 +513,27 @@ struct thread_Settings {
 #define setNoConnectSync(settings) settings->flags_extend |= FLAG_NOCONNECTSYNC
 #define setSumOnly(settings)       settings->flags_extend |= FLAG_SUMONLY
 #define setFrameInterval(settings) settings->flags_extend |= FLAG_FRAMEINTERVAL
-#define setIPG(settings) settings->flags_extend |= FLAG_IPG
+#define setIPG(settings)           settings->flags_extend |= FLAG_IPG
+#define setDontRoute(settings)     settings->flags_extend |= FLAG_DONTROUTE
+#define setNearCongest(settings)   settings->flags_extend |= FLAG_NEARCONGEST
+#define setPermitKey(settings)     settings->flags_extend |= FLAG_PERMITKEY
+#define setTCPMSS(settings)        settings->flags_extend |= FLAG_SETTCPMSS
+#define setPeriodicBurst(settings) settings->flags_extend2 |= FLAG_PERIODICBURST
+#define setSumServerDstIP(settings) settings->flags_extend2 |= FLAG_SUMDSTIP
+#define setRxClamp(settings)       settings->flags_extend2 |= FLAG_RXCLAMP
+#define setWritePrefetch(settings) settings->flags_extend2 |= FLAG_WRITEPREFETCH
+#define setTapDev(settings)        settings->flags_extend2 |= FLAG_TAPDEV
+#define setTunDev(settings)        settings->flags_extend2 |= FLAG_TUNDEV
+#define setHideIPs(settings)       settings->flags_extend2 |= FLAG_HIDEIPS
+#define setBounceBack(settings)    settings->flags_extend2 |= FLAG_BOUNCEBACK
+#define setTcpDrain(settings)      settings->flags_extend2 |= FLAG_TCPDRAIN
+#define setOverrideTOS(settings)   settings->flags_extend2 |= FLAG_OVERRIDETOS
 
 #define unsetBuflenSet(settings)   settings->flags &= ~FLAG_BUFLENSET
 #define unsetCompat(settings)      settings->flags &= ~FLAG_COMPAT
 #define unsetDaemon(settings)      settings->flags &= ~FLAG_DAEMON
-#define unsetIPV6(settings)        settings->flags &= ~FLAG_DOMAIN
+#define unsetIPV6(settings)        settings->flags &= ~FLAG_DOMAINV6
+#define unsetIPV4(settings)        settings->flags_extend2 &= ~FLAG_DOMAINV4
 #define unsetFileInput(settings)   settings->flags &= ~FLAG_FILEINPUT
 #define unsetNoDelay(settings)     settings->flags &= ~FLAG_NODELAY
 #define unsetPrintMSS(settings)    settings->flags &= ~FLAG_PRINTMSS
@@ -461,29 +557,46 @@ struct thread_Settings {
 #define unsetRealtime(settings)    settings->flags &= ~FLAG_REALTIME
 #define unsetBWSet(settings)       settings->flags &= ~FLAG_BWSET
 #define unsetEnhanced(settings)    settings->flags &= ~FLAG_ENHANCEDREPORT
-#define unsetServerModeTime(settings)    settings->flags &= ~FLAG_SERVERMODETIME
-#define unsetPeerVerDetect(settings)    settings->flags_extend &= ~FLAG_PEERVER
+#define unsetServerModeTime(settings) settings->flags &= ~FLAG_SERVERMODETIME
+#define unsetPeerVerDetect(settings)  settings->flags_extend &= ~FLAG_PEERVER
 #define unsetSeqNo64b(settings)    settings->flags_extend &= ~FLAG_SEQNO64
 #define unsetReverse(settings)     settings->flags_extend &= ~FLAG_REVERSE
-#define unsetFullDuplex(settings)       settings->flags_extend &= ~FLAG_FULLDUPLEX
+#define unsetFullDuplex(settings)  settings->flags_extend &= ~FLAG_FULLDUPLEX
 #define unsetServerReverse(settings) settings->flags_extend &= ~FLAG_SERVERREVERSE
 #define unsetIsochronous(settings)  settings->flags_extend &= ~FLAG_ISOCHRONOUS
-#define unsetRxHistogram(settings)    settings->flags_extend &= ~FLAG_RXHISTOGRAM
+#define unsetHistogram(settings)  settings->flags_extend &= ~FLAG_HISTOGRAM
 #define unsetL2LengthCheck(settings)  settings->flags_extend &= ~FLAG_L2LENGTHCHECK
 #define unsetIncrDstIP(settings)    settings->flags_extend &= ~FLAG_INCRDSTIP
+#define unsetIncrSrcIP(settings)    settings->flags_extend &= ~FLAG_INCRSRCIP
+#define unsetIncrDstPort(settings)  settings->flags_extend &= ~FLAG_INCRDSTPORT
+#define unsetIncrSrcPort(settings)  settings->flags_extend2 &= ~FLAG_INCRSRCPORT
 #define unsetTxStartTime(settings)  settings->flags_extend &= ~FLAG_TXSTARTTIME
 #define unsetTxHoldback(settings)   settings->flags_extend &= ~FLAG_TXHOLDBACK
 #define unsetVaryLoad(settings)     settings->flags_extend &= ~FLAG_VARYLOAD
 #define unsetFQPacing(settings)     settings->flags_extend &= ~FLAG_FQPACING
 #define unsetTripTime(settings)     settings->flags_extend &= ~FLAG_TRIPTIME
+#define unsetSmallTripTime(settings) settings->flags_extend2 &= ~FLAG_SMALLTRIPTIME
 #define unsetModeInfinite(settings) settings->flags_extend &= ~FLAG_MODEINFINITE
 #define unsetConnectOnly(settings)  settings->flags_extend &= ~FLAG_CONNECTONLY
 #define unsetWriteAck(settings)     settings->flags_extend &= ~FLAG_WRITEACK
 #define unsetNoUDPfin(settings)     settings->flags_extend &= ~FLAG_NOUDPFIN
 #define unsetNoConnectSync(settings) settings->flags_extend &= ~FLAG_NOCONNECTSYNC
-#define unsetSumOnly(settings) settings->flags_extend &= ~FLAG_SUMONLY
+#define unsetSumOnly(settings)       settings->flags_extend &= ~FLAG_SUMONLY
 #define unsetFrameInterval(settings) settings->flags_extend &= ~FLAG_FRAMEINTERVAL
-#define unsetIPG(settings) settings->flags_extend &= ~FLAG_IPG
+#define unsetIPG(settings)           settings->flags_extend &= ~FLAG_IPG
+#define unsetDontRoute(settings)     settings->flags_extend &= ~FLAG_DONTROUTE
+#define unsetPermitKey(settings)     settings->flags_extend &= ~FLAG_PERMITKEY
+#define unsetTCPMSS(settings)        settings->flags_extend &= ~FLAG_SETTCPMSS
+#define unsetPeriodicBurst(settings) settings->flags_extend2 &= ~FLAG_PERIODICBURST
+#define unsetSumServerDstIP(settings) settings->flags_extend2 &= ~FLAG_SUMDSTIP
+#define unsetRxClamp(settings)       settings->flags_extend2 &= ~FLAG_RXCLAMP
+#define unsetWritePrefetch(settings) settings->flags_extend2 &= ~FLAG_WRITEPREFETCH
+#define unsetTapDev(settings)        settings->flags_extend2 &= ~FLAG_TAPDEV
+#define unsetTunDev(settings)        settings->flags_extend2 &= ~FLAG_TUNDEV
+#define unsetHideIPs(settings)       settings->flags_extend2 &= ~FLAG_HIDEIPS
+#define unsetBounceBack(settings)    settings->flags_extend2 &= ~FLAG_BOUNCEBACK
+#define unsetTcpDrain(settings)      settings->flags_extend2 &= ~FLAG_TCPDRAIN
+#define unsetOverrideTOS(settings)   settings->flags_extend2 &= ~FLAG_OVERRIDETOS
 
 // set to defaults
 void Settings_Initialize(struct thread_Settings* main);
@@ -510,12 +623,12 @@ void Settings_GetUpperCaseArg(const char *,char *);
 void Settings_GenerateListenerSettings(struct thread_Settings *client, struct thread_Settings **listener);
 
 // generate settings for speaker instance
- void Settings_GenerateClientSettings(struct thread_Settings *server, struct thread_Settings **client, void * mBuf);
+void Settings_GenerateClientSettings(struct thread_Settings *server, struct thread_Settings **client, void * mBuf);
 
 // generate client header for server
 int Settings_GenerateClientHdr(struct thread_Settings *client, void * hdr, struct timeval startTime);
 
-int Settings_ClientHdrPeekLen(uint32_t flags);
+int Settings_ClientTestHdrLen(uint32_t flags, struct thread_Settings *inSettings);
 
 #ifdef __cplusplus
 } /* end extern "C" */
