@@ -68,7 +68,7 @@ public:
     /* -------------------------------------------------------------------
      * Create a timestamp, with the current time in it.
      * ------------------------------------------------------------------- */
-    Timestamp( void ) {
+    Timestamp(void) {
         setnow();
     }
 
@@ -83,21 +83,21 @@ public:
     /* -------------------------------------------------------------------
      * Create a timestamp, with the given seconds/microseconds
      * ------------------------------------------------------------------- */
-    Timestamp( long sec, long usec ) {
-        set( sec, usec );
+    Timestamp(long sec, long usec) {
+        set(sec, usec);
     }
 
     /* -------------------------------------------------------------------
      * Create a timestamp, with the given seconds
      * ------------------------------------------------------------------- */
-    Timestamp( double sec ) {
-        set( sec );
+    Timestamp(double sec) {
+        set(sec);
     }
 
     /* -------------------------------------------------------------------
      * Set timestamp to current time.
      * ------------------------------------------------------------------- */
-    void inline setnow( void ) {
+    void inline setnow(void) {
 #ifdef HAVE_CLOCK_GETTIME
 	struct timespec t1;
 	clock_gettime(CLOCK_REALTIME, &t1);
@@ -111,9 +111,9 @@ public:
     /* -------------------------------------------------------------------
      * Set timestamp to the given seconds/microseconds
      * ------------------------------------------------------------------- */
-    void set( long sec, long usec ) {
-        assert( sec  >= 0 );
-        assert( usec >= 0  &&  usec < kMillion );
+    void set(long sec, long usec) {
+        assert(sec  >= 0);
+        assert(usec >= 0  &&  usec < kMillion);
 
         mTime.tv_sec  = sec;
         mTime.tv_usec = usec;
@@ -122,7 +122,7 @@ public:
     /* -------------------------------------------------------------------
      * Set timestamp to the given seconds
      * ------------------------------------------------------------------- */
-    void set( double sec ) {
+    void set(double sec) {
         mTime.tv_sec  = (long) sec;
         mTime.tv_usec = (long) ((sec - mTime.tv_sec) * kMillion);
     }
@@ -130,21 +130,21 @@ public:
     /* -------------------------------------------------------------------
      * return seconds portion of timestamp
      * ------------------------------------------------------------------- */
-    long inline getSecs( void ) {
+    long inline getSecs(void) {
         return mTime.tv_sec;
     }
 
     /* -------------------------------------------------------------------
      * return microseconds portion of timestamp
      * ------------------------------------------------------------------- */
-    long inline getUsecs( void ) {
+    long inline getUsecs(void) {
         return mTime.tv_usec;
     }
 
     /* -------------------------------------------------------------------
      * return timestamp as a floating point seconds
      * ------------------------------------------------------------------- */
-    double get( void ) {
+    double get(void) {
         return mTime.tv_sec + mTime.tv_usec / ((double) kMillion);
     }
 
@@ -152,7 +152,7 @@ public:
      * subtract the right timestamp from my timestamp.
      * return the difference in microseconds.
      * ------------------------------------------------------------------- */
-    long subUsec( Timestamp right ) {
+    long subUsec(Timestamp right) {
         return(mTime.tv_sec  - right.mTime.tv_sec) * kMillion +
         (mTime.tv_usec - right.mTime.tv_usec);
     }
@@ -161,7 +161,7 @@ public:
      * subtract the right timestamp from my timestamp.
      * return the difference in microseconds.
      * ------------------------------------------------------------------- */
-    long subUsec( timeval right ) {
+    long subUsec(timeval right) {
         return(mTime.tv_sec  - right.tv_sec) * kMillion +
         (mTime.tv_usec - right.tv_usec);
     }
@@ -170,7 +170,7 @@ public:
      * subtract my timestamp from the right timestamp
      * return the difference in microseconds.
      * ------------------------------------------------------------------- */
-    long mysubUsec( timeval right ) {
+    long mysubUsec(timeval right) {
         return(right.tv_sec - mTime.tv_sec) * kMillion +
         (right.tv_usec - mTime.tv_usec);
     }
@@ -180,7 +180,6 @@ public:
      * ------------------------------------------------------------------- */
     long delta_usec(void) {
         struct timeval previous = mTime;
-
         setnow();
         return subUsec(previous);
     }
@@ -189,7 +188,7 @@ public:
      * subtract the right timestamp from my timestamp.
      * return the difference in seconds as a floating point.
      * ------------------------------------------------------------------- */
-    double subSec( Timestamp right ) {
+    double subSec(Timestamp right) {
         return(mTime.tv_sec  - right.mTime.tv_sec) +
         (mTime.tv_usec - right.mTime.tv_usec) / ((double) kMillion);
     }
@@ -197,46 +196,67 @@ public:
     /* -------------------------------------------------------------------
      * add the right timestamp to my timestamp.
      * ------------------------------------------------------------------- */
-    void add( Timestamp right ) {
+    void add(Timestamp right) {
         mTime.tv_sec  += right.mTime.tv_sec;
         mTime.tv_usec += right.mTime.tv_usec;
 
         // watch for under- and overflow
-        if ( mTime.tv_usec < 0 ) {
+        if (mTime.tv_usec < 0) {
             mTime.tv_usec += kMillion;
             mTime.tv_sec--;
         }
-        if ( mTime.tv_usec >= kMillion ) {
+        if (mTime.tv_usec >= kMillion) {
             mTime.tv_usec -= kMillion;
             mTime.tv_sec++;
         }
 
-        assert( mTime.tv_usec >= 0  &&
-                mTime.tv_usec <  kMillion );
+        assert(mTime.tv_usec >= 0  &&
+                mTime.tv_usec <  kMillion);
+    }
+
+    /* -------------------------------------------------------------------
+     * add the right timestamp to my timestamp.
+     * ------------------------------------------------------------------- */
+    void add (struct timeval *right) {
+        mTime.tv_sec  += right->tv_sec;
+        mTime.tv_usec += right->tv_usec;
+
+        // watch for under- and overflow
+        if (mTime.tv_usec < 0) {
+            mTime.tv_usec += kMillion;
+            mTime.tv_sec--;
+        }
+        if (mTime.tv_usec >= kMillion) {
+            mTime.tv_usec -= kMillion;
+            mTime.tv_sec++;
+        }
+
+        assert(mTime.tv_usec >= 0  &&
+                mTime.tv_usec <  kMillion);
     }
 
     /* -------------------------------------------------------------------
      * add the seconds to my timestamp.
      * TODO optimize?
      * ------------------------------------------------------------------- */
-    void add( double sec ) {
+    void add(double sec) {
         mTime.tv_sec  += (long) sec;
-        mTime.tv_usec += (long) ((sec - ((long) sec )) * kMillion);
+        mTime.tv_usec += (long) ((sec - ((long) sec)) * kMillion);
 
         // watch for overflow
-        if ( mTime.tv_usec >= kMillion ) {
+        if (mTime.tv_usec >= kMillion) {
             mTime.tv_usec -= kMillion;
             mTime.tv_sec++;
         }
 
-        assert( mTime.tv_usec >= 0  &&
-                mTime.tv_usec <  kMillion );
+        assert(mTime.tv_usec >= 0  &&
+                mTime.tv_usec <  kMillion);
     }
 
     /* -------------------------------------------------------------------
      * add micro seconds to my timestamp.
      * ------------------------------------------------------------------- */
-    void add(unsigned int usec ) {
+    void add(unsigned int usec) {
         mTime.tv_usec += usec;
 	mTime.tv_sec += mTime.tv_usec / kMillion;
 	mTime.tv_usec = mTime.tv_usec % kMillion;
@@ -246,29 +266,29 @@ public:
     /* -------------------------------------------------------------------
      * return true if my timestamp is before the right timestamp.
      * ------------------------------------------------------------------- */
-    bool before( timeval right ) {
+    bool before(timeval right) {
         return mTime.tv_sec < right.tv_sec  ||
         (mTime.tv_sec == right.tv_sec &&
          mTime.tv_usec < right.tv_usec);
     }
-    bool before( Timestamp right ) { return before(right.mTime); }
+    bool before(Timestamp right) { return before(right.mTime); }
 
     /* -------------------------------------------------------------------
      * return true if my timestamp is after the right timestamp.
      * ------------------------------------------------------------------- */
-    bool after( timeval right ) {
+    bool after(timeval right) {
         return mTime.tv_sec > right.tv_sec  ||
         (mTime.tv_sec == right.tv_sec &&
          mTime.tv_usec > right.tv_usec);
     }
-    bool after( Timestamp right ) { return after(right.mTime); }
+    bool after(Timestamp right) { return after(right.mTime); }
 
     /**
      * This function returns the fraction of time elapsed after the beginning
      * till the end
      */
     double fraction(Timestamp currentTime, Timestamp endTime) {
-        if ( (currentTime.after(*this)) && (endTime.after(currentTime)) ) {
+        if ((currentTime.after(*this)) && (endTime.after(currentTime))) {
             return(((double)currentTime.subUsec(*this)) /
                    ((double)endTime.subUsec(*this)));
         } else {
@@ -276,7 +296,6 @@ public:
         }
     }
 
-
 protected:
     enum {
         kMillion = 1000000
