@@ -55,11 +55,12 @@
 namespace Isochronous {
     class FrameCounter {
     public :
-        FrameCounter(double, Timestamp start);
+        FrameCounter(double, const Timestamp& start);
         FrameCounter(double);
-	unsigned int get(void);
+        ~FrameCounter();
+	unsigned int get(void) const;
 	unsigned int get(long *);
-	unsigned int get(Timestamp);
+	unsigned int get(const Timestamp&) const;
 	unsigned int period_us(void);
 	unsigned int wait_tick(void);
 	unsigned int wait_sync(long sec, long usec);
@@ -68,14 +69,19 @@ namespace Isochronous {
 	void reset(void);
         Timestamp next_slot(void);
 	unsigned int slip;
-
     private :
 	double frequency;
 	Timestamp startTime;
 	Timestamp nextslotTime;
-	unsigned int period;  // units microseconds
+	unsigned int period;  // units microseconds or 100 ns
 	unsigned int lastcounter;
 	unsigned int slot_counter;
+#ifdef WIN32
+	int mySetWaitableTimer (long delay_time);
+	HANDLE my_timer;	// Timer handle
+	LARGE_INTEGER delay;    // units is 100 nanoseconds
+#endif
+
     }; // end class FrameCounter
 }
 #endif // ISOCHRONOUS_H
