@@ -56,6 +56,14 @@ extern "C" {
 
 #define ACKRING_DEFAULTSIZE 100
 
+struct reportstruct_tcpstats {
+    bool isValid;
+    int cwnd;
+    int rtt;
+    int rttvar;
+    intmax_t retry_tot;
+};
+
 struct ReportStruct {
     intmax_t packetID;
     intmax_t packetLen;
@@ -75,6 +83,10 @@ struct ReportStruct {
     intmax_t burstperiod;
     intmax_t remaining;
     int transit_ready;
+    int writecnt;
+    struct reportstruct_tcpstats tcpstats;
+    double select_delay;
+    long drain_time;
 };
 
 struct PacketRing {
