@@ -53,18 +53,39 @@ extern "C" {
 #endif
 /*
  * Message header flags
- *
+ */
+
+#define TAPBYTESSLOP 512
+
+/* key for permit */
+#define HEADER_KEYCHECK     0x10000000
+#define HEADER_KEYL1        0x01000000
+#define HEADER_KEYL2        0x02000000
+#define HEADER_KEYLEN_MASK  0x0000FFFE // Lower bit used by V1 hdr
+#define DEFAULT_PERMITKEY_LEN 16
+#define MIN_PERMITKEY_LEN      4
+#define MAX_PERMITKEY_LEN    128
+
+/*
  * base flags, keep compatible with older versions
  */
-#define HEADER_VERSION1     0x80000000
-#define HEADER_EXTEND       0x40000000
-#define HEADER_UDPTESTS     0x20000000
-#define HEADER_SEQNO64B     0x08000000
-#define HEADER_VERSION2     0x04000000
-#define HEADER_V2PEERDETECT 0x02000000
-#define HEADER_LEN_BIT      0x00010000
+#define HEADER_VERSION1      0x80000000
+#define HEADER_EXTEND        0x40000000
+#define HEADER_UDPTESTS      0x20000000
+#define HEADER_BOUNCEBACK    0x10000000
+#define HEADER_SEQNO64B      0x08000000
+#define HEADER_VERSION2      0x04000000
+#define HEADER_V2PEERDETECT  0x02000000
+#define HEADER_UDPAVOID2     0x02000000
+#define HEADER_UDPAVOID1     0x01000000
+
+#define HEADER32_SMALL_TRIPTIMES 0x00020000
+#define HEADER_LEN_BIT       0x00010000
+#define HEADER_LEN_MASK      0x000001FE
+#define MAX_HEADER_LEN       256
 #define SERVER_HEADER_EXTEND 0x40000000
-#define RUN_NOW             0x00000001
+#define RUN_NOW              0x00000001
+#define HEADER16_SMALL_TRIPTIMES 0x0002 // use is 16 bits and not 32 bits
 
 // newer flags available per HEADER_EXTEND
 // Below flags are used to pass test settings in *every* UDP packet
@@ -74,13 +95,16 @@ extern "C" {
 #define HEADER_L2LENCHECK     0x0004
 #define HEADER_NOUDPFIN       0x0008
 #define HEADER_TRIPTIME       0x0010
-#define HEADER_TOSSTARTTIME   0x0020
+#define HEADER_UNUSED2        0x0020
 #define HEADER_ISOCH_SETTINGS 0x0040
 #define HEADER_UNITS_PPS      0x0080
 #define HEADER_BWSET          0x0100
 #define HEADER_FQRATESET      0x0200
 #define HEADER_REVERSE        0x0400
 #define HEADER_FULLDUPLEX     0x0800
+#define HEADER_EPOCH_START    0x1000
+#define HEADER_PERIODICBURST  0x2000
+#define HEADER_WRITEPREFETCH  0x4000
 
 // later features
 #define HDRXACKMAX 2500000 // default 2.5 seconds, units microseconds
@@ -98,6 +122,10 @@ enum MsgType {
     SERVERHDRACK
 };
 
+#define MINIPERFPAYLOAD 18
+// Minimum IPv4 frame size = 18 (Ethernet) + 20 (IPv4) + 8 (UDP) + 18 (payload) = 64 bytes
+// Minimum IPv6 frame size = 18 (Ethernet) + 40 (IPv6) + 8 (UDP) + 18 (payload) = 84 bytes
+
 /*
  * Structures below will be passed as network i/o
  * between the client, listener and server
@@ -158,6 +186,48 @@ struct client_hdr_v1 {
     int32_t mAmount;
 };
 
+//
+// Payload used for the bounce back feature
+//
+// Notes:
+//    o) flags is standard iperf 2 flags for test exchange info
+//    o) burst size is the payload size and size of bounce back write
+//    o) burst id is a running seq no
+//    o) send is the write timestamp
+//    o) bb_r2w_hold is an optional delay value between the read and bb write, typically expected as zero
+//    o) drain is the time from write to os indicating remote received payload (TCP only)
+//    o) drain times are deltas and not timestamps, computed by each end, units is microseconds
+//    o) triptimes will support OWD measurements in each direction (useful for asymmetry testing)
+//    o) min payload
+//         - seven 32b or 28 bytes with round trip only,
+//	   - nine 32b or 36 bytes when including drain times,
+//	   - eleven 32b or 44 bytes when including trip-times support
+//    o) no need for a bb read timestamp to be passed in the payload
+//    o) OWD calculations require e2e clock sync and --trip-times cli option
+//    o) no need to copy bb payload as rx buffer with be used for bounce back write
+//    o) single threaded design
+//    o) these are packed, be careful that the union doesn't break this
+//
+
+struct bb_ts {
+    uint32_t sec;
+    uint32_t usec;
+};
+union bb_r2w_info {
+    struct bb_ts bb_r2w_hold;
+    struct bb_ts bb_send;
+};
+struct bounce_back_datagram_hdr {
+    uint32_t flags;
+    uint32_t burst_size;
+    uint32_t burst_id;
+    struct bb_ts send_ts;
+    union bb_r2w_info bb_r2w; // up to here is mandatory
+    uint32_t drain; //units of usecs
+    uint32_t bb_drain;
+    struct bb_ts bb_read;
+};
+
 struct client_hdrext_isoch_settings {
     int32_t FPSl;
     int32_t FPSu;
@@ -180,7 +250,7 @@ struct client_hdrext {
     uint16_t tos;
     uint32_t lRate;
     uint32_t uRate;
-    int32_t Realtime;
+    uint32_t TCPWritePrefetch;
 };
 
 struct client_hdrext_starttime_fq {
@@ -366,6 +436,11 @@ struct isoch_payload {
     uint32_t reserved;
 };
 
+struct permitKey {
+    uint16_t length;
+    char value[MAX_PERMITKEY_LEN];
+};
+
 struct client_udp_testhdr {
     struct UDP_datagram seqno_ts;
     struct client_hdr_v1 base;
@@ -375,6 +450,20 @@ struct client_udp_testhdr {
     struct client_hdrext_isoch_settings isoch_settings;
 };
 
+struct client_udpsmall_testhdr {
+    struct UDP_datagram seqno_ts;
+    uint16_t flags;
+};
+
+struct client_hdr_ack_ts {
+    uint32_t sent_tv_sec;
+    uint32_t sent_tv_usec;
+    uint32_t sentrx_tv_sec;
+    uint32_t sentrx_tv_usec;
+    uint32_t ack_tv_sec;
+    uint32_t ack_tv_usec;
+};
+
 struct client_hdr_ack {
     struct hdr_typelen typelen;
     uint32_t flags;
@@ -382,6 +471,7 @@ struct client_hdr_ack {
     uint32_t version_l;
     uint32_t reserved1;
     uint32_t reserved2;
+    struct client_hdr_ack_ts ts;
 };
 
 /*
@@ -417,7 +507,7 @@ struct client_hdr_ack {
  *                +--------+--------+--------+--------+
  *            14  |        rate units                 |
  *                +--------+--------+--------+--------+
- *            15  |        realtime   (0.13)          |
+ *            15  |        write prefetch   (0.13)    |
  *                +--------+--------+--------+--------+
  *            16  |        reserved (0.14 start)      |
  *                +--------+--------+--------+--------+
@@ -451,6 +541,7 @@ struct client_tcp_testhdr {
     struct client_hdrext extend;
     struct client_hdrext_starttime_fq start_fq;
     struct client_hdrext_isoch_settings isoch_settings;
+    struct permitKey permitkey;
 };
 
 /*
@@ -499,7 +590,7 @@ struct server_hdr_extension {
     int32_t IPGsum;
 };
 
-  // Extension for 64bit datagram counts
+// Extension for 64bit datagram counts
 struct server_hdr_extension2 {
     int32_t error_cnt2;
     int32_t outorder_cnt2;
@@ -518,7 +609,8 @@ struct server_hdr {
 #define SIZEOF_UDPHDRMSG_EXT (sizeof(struct client_udp_testhdr))
 #define SIZEOF_TCPHDRMSG_V1 (sizeof(struct client_hdr_v1))
 #define SIZEOF_TCPHDRMSG_EXT (sizeof(struct client_tcp_testhdr))
-#define MINMBUFALLOCSIZE ((int) sizeof(struct client_udp_testhdr))
+#define MINMBUFALLOCSIZE (int) (sizeof(struct client_tcp_testhdr))
+#define MINTRIPTIMEPLAYOAD (int) (sizeof(struct client_udp_testhdr) - sizeof(struct client_hdrext_isoch_settings))
 #ifdef __cplusplus
 } /* end extern "C" */
 #endif
