@@ -73,9 +73,16 @@ int getsock_tcp_windowsize(int inSock, int inSend);
 void setsock_tcp_mss(int inSock, int inMSS);
 int  getsock_tcp_mss(int inSock);
 bool setsock_blocking(int fd, bool blocking);
-
+#if HAVE_DECL_TCP_WINDOW_CLAMP
+int  getsock_tcp_windowclamp(int inSock);
+int  setsock_tcp_windowclamp(int inSock, int clampsize);
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+int  getsock_tcp_notsent_low_watermark(int inSock);
+int  setsock_tcp_notsent_low_watermark(int inSock, int clampsize);
+#endif
 int recvn(int inSock, char *outBuf, int inLen, int flags);
-int writen(int inSock, const void *inBuf, int inLen);
+int writen(int inSock, const void *inBuf, int inLen, int *count);
 
 void disarm_itimer(void);
 /* -------------------------------------------------------------------
@@ -233,6 +240,30 @@ void redirect(const char *inOutputFileName);
     }                                           \
   } while(false)
 
+
+// Readn and write error macros
+// Define fatal and nonfatal write errors
+#ifdef WIN32
+#define FATALTCPREADERR(errno) (WSAGetLastError() != WSAEWOULDBLOCK)
+#define FATALUDPREADERR(errno)  (((errno = WSAGetLastError()) != WSAEWOULDBLOCK))
+#define FATALTCPWRITERR(errno)  ((errno = WSAGetLastError()) != WSAETIMEDOUT)
+#define NONFATALTCPWRITERR(errno) ((errno = WSAGetLastError()) == WSAETIMEDOUT)
+#define FATALUDPWRITERR(errno)  (((errno = WSAGetLastError()) != WSAETIMEDOUT) \
+				 && (errno != WSAECONNREFUSED))
+#else
+#define FATALTCPREADERR(errno) ((errno != EAGAIN) && (errno != EWOULDBLOCK) && (errno != EINTR))
+#define FATALUDPREADERR(errno) ((errno != EAGAIN) && (errno != EWOULDBLOCK) && \
+				(errno != EINTR))
+#define FATALTCPWRITERR(errno)  (errno != EAGAIN && errno != EWOULDBLOCK && errno != EINTR)
+#define NONFATALTCPWRITERR(errno)  (errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR)
+#define FATALUDPWRITERR(errno) 	((errno != EAGAIN) && (errno != EWOULDBLOCK) && (errno != EINTR) \
+				 && (errno != ECONNREFUSED) && (errno != ENOBUFS))
+#endif
+
+#ifdef WIN32
+#else
+#endif
+
 #ifdef __cplusplus
 } /* end extern "C" */
 #endif
