@@ -66,6 +66,7 @@
 #include "version.h"
 #include "payloads.h"
 #include "active_hosts.h"
+#include "gettcpinfo.h"
 
 // const double kSecs_to_usecs = 1e6;
 const double kSecs_to_nsecs = 1e9;
@@ -76,10 +77,9 @@ const int    kBytes_to_Bits = 8;
 
 Client::Client (thread_Settings *inSettings) {
 #ifdef HAVE_THREAD_DEBUG
-  thread_debug("Client constructor with thread %p sum=%p (flags=%x)", (void *) inSettings, (void *)inSettings->mSumReport, inSettings->flags);
+    thread_debug("Client constructor with thread %p sum=%p (flags=%x)", (void *) inSettings, (void *)inSettings->mSumReport, inSettings->flags);
 #endif
     mSettings = inSettings;
-    mBuf = NULL;
     myJob = NULL;
     myReport = NULL;
     framecounter = NULL;
@@ -97,13 +97,7 @@ Client::Client (thread_Settings *inSettings) {
 	unsetPeerVerDetect(mSettings);
     }
 
-    int payloadsize = (mSettings->mBufLen > MINMBUFALLOCSIZE) ? mSettings->mBufLen : MINMBUFALLOCSIZE; // defined in payloads.h
-    mBuf = new char[payloadsize]; // defined in payloads.h
-#ifdef HAVE_THREAD_DEBUG
-    thread_debug("Client constructor: MBUF malloc %d bytes (%p)", payloadsize, (void *) mBuf);
-#endif
-    FAIL_errno(mBuf == NULL, "No memory for buffer\n", mSettings);
-    pattern(mBuf, mSettings->mBufLen);
+    pattern(mSettings->mBuf, mSettings->mBufLen);
     if (isFileInput(mSettings)) {
         if (!isSTDIN(mSettings))
             Extractor_Initialize(mSettings->mFileName, mSettings->mBufLen, mSettings);
@@ -118,7 +112,7 @@ Client::Client (thread_Settings *inSettings) {
 	FAIL_errno(!(mSettings->mFPS > 0.0), "Invalid value for frames per second in the isochronous settings\n", mSettings);
     }
     peerclose = false;
-    isburst = (isIsochronous(mSettings) || isPeriodicBurst(mSettings) || (isTripTime(mSettings) && !isUDP(mSettings)));
+    isburst = (isIsochronous(mSettings) || isPeriodicBurst(mSettings) || ((isTripTime(mSettings) || isTcpDrain(mSettings)) && !isUDP(mSettings)));
 } // end Client
 
 /* -------------------------------------------------------------------
@@ -130,7 +124,6 @@ Client::~Client () {
 		 mySocket, (void *) mSettings->reporthdr, \
 		 (isServerReverse(mSettings) ? "true" : "false"), (isFullDuplex(mSettings) ? "true" : "false"));
 #endif
-    DELETE_ARRAY(mBuf);
     DELETE_PTR(framecounter);
 } // end ~Client
 
@@ -142,11 +135,10 @@ Client::~Client () {
  * ------------------------------------------------------------------- */
 bool Client::my_connect (bool close_on_fail) {
     int rc;
-    double connecttime = -1.0;
     // create an internet socket
     int type = (isUDP(mSettings) ? SOCK_DGRAM : SOCK_STREAM);
     int domain = (SockAddr_isIPv6(&mSettings->peer) ?
-#ifdef HAVE_IPV6
+#if HAVE_IPV6
                   AF_INET6
 #else
                   AF_INET
@@ -169,6 +161,7 @@ bool Client::my_connect (bool close_on_fail) {
 
     // connect socket
     connected = false;
+    my_init_cond.connecttime = -1;
     if (!isUDP(mSettings)) {
 	int trycnt = mSettings->mConnectRetries + 1;
 	while (trycnt > 0) {
@@ -187,8 +180,8 @@ bool Client::my_connect (bool close_on_fail) {
 		}
 	    } else {
 		connect_done.setnow();
-		connecttime = 1e3 * connect_done.subSec(connect_start);
-		mSettings->connecttime = connecttime;
+		my_init_cond.connecttime = 1e3 * connect_done.subSec(connect_start);
+		mSettings->connecttime = my_init_cond.connecttime;
 		connected = true;
 		break;
 	    }
@@ -196,12 +189,22 @@ bool Client::my_connect (bool close_on_fail) {
     } else {
 	rc = connect(mySocket, reinterpret_cast<sockaddr*>(&mSettings->peer),
 		     SockAddr_get_sizeof_sockaddr(&mSettings->peer));
-	connecttime = 0.0; // UDP doesn't have a 3WHS
+	my_init_cond.connecttime = 0.0; // UDP doesn't have a 3WHS
         WARN_errno((rc == SOCKET_ERROR), "udp connect");
 	if (rc != SOCKET_ERROR)
 	    connected = true;
     }
+    my_init_cond.rtt = -1;
+    my_init_cond.cwnd = -1;
     if (connected) {
+#if HAVE_TCP_STATS
+        assert(reportstruct);
+	if (!isUDP(mSettings) && connected) {
+	    gettcpinfo(mySocket, reportstruct);
+	    my_init_cond.rtt = reportstruct->tcpstats.rtt;
+	    my_init_cond.cwnd = reportstruct->tcpstats.cwnd;
+	}
+#endif
 	// Set the send timeout for the very first write which has the test exchange
 	int sosndtimer = TESTEXCHANGETIMEOUT; // 4 sec in usecs
 	SetSocketOptionsSendTimeout(mSettings, sosndtimer);
@@ -212,7 +215,6 @@ bool Client::my_connect (bool close_on_fail) {
 	    mSettings->mBurstIPG = get_delay_target() / 1e3; // this is being set for the settings report only
 	}
     } else {
-	connecttime = -1;
 	if (mySocket != INVALID_SOCKET) {
 	    int rc = close(mySocket);
 	    WARN_errno(rc == SOCKET_ERROR, "client connect close");
@@ -226,16 +228,16 @@ bool Client::my_connect (bool close_on_fail) {
 	setNoSettReport(mSettings);
     }
     // Post the connect report unless peer version exchange is set
-    if (isConnectionReport(mSettings) && !isSumOnly(mSettings) && !isPeerVerDetect(mSettings)) {
+    if (isConnectionReport(mSettings) && !isSumOnly(mSettings)) {
 	if (connected) {
-	    struct ReportHeader *reporthdr = InitConnectionReport(mSettings, connecttime);
+	    struct ReportHeader *reporthdr = InitConnectionReport(mSettings, &my_init_cond);
 	    struct ConnectionInfo *cr = static_cast<struct ConnectionInfo *>(reporthdr->this_report);
 	    cr->connect_timestamp.tv_sec = connect_start.getSecs();
 	    cr->connect_timestamp.tv_usec = connect_start.getUsecs();
 	    assert(reporthdr);
 	    PostReport(reporthdr);
 	} else {
-	    PostReport(InitConnectionReport(mSettings, -1));
+	    PostReport(InitConnectionReport(mSettings, &my_init_cond));
 	}
     }
     return connected;
@@ -254,28 +256,10 @@ void Client::TxDelay () {
     }
 }
 
-// return true of tcpi stats were sampled
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-inline bool Client::myReportPacket (bool sample_tcpi) {
-    bool rc = false;
-    if (sample_tcpi) {
-	rc = ReportPacket(myReport, reportstruct, &my_tcpi_stats);
-    } else {
-	ReportPacket(myReport, reportstruct, NULL);
-    }
-    reportstruct->packetLen = 0;
-    return rc;
-}
-inline void Client::myReportPacket () {
-    ReportPacket(myReport, reportstruct, NULL);
-    reportstruct->packetLen = 0;
-}
-#else
 inline void Client::myReportPacket (void) {
     ReportPacket(myReport, reportstruct);
     reportstruct->packetLen = 0;
 }
-#endif
 
 
 // There are multiple startup synchronizations, this code
@@ -289,6 +273,7 @@ int Client::StartSynch () {
     myJob = InitIndividualReport(mSettings);
     myReport = static_cast<struct ReporterData *>(myJob->this_report);
     myReport->info.common->socket=mySocket;
+    myReport->info.isEnableTcpInfo = false; // default here, set in init traffic actions
 
     // Perform delays, usually between connect() and data xfer though before connect
     // Two delays are supported:
@@ -312,7 +297,7 @@ int Client::StartSynch () {
 		    timeout.tv_sec = 0;
 		    timeout.tv_usec = rand() % 20000; // randomize IPG a bit
 		    if (select(mySocket + 1, &set, NULL, NULL, &timeout) == 0) {
-			SendFirstPayload();
+			reportstruct->packetLen = SendFirstPayload();
 			// printf("**** resend sock=%d count=%d\n", mySocket, resend_udp);
 		    } else {
 			break;
@@ -341,11 +326,28 @@ int Client::StartSynch () {
 	    return -1;
     }
     SetReportStartTime();
+#if HAVE_TCP_STATS
+    if (!isUDP(mSettings)) {
+	// Near congestion and peridiodic need sampling on every report packet
+	if (isNearCongest(mSettings) || isPeriodicBurst(mSettings)) {
+	    myReport->info.isEnableTcpInfo = true;
+	    myReport->info.ts.nextTCPStampleTime.tv_sec = 0;
+	    myReport->info.ts.nextTCPStampleTime.tv_usec = 0;
+	} else if (isEnhanced(mSettings)) {
+	    myReport->info.isEnableTcpInfo = true;
+	    myReport->info.ts.nextTCPStampleTime = myReport->info.ts.nextTime;
+	}
+    }
+#endif
+
     if (reportstruct->packetLen > 0) {
 	reportstruct->packetTime = myReport->info.ts.startTime;
 	reportstruct->sentTime = reportstruct->packetTime;
 	reportstruct->prevSentTime = reportstruct->packetTime;
 	reportstruct->prevPacketTime = myReport->info.ts.prevpacketTime;
+	if (isModeAmount(mSettings)) {
+	    mSettings->mAmount -= reportstruct->packetLen;
+	}
 	myReportPacket();
 	myReport->info.ts.prevpacketTime = reportstruct->packetTime;
 	reportstruct->packetID++;
@@ -461,9 +463,9 @@ void Client::InitTrafficLoop () {
     if (isPeriodicBurst(mSettings) && (mSettings->mFPS > 0.0)) {
 	sosndtimer = static_cast<int>(round(250000.0 / mSettings->mFPS));
     } else if (mSettings->mInterval > 0) {
-	sosndtimer = static_cast<int>(mSettings->mInterval / 2);
+        sosndtimer = static_cast<int>(round(0.5 * mSettings->mInterval));
     } else {
-	sosndtimer = static_cast<int>((mSettings->mAmount * 10000) / 2);
+	sosndtimer = static_cast<int>(mSettings->mAmount * 5e3);
     }
     SetSocketOptionsSendTimeout(mSettings, sosndtimer);
     // set the lower bounds delay based of the socket timeout timer
@@ -479,10 +481,8 @@ void Client::InitTrafficLoop () {
         mEndTime.setnow();
         mEndTime.add(mSettings->mAmount / 100.0);
     }
-    readAt = mBuf;
+    readAt = mSettings->mBuf;
     lastPacketTime.set(myReport->info.ts.startTime.tv_sec, myReport->info.ts.startTime.tv_usec);
-    if (isConnectionReport(mSettings) && isPeerVerDetect(mSettings) && !isSumOnly(mSettings))
-	PostReport(InitConnectionReport(mSettings, mSettings->connecttime));
     reportstruct->errwrite=WriteNoErr;
     reportstruct->emptyreport=0;
     reportstruct->packetLen = 0;
@@ -530,10 +530,15 @@ void Client::Run () {
 	}
     } else {
 	// Launch the approprate TCP traffic loop
-	if (mSettings->mAppRate > 0)
+	if (mSettings->mAppRate > 0) {
 	    RunRateLimitedTCP();
-	else if (isNearCongest(mSettings)) {
+	} else if (isNearCongest(mSettings)) {
 	    RunNearCongestionTCP();
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	} else if (isWritePrefetch(mSettings) && \
+		   !isIsochronous(mSettings) && !isPeriodicBurst(mSettings)) {
+	    RunWriteEventsTCP();
+#endif
 	} else {
 	    RunTCP();
 	}
@@ -547,10 +552,12 @@ void Client::RunTCP () {
     int burst_remaining = 0;
     int burst_id = 1;
     int writelen = mSettings->mBufLen;
+
     now.setnow();
     reportstruct->packetTime.tv_sec = now.getSecs();
     reportstruct->packetTime.tv_usec = now.getUsecs();
     while (InProgress()) {
+	reportstruct->writecnt = 0;
         if (isModeAmount(mSettings)) {
 	    writelen = ((mSettings->mAmount < static_cast<unsigned>(mSettings->mBufLen)) ? mSettings->mAmount : mSettings->mBufLen);
 	}
@@ -586,6 +593,11 @@ void Client::RunTCP () {
 			PostNullEvent();
 		    }
 		}
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+		if (isWritePrefetch(mSettings)) {
+		    AwaitWriteSelectEventTCP();
+		}
+#endif
 	    }
 	    now.setnow();
 	    reportstruct->packetTime.tv_sec = now.getSecs();
@@ -595,15 +607,25 @@ void Client::RunTCP () {
 	    myReport->info.ts.prevsendTime = reportstruct->packetTime;
 	    writelen = (mSettings->mBufLen > burst_remaining) ? burst_remaining : mSettings->mBufLen;
 	    // perform write, full header must succeed
-	    reportstruct->packetLen = writen(mySocket, mBuf, writelen);
-	    FAIL_errno(reportstruct->packetLen < (intmax_t) sizeof(struct TCP_burst_payload), "burst writen", mSettings);
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	    if (isTcpDrain(mSettings))
+		drain_start.setnow();
+#endif
+	    reportstruct->packetLen = writen(mySocket, mSettings->mBuf, writelen, &reportstruct->writecnt);
+	    FAIL_errno(reportstruct->packetLen < (intmax_t) sizeof(struct TCP_burst_payload), "burst written", mSettings);
 	} else {
 	    // printf("pl=%ld\n",reportstruct->packetLen);
 	    // perform write
 	    if (isburst)
 		writelen = (mSettings->mBufLen > burst_remaining) ? burst_remaining : mSettings->mBufLen;
-	    reportstruct->packetLen = write(mySocket, mBuf, writelen);
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	    if (isWritePrefetch(mSettings)) {
+		AwaitWriteSelectEventTCP();
+	    }
+#endif
+	    reportstruct->packetLen = write(mySocket, mSettings->mBuf, writelen);
 	    now.setnow();
+	    reportstruct->writecnt++;
 	    reportstruct->packetTime.tv_sec = now.getSecs();
 	    reportstruct->packetTime.tv_usec = now.getUsecs();
 	    reportstruct->sentTime = reportstruct->packetTime;
@@ -632,6 +654,11 @@ void Client::RunTCP () {
 		    reportstruct->transit_ready = 0;
 		} else {
 		    reportstruct->transit_ready = 1;
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+		    if (isTcpDrain(mSettings)) {
+			tcp_drain();
+		    }
+#endif
 		}
 	    }
 	}
@@ -660,6 +687,7 @@ void Client::RunNearCongestionTCP () {
     reportstruct->packetTime.tv_sec = now.getSecs();
     reportstruct->packetTime.tv_usec = now.getUsecs();
     while (InProgress()) {
+	reportstruct->writecnt = 0;
         if (isModeAmount(mSettings)) {
 	    reportstruct->packetLen = ((mSettings->mAmount < static_cast<unsigned>(mSettings->mBufLen)) ? mSettings->mAmount : mSettings->mBufLen);
 	} else {
@@ -674,9 +702,14 @@ void Client::RunNearCongestionTCP () {
 	    WriteTcpTxHdr(reportstruct, burst_remaining, burst_id++);
 	    reportstruct->sentTime = reportstruct->packetTime;
 	    myReport->info.ts.prevsendTime = reportstruct->packetTime;
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	    if (isTcpDrain(mSettings))
+		drain_start.setnow();
+#endif
 	    // perform write
 	    int writelen = (mSettings->mBufLen > burst_remaining) ? burst_remaining : mSettings->mBufLen;
-	    reportstruct->packetLen = write(mySocket, mBuf, writelen);
+	    reportstruct->packetLen = write(mySocket, mSettings->mBuf, writelen);
+	    reportstruct->writecnt++;
 	    assert(reportstruct->packetLen >= (intmax_t) sizeof(struct TCP_burst_payload));
 	    goto ReportNow;
 	}
@@ -685,8 +718,9 @@ void Client::RunNearCongestionTCP () {
 	}
 	// printf("pl=%ld\n",reportstruct->packetLen);
 	// perform write
-	reportstruct->packetLen = write(mySocket, mBuf, reportstruct->packetLen);
+	reportstruct->packetLen = write(mySocket, mSettings->mBuf, reportstruct->packetLen);
 	now.setnow();
+	reportstruct->writecnt++;
 	reportstruct->packetTime.tv_sec = now.getSecs();
 	reportstruct->packetTime.tv_usec = now.getUsecs();
 	reportstruct->sentTime = reportstruct->packetTime;
@@ -721,17 +755,19 @@ void Client::RunNearCongestionTCP () {
 		mSettings->mAmount = 0;
 	    }
 	}
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
 	// apply placing after write burst completes
-	if (reportstruct->transit_ready && myReportPacket(true)) {
-	    int pacing_timer = static_cast<int>(std::ceil(static_cast<double>(my_tcpi_stats.tcpi_rtt) * mSettings->rtt_nearcongest_divider));
-//		printf("**** delaytime = %d\n", delaytime);
-	    delay_loop(pacing_timer);
-	} else
+	if (reportstruct->transit_ready) {
+	    myReportPacket(); // this will set the tcpstats in the report struct
+	    // pacing timer is weighted by the RTT (set to 1 when RTT is not supported)
+	    int pacing_timer = 0;
+#if HAVE_TCP_STATS
+	    pacing_timer = static_cast<int>(std::ceil(static_cast<double>(reportstruct->tcpstats.rtt) * mSettings->rtt_nearcongest_weight_factor));
+#else
+	    pacing_timer = static_cast<int>(100 * mSettings->rtt_nearcongest_weight_factor);
 #endif
-        {
-	   myReportPacket();
-        }
+	    if (pacing_timer)
+		delay_loop(pacing_timer);
+	}
     }
     FinishTrafficActions();
 }
@@ -753,6 +789,7 @@ void Client::RunRateLimitedTCP () {
     reportstruct->packetTime.tv_sec = now.getSecs();
     reportstruct->packetTime.tv_usec = now.getUsecs();
     while (InProgress() && !fatalwrite_err) {
+	reportstruct->writecnt = 0;
 	// Add tokens per the loop time
 	time2.setnow();
         if (isVaryLoad(mSettings)) {
@@ -783,7 +820,7 @@ void Client::RunRateLimitedTCP () {
 		    reportstruct->sentTime = reportstruct->packetTime;
 		    burst_remaining = burst_size;
 		    // perform write
-		    n = writen(mySocket, mBuf, sizeof(struct TCP_burst_payload));
+		    n = writen(mySocket, mSettings->mBuf, sizeof(struct TCP_burst_payload), &reportstruct->writecnt);
 		    WARN(n != sizeof(struct TCP_burst_payload), "burst hdr write failed");
 		    burst_remaining -= n;
 		    reportstruct->packetLen -= n;
@@ -792,7 +829,8 @@ void Client::RunRateLimitedTCP () {
 		    reportstruct->packetLen = burst_remaining;
 		}
 	    }
-	    int len = write(mySocket, mBuf, reportstruct->packetLen);
+	    int len = write(mySocket, mSettings->mBuf, reportstruct->packetLen);
+	    reportstruct->writecnt++;
 	    if (len < 0) {
 	        if (NONFATALTCPWRITERR(errno)) {
 		    reportstruct->errwrite=WriteErrAccount;
@@ -838,6 +876,91 @@ void Client::RunRateLimitedTCP () {
     FinishTrafficActions();
 }
 
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+inline bool Client::AwaitWriteSelectEventTCP (void) {
+    int rc;
+    struct timeval timeout;
+    fd_set writeset;
+    FD_ZERO(&writeset);
+    FD_SET(mySocket, &writeset);
+    if (isModeTime(mSettings)) {
+        Timestamp write_event_timeout(0,0);
+	if (mSettings->mInterval && (mSettings->mIntervalMode == kInterval_Time)) {
+	    write_event_timeout.add((double) mSettings->mInterval / 1e6 * 2.0);
+	} else {
+	    write_event_timeout.add((double) mSettings->mAmount / 1e2 * 4.0);
+	}
+	timeout.tv_sec = write_event_timeout.getSecs();
+        timeout.tv_usec = write_event_timeout.getUsecs();
+    } else {
+	timeout.tv_sec = 10; // longest is 10 seconds
+	timeout.tv_usec = 0;
+    }
+
+    Timestamp t1;
+    if ((rc = select(mySocket + 1, NULL, &writeset, NULL, &timeout)) <= 0) {
+        reportstruct->select_delay = -1;
+	WARN_errno((rc < 0), "select");
+#ifdef HAVE_THREAD_DEBUG
+	if (rc == 0)
+	    thread_debug("AwaitWrite timeout");
+#endif
+	return false;
+    }
+    Timestamp t2;
+    reportstruct->select_delay = t2.subSec(t1);
+    //    printf("*****t1 = %f\n", t2.subSec(t1));
+    return true;
+}
+
+void Client::RunWriteEventsTCP () {
+    int burst_id = 0;
+    int writelen = mSettings->mBufLen;
+
+    now.setnow();
+    reportstruct->packetTime.tv_sec = now.getSecs();
+    reportstruct->packetTime.tv_usec = now.getUsecs();
+    while (InProgress()) {
+        if (isModeAmount(mSettings)) {
+	    writelen = ((mSettings->mAmount < static_cast<unsigned>(mSettings->mBufLen)) ? mSettings->mAmount : mSettings->mBufLen);
+	}
+	now.setnow();
+	bool rc = AwaitWriteSelectEventTCP();
+	reportstruct->emptyreport = (rc == false) ? 1 : 0;
+        if (rc) {
+	    now.setnow();
+	    reportstruct->packetTime.tv_sec = now.getSecs();
+	    reportstruct->packetTime.tv_usec = now.getUsecs();
+	    WriteTcpTxHdr(reportstruct, writelen, ++burst_id);
+	    reportstruct->sentTime = reportstruct->packetTime;
+	    reportstruct->packetLen = writen(mySocket, mSettings->mBuf, writelen, &reportstruct->writecnt);
+	    if (reportstruct->packetLen <= 0) {
+		WARN_errno((reportstruct->packetLen < 0), "event writen()");
+		if (reportstruct->packetLen == 0) {
+		    peerclose = true;
+		}
+		reportstruct->packetLen = 0;
+		reportstruct->emptyreport = 1;
+	    }
+	}
+	if (isModeAmount(mSettings) && !reportstruct->emptyreport) {
+	    /* mAmount may be unsigned, so don't let it underflow! */
+	    if (mSettings->mAmount >= static_cast<unsigned long>(reportstruct->packetLen)) {
+		mSettings->mAmount -= static_cast<unsigned long>(reportstruct->packetLen);
+	    } else {
+		mSettings->mAmount = 0;
+	    }
+	}
+	if (!one_report) {
+	    myReportPacket();
+	}
+    }
+    FinishTrafficActions();
+}
+#endif
+void Client::RunBounceBackTCP () {
+
+}
 /*
  * UDP send loop
  */
@@ -859,7 +982,7 @@ double Client::get_delay_target () {
 }
 
 void Client::RunUDP () {
-    struct UDP_datagram* mBuf_UDP = reinterpret_cast<struct UDP_datagram*>(mBuf);
+    struct UDP_datagram* mBuf_UDP = reinterpret_cast<struct UDP_datagram*>(mSettings->mBuf);
     int currLen;
 
     double delay_target = get_delay_target();
@@ -934,9 +1057,9 @@ void Client::RunUDP () {
 	reportstruct->emptyreport = 0;
 	// perform write
 	if (isModeAmount(mSettings)) {
-	    currLen = write(mySocket, mBuf, (mSettings->mAmount < static_cast<unsigned>(mSettings->mBufLen)) ? mSettings->mAmount : mSettings->mBufLen);
+	    currLen = write(mySocket, mSettings->mBuf, (mSettings->mAmount < static_cast<unsigned>(mSettings->mBufLen)) ? mSettings->mAmount : mSettings->mBufLen);
 	} else {
-	    currLen = write(mySocket, mBuf, mSettings->mBufLen);
+	    currLen = write(mySocket, mSettings->mBuf, mSettings->mBufLen);
 	}
 	if (currLen < 0) {
 	    reportstruct->packetID--;
@@ -981,9 +1104,9 @@ void Client::RunUDP () {
  * UDP isochronous send loop
  */
 void Client::RunUDPIsochronous () {
-    struct UDP_datagram* mBuf_UDP = reinterpret_cast<struct UDP_datagram*>(mBuf);
+    struct UDP_datagram* mBuf_UDP = reinterpret_cast<struct UDP_datagram*>(mSettings->mBuf);
     // skip over the UDP datagram (seq no, timestamp) to reach the isoch fields
-    struct client_udp_testhdr *udp_payload = reinterpret_cast<client_udp_testhdr *>(mBuf);
+    struct client_udp_testhdr *udp_payload = reinterpret_cast<client_udp_testhdr *>(mSettings->mBuf);
 
     double delay_target = mSettings->mBurstIPG * 1000000;  // convert from milliseconds to nanoseconds
     double delay = 0;
@@ -1062,11 +1185,11 @@ void Client::RunUDPIsochronous () {
 	    if (isModeAmount(mSettings) && (mSettings->mAmount < static_cast<unsigned>(mSettings->mBufLen))) {
 	        udp_payload->isoch.remaining = htonl(mSettings->mAmount);
 		reportstruct->remaining=mSettings->mAmount;
-	        currLen = write(mySocket, mBuf, mSettings->mAmount);
+	        currLen = write(mySocket, mSettings->mBuf, mSettings->mAmount);
 	    } else {
 	        udp_payload->isoch.remaining = htonl(bytecnt);
 		reportstruct->remaining=bytecnt;
-	        currLen = write(mySocket, mBuf, (bytecnt < mSettings->mBufLen) ? bytecnt : mSettings->mBufLen);
+	        currLen = write(mySocket, mSettings->mBuf, (bytecnt < mSettings->mBufLen) ? bytecnt : mSettings->mBufLen);
 	    }
 
 	    if (currLen < 0) {
@@ -1123,7 +1246,7 @@ void Client::RunUDPIsochronous () {
 // end RunUDPIsoch
 
 inline void Client::WritePacketID (intmax_t packetID) {
-    struct UDP_datagram * mBuf_UDP = reinterpret_cast<struct UDP_datagram *>(mBuf);
+    struct UDP_datagram * mBuf_UDP = reinterpret_cast<struct UDP_datagram *>(mSettings->mBuf);
     // store datagram ID into buffer
 #ifdef HAVE_INT64_T
     // Pack signed 64bit packetID into unsigned 32bit id1 + unsigned
@@ -1146,7 +1269,7 @@ inline void Client::WritePacketID (intmax_t packetID) {
 }
 
 inline void Client::WriteTcpTxHdr (struct ReportStruct *reportstruct, int burst_size, int burst_id) {
-    struct TCP_burst_payload * mBuf_burst = reinterpret_cast<struct TCP_burst_payload *>(mBuf);
+    struct TCP_burst_payload * mBuf_burst = reinterpret_cast<struct TCP_burst_payload *>(mSettings->mBuf);
     // store packet ID into buffer
     reportstruct->packetID += burst_size;
     mBuf_burst->start_tv_sec = htonl(myReport->info.ts.startTime.tv_sec);
@@ -1182,7 +1305,7 @@ inline void Client::WriteTcpTxHdr (struct ReportStruct *reportstruct, int burst_
 //    printf("**** Write tcp burst header size= %d id = %d\n", burst_size, burst_id);
 }
 
-inline bool Client::InProgress () {
+inline bool Client::InProgress (void) {
     // Read the next data block from
     // the file if it's file input
     if (isFileInput(mSettings)) {
@@ -1194,6 +1317,29 @@ inline bool Client::InProgress () {
 	(isModeAmount(mSettings) && (mSettings->mAmount <= 0)));
 }
 
+inline void Client::tcp_drain (void) {
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    AwaitWriteSelectEventTCP();
+    drain_end.setnow();
+    reportstruct->drain_time = drain_end.subUsec(drain_start);
+#ifdef HAVE_THREAD_DEBUG
+    thread_debug("Drain time  = %f", reportstruct->drain_time);
+#endif
+#endif
+}
+
+inline void Client::tcp_shutdown (void) {
+    if ((mySocket != INVALID_SOCKET) && isConnected()) {
+	int rc = shutdown(mySocket, SHUT_WR);
+#ifdef HAVE_THREAD_DEBUG
+	thread_debug("Client calls shutdown() SHUTW_WR on tcp socket %d", mySocket);
+#endif
+	WARN_errno(rc == SOCKET_ERROR, "shutdown");
+	if (!rc && !isFullDuplex(mSettings))
+	    AwaitServerCloseEvent();
+    }
+}
+
 /*
  * Common things to do to finish a traffic thread
  *
@@ -1210,18 +1356,7 @@ void Client::FinishTrafficActions () {
     disarm_itimer();
     // Shutdown the TCP socket's writes as the event for the server to end its traffic loop
     if (!isUDP(mSettings)) {
-	if ((mySocket != INVALID_SOCKET) && isConnected()) {
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-	  // gettcpistats(myReport, true, NULL);
-#endif
-	    int rc = shutdown(mySocket, SHUT_WR);
-#ifdef HAVE_THREAD_DEBUG
-	    thread_debug("Client calls shutdown() SHUTW_WR on tcp socket %d", mySocket);
-#endif
-	    WARN_errno(rc == SOCKET_ERROR, "shutdown");
-	    if (!rc && !isFullDuplex(mSettings))
-		AwaitServerCloseEvent();
-	}
+	tcp_shutdown();
 	now.setnow();
 	reportstruct->packetTime.tv_sec = now.getSecs();
 	reportstruct->packetTime.tv_usec = now.getUsecs();
@@ -1244,10 +1379,10 @@ void Client::FinishTrafficActions () {
 	// but didn't count our first datagram, so we're even now.
 	// The negative datagram ID signifies termination to the server.
 	WritePacketID(-reportstruct->packetID);
-	struct UDP_datagram * mBuf_UDP = reinterpret_cast<struct UDP_datagram *>(mBuf);
+	struct UDP_datagram * mBuf_UDP = reinterpret_cast<struct UDP_datagram *>(mSettings->mBuf);
 	mBuf_UDP->tv_sec = htonl(reportstruct->packetTime.tv_sec);
 	mBuf_UDP->tv_usec = htonl(reportstruct->packetTime.tv_usec);
-	int len = write(mySocket, mBuf, mSettings->mBufLen);
+	int len = write(mySocket, mSettings->mBuf, mSettings->mBufLen);
 #ifdef HAVE_THREAD_DEBUG
 	thread_debug("UDP client sent final packet per negative seqno %ld", -reportstruct->packetID);
 #endif
@@ -1267,7 +1402,7 @@ void Client::FinishTrafficActions () {
     }
     if (do_close) {
 #if HAVE_THREAD_DEBUG
-	thread_debug("UDP client close sock=%d", mySocket);
+	thread_debug("client close sock=%d", mySocket);
 #endif
 	int rc = close(mySocket);
 	WARN_errno(rc == SOCKET_ERROR, "client close");
@@ -1283,18 +1418,20 @@ void Client::FinishTrafficActions () {
  * stats to displayed on the client.  Attempt to re-transmit
  * until the fin is received
  * ------------------------------------------------------------------- */
+#define RETRYTIMER 10000 //units of us
+#define RETRYCOUNT (2 * 1000000 / RETRYTIMER) // 2 seconds worth of retries
 void Client::AwaitServerFinPacket () {
     int rc;
     fd_set readSet;
     struct timeval timeout;
     int ack_success = 0;
-    int count = 20 ;
+    int count = RETRYCOUNT;
     while (--count >= 0) {
         // wait until the socket is readable, or our timeout expires
         FD_ZERO(&readSet);
         FD_SET(mySocket, &readSet);
         timeout.tv_sec  = 0;
-        timeout.tv_usec = 100000; // 100 milliseconds
+        timeout.tv_usec = RETRYTIMER;
         rc = select(mySocket+1, &readSet, NULL, NULL, &timeout);
         FAIL_errno(rc == SOCKET_ERROR, "select", mSettings);
         // rc= zero means select's read timed out
@@ -1302,7 +1439,7 @@ void Client::AwaitServerFinPacket () {
 	    // try to trigger another FIN by resending a negative seq no
 	    WritePacketID(-(++reportstruct->packetID));
 	    // write data
-	    rc = write(mySocket, mBuf, mSettings->mBufLen);
+	    rc = write(mySocket, mSettings->mBuf, mSettings->mBufLen);
 	    WARN_errno(rc < 0, "write-fin");
 #ifdef HAVE_THREAD_DEBUG
 	    thread_debug("UDP client retransmit final packet per negative seqno %ld", -reportstruct->packetID);
@@ -1311,11 +1448,11 @@ void Client::AwaitServerFinPacket () {
             // socket ready to read, this packet size
 	    // is set by the server.  Assume it's large enough
 	    // to contain the final server packet
-	    rc = read(mySocket, mBuf, MAXUDPBUF);
+	    rc = read(mySocket, mSettings->mBuf, MAXUDPBUF);
 
 	    // dump any 2.0.13 client acks sent at the start of traffic
 	    if (rc == sizeof(client_hdr_ack)) {
-		struct client_hdr_ack *ack =  reinterpret_cast<struct client_hdr_ack *>(mBuf);
+		struct client_hdr_ack *ack =  reinterpret_cast<struct client_hdr_ack *>(mSettings->mBuf);
 		if (ntohl(ack->typelen.type) == CLIENTHDRACK) {
 		    // printf("**** dump stale ack \n");
 		    continue;
@@ -1329,7 +1466,7 @@ void Client::AwaitServerFinPacket () {
 		thread_debug("UDP client received server relay report ack (%d)", -reportstruct->packetID);
 #endif
 		if (mSettings->mReportMode != kReport_CSV) {
-		    PostReport(InitServerRelayUDPReport(mSettings, reinterpret_cast<server_hdr*>(reinterpret_cast<UDP_datagram*>(mBuf) + 1)));
+		    PostReport(InitServerRelayUDPReport(mSettings, reinterpret_cast<server_hdr*>(reinterpret_cast<UDP_datagram*>(mSettings->mBuf) + 1)));
 		}
 		break;
 	    }
@@ -1367,13 +1504,16 @@ void Client::AwaitServerCloseEvent () {
 	amount_usec = MINAWAITCLOSEUSECS;
     SetSocketOptionsReceiveTimeout(mSettings, amount_usec);
     int rc;
-    while ((rc = recv(mySocket, mBuf, mSettings->mBufLen, 0) > 0)) {};
+    while ((rc = recv(mySocket, mSettings->mBuf, mSettings->mBufLen, 0) > 0)) {};
     if (rc < 0)
 	WARN_errno(1, "client await server close");
+
+    if (rc==0) {
+	connected = false;
 #ifdef HAVE_THREAD_DEBUG
-    if (rc==0)
 	thread_debug("Client detected server close %d", mySocket);
 #endif
+    }
 }
 
 int Client::SendFirstPayload () {
@@ -1387,32 +1527,52 @@ int Client::SendFirstPayload () {
 	    reportstruct->packetTime.tv_usec = now.getUsecs();
 	}
 	if (isTxStartTime(mSettings)) {
-	    pktlen += Settings_GenerateClientHdr(mSettings, (void *) mBuf, mSettings->txstart_epoch);
+	    pktlen = Settings_GenerateClientHdr(mSettings, (void *) mSettings->mBuf, mSettings->txstart_epoch);
 	} else {
-	    pktlen += Settings_GenerateClientHdr(mSettings, (void *) mBuf, reportstruct->packetTime);
+	    pktlen = Settings_GenerateClientHdr(mSettings, (void *) mSettings->mBuf, reportstruct->packetTime);
 	}
 	if (pktlen > 0) {
 	    if (isUDP(mSettings)) {
-		struct client_udp_testhdr *tmphdr = reinterpret_cast<struct client_udp_testhdr *>(mBuf);
+		struct client_udp_testhdr *tmphdr = reinterpret_cast<struct client_udp_testhdr *>(mSettings->mBuf);
 		WritePacketID(reportstruct->packetID);
 		tmphdr->seqno_ts.tv_sec  = htonl(reportstruct->packetTime.tv_sec);
 		tmphdr->seqno_ts.tv_usec = htonl(reportstruct->packetTime.tv_usec);
 		udp_payload_minimum = pktlen;
 #if HAVE_DECL_MSG_DONTWAIT
-		pktlen = send(mySocket, mBuf, (pktlen > mSettings->mBufLen) ? pktlen : mSettings->mBufLen, MSG_DONTWAIT);
+		pktlen = send(mySocket, mSettings->mBuf, (pktlen > mSettings->mBufLen) ? pktlen : mSettings->mBufLen, MSG_DONTWAIT);
 #else
-		pktlen = send(mySocket, mBuf, (pktlen > mSettings->mBufLen) ? pktlen : mSettings->mBufLen, 0);
+		pktlen = send(mySocket, mSettings->mBuf, (pktlen > mSettings->mBufLen) ? pktlen : mSettings->mBufLen, 0);
 #endif
 		apply_first_udppkt_delay = true;
 	    } else {
+#if HAVE_DECL_TCP_NODELAY
+		if (!isNoDelay(mSettings) && isPeerVerDetect(mSettings) && isTripTime(mSettings)) {
+		    int optflag=1;
+		    int rc;
+		    // Disable Nagle to reduce latency of this intial message
+		    if ((rc = setsockopt(mSettings->mSock, IPPROTO_TCP, TCP_NODELAY, reinterpret_cast<char *>(&optflag), sizeof(int))) < 0) {
+			WARN_errno(rc < 0, "tcpnodelay");
+		    }
+		}
+#endif
 #if HAVE_DECL_MSG_DONTWAIT
-		pktlen = send(mySocket, mBuf, pktlen, MSG_DONTWAIT);
+		pktlen = send(mySocket, mSettings->mBuf, pktlen, MSG_DONTWAIT);
 #else
-		pktlen = send(mySocket, mBuf, pktlen, 0);
+		pktlen = send(mySocket, mSettings->mBuf, pktlen, 0);
 #endif
 		if (isPeerVerDetect(mSettings) && !isServerReverse(mSettings)) {
 		    PeerXchange();
 		}
+#if HAVE_DECL_TCP_NODELAY
+		if (!isNoDelay(mSettings) && isPeerVerDetect(mSettings) && isTripTime(mSettings)) {
+		    int optflag=0;
+		    int rc;
+		    // Disable Nagle to reduce latency of this intial message
+		    if ((rc = setsockopt(mSettings->mSock, IPPROTO_TCP, TCP_NODELAY, reinterpret_cast<char *>(&optflag), sizeof(int))) < 0) {
+			WARN_errno(rc < 0, "tcpnodelay");
+		    }
+		}
+#endif
 	    }
 	    WARN_errno(pktlen < 0, "send_hdr");
 	}
@@ -1426,10 +1586,23 @@ void Client::PeerXchange () {
     /*
      * Hang read and see if this is a header ack message
      */
-    if ((n = recvn(mySocket, reinterpret_cast<char *>(&ack), sizeof(client_hdr_ack), 0)) == sizeof(client_hdr_ack)) {
-	if (ntohl(ack.typelen.type) == CLIENTHDRACK && ntohl(ack.typelen.length) == sizeof(client_hdr_ack)) {
+    int readlen = isTripTime(mSettings) ? sizeof(struct client_hdr_ack) : (sizeof(struct client_hdr_ack) - sizeof(struct client_hdr_ack_ts));
+    if ((n = recvn(mySocket, reinterpret_cast<char *>(&ack), readlen, 0)) == readlen) {
+	if (ntohl(ack.typelen.type) == CLIENTHDRACK) {
 	    mSettings->peer_version_u = ntohl(ack.version_u);
 	    mSettings->peer_version_l = ntohl(ack.version_l);
+	    if (isTripTime(mSettings)) {
+		Timestamp now;
+		Timestamp senttx(ntohl(ack.ts.sent_tv_sec), ntohl(ack.ts.sent_tv_usec));
+		Timestamp sentrx(ntohl(ack.ts.sentrx_tv_sec), ntohl(ack.ts.sentrx_tv_usec));
+		Timestamp acktx(ntohl(ack.ts.ack_tv_sec), ntohl(ack.ts.ack_tv_usec));
+		Timestamp ackrx(now.getSecs(), now.getUsecs());
+		double str = (sentrx.get() - senttx.get()) * 1e3;
+		double atr = (now.get() - acktx.get()) * 1e3;
+		double rtt = str + atr;
+		double halfrtt = rtt / 2.0;
+		fprintf(stderr,"%sClock sync check (ms): RTT/Half=(%0.3f/%0.3f) OWD-send/ack/asym=(%0.3f/%0.3f/%0.3f)\n",mSettings->mTransferIDStr, rtt, halfrtt, str, atr, (str-atr));
+	    }
 	}
     } else {
 	WARN_errno(1, "recvack");
