@@ -83,9 +83,7 @@
 #include "SocketAddr.h"
 #include "payloads.h"
 #include "delay.h"
-#if (defined HAVE_SSM_MULTICAST) && (defined HAVE_NET_IF_H)
-#include <net/if.h>
-#endif
+
 /* -------------------------------------------------------------------
 
  * Stores local hostname and socket info.
@@ -102,10 +100,6 @@ Listener::Listener (thread_Settings *inSettings) {
      * 3) Server thread
      */
     mSettings = inSettings;
-    // alloc and initialize the buffer (mBuf) used for test messages in the payload
-    mBufLen = (mSettings->mBufLen > MINMBUFALLOCSIZE) ? mSettings->mBufLen : MINMBUFALLOCSIZE;
-    mBuf = new char[mBufLen]; // defined in payloads.h
-    FAIL_errno(mBuf == NULL, "No memory for buffer\n", mSettings);
 } // end Listener
 
 /* -------------------------------------------------------------------
@@ -119,7 +113,6 @@ Listener::~Listener () {
         int rc = close(ListenSocket);
         WARN_errno(rc == SOCKET_ERROR, "listener close");
     }
-    DELETE_ARRAY(mBuf);
 } // end ~Listener
 
 /* -------------------------------------------------------------------
@@ -288,9 +281,9 @@ void Listener::Run () {
 	// Note 2: The mBuf read is a peek so the server's traffic thread started later
 	// will also process the first message from an accounting perspective.
 	// This is required for accurate traffic statistics
-	if (!isCompat(server) && !apply_client_settings(server)) {
+	if (!isCompat(server) && (isConnectOnly(server) || !apply_client_settings(server))) {
 	    if (isConnectionReport(server) && !isSumOnly(server)) {
-		struct ReportHeader *reporthdr = InitConnectionReport(server, 0);
+		struct ReportHeader *reporthdr = InitConnectionReport(server, NULL);
 		struct ConnectionInfo *cr = static_cast<struct ConnectionInfo *>(reporthdr->this_report);
 		cr->connect_timestamp.tv_sec = server->accept_time.tv_sec;
 		cr->connect_timestamp.tv_usec = server->accept_time.tv_usec;
@@ -308,8 +301,8 @@ void Listener::Run () {
 	}
 	// server settings flags should now be set per the client's first message exchange
 	// so the server setting's flags per the client can now be checked
-	if (isUDP(server) && !isCompat(mSettings)  && (isL2LengthCheck(mSettings) || isL2LengthCheck(server))) {
-	    if (!L2_setup(server, server->mSock)) {
+	if (isUDP(server)){
+	    if (!isCompat(mSettings) && !isTapDev(mSettings) && (isL2LengthCheck(mSettings) || isL2LengthCheck(server)) && !L2_setup(server, server->mSock)) {
 		// Requested L2 testing but L2 setup failed
 		Iperf_remove_host(server);
 		if (DecrSumReportRefCounter(server->mSumReport) <= 0) {
@@ -324,6 +317,7 @@ void Listener::Run () {
 	if (isUDP(server) && isCompat(mSettings)) {
 	    setSeqNo64b(server);
 	}
+	setTransferID(server, 0);
 
 	// Read any more test settings and test values (not just the flags) and instantiate
 	// any settings objects for client threads (e.g. bidir or full duplex)
@@ -338,7 +332,7 @@ void Listener::Run () {
 	    (isFullDuplex(server) || isServerReverse(server) || (server->mMode != kTest_Normal))) {
 	    thread_Settings *listener_client_settings = NULL;
 	    // read client header for reverse settings
-	    Settings_GenerateClientSettings(server, &listener_client_settings, mBuf);
+	    Settings_GenerateClientSettings(server, &listener_client_settings, server->mBuf);
 	    if (listener_client_settings) {
 		if (server->mMode != kTest_Normal)
 		    listener_client_settings->mTransferID = 0;
@@ -378,7 +372,7 @@ void Listener::Run () {
 	}
 	setTransferID(server, 0);
 	if (isConnectionReport(server) && !isSumOnly(server)) {
-	    struct ReportHeader *reporthdr = InitConnectionReport(server, 0);
+	    struct ReportHeader *reporthdr = InitConnectionReport(server, NULL);
 	    struct ConnectionInfo *cr = static_cast<struct ConnectionInfo *>(reporthdr->this_report);
 	    cr->connect_timestamp.tv_sec = server->accept_time.tv_sec;
 	    cr->connect_timestamp.tv_usec = server->accept_time.tv_usec;
@@ -403,15 +397,27 @@ void Listener::Run () {
  * ------------------------------------------------------------------- */
 void Listener::my_listen () {
     int rc;
-
+    int type;
+    int domain;
     SockAddr_localAddr(mSettings);
 
-    // create an AF_INET socket for the accepts
-    // for the case of L2 testing and UDP, a new AF_PACKET
-    // will be created to supercede this one
-    int type = (isUDP(mSettings)  ?  SOCK_DGRAM  :  SOCK_STREAM);
-    int domain = (SockAddr_isIPv6(&mSettings->local) ?
-#ifdef HAVE_IPV6
+#if (((HAVE_TUNTAP_TUN) || (HAVE_TUNTAP_TAP)) && (AF_PACKET))
+    if (isTapDev(mSettings)) {
+	ListenSocket = socket(AF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
+	FAIL_errno(ListenSocket == SOCKET_ERROR, "tuntap socket()", mSettings);
+	mSettings->mSock = ListenSocket;
+	rc = SockAddr_v4_Accept_BPF(ListenSocket, mSettings->mPort);
+	WARN_errno((rc == SOCKET_ERROR), "tap accept bpf");
+	SetSocketOptions(mSettings);
+    } else
+#endif
+    {
+	// create an AF_INET socket for the accepts
+	// for the case of L2 testing and UDP, a new AF_PACKET
+	// will be created to supercede this one
+	type = (isUDP(mSettings)  ?  SOCK_DGRAM  :  SOCK_STREAM);
+	domain = (SockAddr_isIPv6(&mSettings->local) ?
+#if HAVE_IPV6
 		  AF_INET6
 #else
 		  AF_INET
@@ -419,37 +425,36 @@ void Listener::my_listen () {
 		  : AF_INET);
 
 #ifdef WIN32
-    if (SockAddr_isMulticast(&mSettings->local)) {
-	// Multicast on Win32 requires special handling
-	ListenSocket = WSASocket(domain, type, 0, 0, 0, WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF);
-	WARN_errno(ListenSocket == INVALID_SOCKET, "socket");
-
-    } else
-#endif
-	{
-	    ListenSocket = socket(domain, type, 0);
+	if (SockAddr_isMulticast(&mSettings->local)) {
+	    // Multicast on Win32 requires special handling
+	    ListenSocket = WSASocket(domain, type, 0, 0, 0, WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF);
 	    WARN_errno(ListenSocket == INVALID_SOCKET, "socket");
-	}
-    mSettings->mSock = ListenSocket;
 
-    SetSocketOptions(mSettings);
-
-    // reuse the address, so we can run if a former server was killed off
-    int boolean = 1;
-    Socklen_t len = sizeof(boolean);
-    rc = setsockopt(ListenSocket, SOL_SOCKET, SO_REUSEADDR, reinterpret_cast<char*>(&boolean), len);
-    // bind socket to server address
-#ifdef WIN32
-    if (SockAddr_isMulticast(&mSettings->local)) {
-	// Multicast on Win32 requires special handling
-	rc = WSAJoinLeaf(ListenSocket, (sockaddr*) &mSettings->local, mSettings->size_local,0,0,0,0,JL_BOTH);
-	WARN_errno(rc == SOCKET_ERROR, "WSAJoinLeaf (aka bind)");
-    } else
+	} else
 #endif
-	{
-	    rc = bind(ListenSocket, reinterpret_cast<sockaddr*>(&mSettings->local), mSettings->size_local);
-	    FAIL_errno(rc == SOCKET_ERROR, "bind", mSettings);
-	}
+	    {
+		ListenSocket = socket(domain, type, 0);
+		WARN_errno(ListenSocket == INVALID_SOCKET, "socket");
+	    }
+	mSettings->mSock = ListenSocket;
+	SetSocketOptions(mSettings);
+	// reuse the address, so we can run if a former server was killed off
+	int boolean = 1;
+	Socklen_t len = sizeof(boolean);
+	rc = setsockopt(ListenSocket, SOL_SOCKET, SO_REUSEADDR, reinterpret_cast<char*>(&boolean), len);
+	// bind socket to server address
+#ifdef WIN32
+	if (SockAddr_isMulticast(&mSettings->local)) {
+	    // Multicast on Win32 requires special handling
+	    rc = WSAJoinLeaf(ListenSocket, (sockaddr*) &mSettings->local, mSettings->size_local,0,0,0,0,JL_BOTH);
+	    WARN_errno(rc == SOCKET_ERROR, "WSAJoinLeaf (aka bind)");
+	} else
+#endif
+	    {
+		rc = bind(ListenSocket, reinterpret_cast<sockaddr*>(&mSettings->local), mSettings->size_local);
+		FAIL_errno(rc == SOCKET_ERROR, "listener bind", mSettings);
+	    }
+    }
 
     // update the reporter thread
     if (isReport(mSettings) && isSettingsReport(mSettings)) {
@@ -537,7 +542,7 @@ void Listener::my_multicast_join () {
 	    WARN_errno(rc == SOCKET_ERROR, "ip_multicast_all");
 #endif
 	} else {
-#if (HAVE_IPV6_MULTICAST && (HAVE_DECL_IPV6_JOIN_GROUP || HAVE_DECL_IPV6_ADD_MEMBERSHIP))
+#if (HAVE_IPV6 && HAVE_IPV6_MULTICAST && (HAVE_DECL_IPV6_JOIN_GROUP || HAVE_DECL_IPV6_ADD_MEMBERSHIP))
 	    struct ipv6_mreq mreq;
 	    memcpy(&mreq.ipv6mr_multiaddr, SockAddr_get_in6_addr(&mSettings->local), sizeof(mreq.ipv6mr_multiaddr));
 	    mreq.ipv6mr_interface = 0;
@@ -571,7 +576,7 @@ void Listener::my_multicast_join () {
 #endif
 
         if (isIPV6(mSettings)) {
-#ifdef HAVE_IPV6_MULTICAST
+#if HAVE_IPV6_MULTICAST
 	    if (mSettings->mSSMMulticastStr) {
 		struct group_source_req group_source_req;
 		struct sockaddr_in6 *group;
@@ -801,7 +806,7 @@ bool Listener::L2_setup (thread_Settings *server, int sockfd) {
     // Now optimize packet flow up the raw socket
     // Establish the flow BPF to forward up only "connected" packets to this raw socket
     if (l->sa_family == AF_INET6) {
-#ifdef HAVE_IPV6
+#if HAVE_IPV6
 	struct in6_addr *v6peer = SockAddr_get_in6_addr(&server->peer);
 	struct in6_addr *v6local = SockAddr_get_in6_addr(&server->local);
 	if (isIPV6(server)) {
@@ -827,6 +832,58 @@ bool Listener::L2_setup (thread_Settings *server, int sockfd) {
 #endif
 }
 
+bool Listener::tap_setup (thread_Settings *server, int sockfd) {
+#if defined(HAVE_IF_TUNTAP) && defined(HAVE_AF_PACKET) && defined(HAVE_DECL_SO_BINDTODEVICE)
+    struct sockaddr *p = reinterpret_cast<sockaddr *>(&server->peer);
+    struct sockaddr *l = reinterpret_cast<sockaddr *>(&server->local);
+    int rc = 0;
+
+    //
+    // Establish a packet (raw) socket to be used by the server thread giving it full L2 packets
+    //
+    struct sockaddr s;
+    socklen_t len = sizeof(s);
+    getpeername(sockfd, &s, &len);
+    if (isIPV6(server)) {
+	server->l4offset = IPV6HDRLEN + sizeof(struct ether_header);
+    } else {
+	server->l4offset = sizeof(struct iphdr) + sizeof(struct ether_header);
+    }
+    // Didn't get a valid socket, return now
+    if (server->mSock < 0) {
+	return false;
+    }
+    // More per thread settings based on using a packet socket
+    server->l4payloadoffset = server->l4offset + sizeof(struct udphdr);
+    server->recvflags = MSG_TRUNC;
+    // Now optimize packet flow up the raw socket
+    // Establish the flow BPF to forward up only "connected" packets to this raw socket
+    if (l->sa_family == AF_INET6) {
+#if HAVE_IPV6
+	struct in6_addr *v6peer = SockAddr_get_in6_addr(&server->peer);
+	struct in6_addr *v6local = SockAddr_get_in6_addr(&server->local);
+	if (isIPV6(server)) {
+	    rc = SockAddr_v6_Connect_BPF(server->mSock, v6local, v6peer, (reinterpret_cast<struct sockaddr_in6 *>(l))->sin6_port, (reinterpret_cast<struct sockaddr_in6 *>(p))->sin6_port);
+	    WARN_errno(rc == SOCKET_ERROR, "l2 connect ipv6 bpf");
+	} else {
+	    // This is an ipv4 address in a v6 family (structure), just pull the lower 32 bits for the v4 addr
+	    rc = SockAddr_v4_Connect_BPF(server->mSock, v6local->s6_addr32[3], v6peer->s6_addr32[3], (reinterpret_cast<struct sockaddr_in6 *>(l))->sin6_port, (reinterpret_cast<struct sockaddr_in6 *>(p))->sin6_port);
+	    WARN_errno(rc == SOCKET_ERROR, "l2 v4in6 connect ip bpf");
+	}
+#else
+	fprintf(stderr, "Unfortunately, IPv6 is not supported on this platform\n");
+	return false;
+#endif /* HAVE_IPV6 */
+    } else {
+	rc = SockAddr_v4_Connect_BPF(server->mSock, (reinterpret_cast<struct sockaddr_in *>(l))->sin_addr.s_addr, (reinterpret_cast<struct sockaddr_in *>(p))->sin_addr.s_addr, (reinterpret_cast<struct sockaddr_in *>(l))->sin_port, (reinterpret_cast<struct sockaddr_in *>(p))->sin_port);
+	WARN_errno(rc == SOCKET_ERROR, "l2 connect ip bpf");
+    }
+    return rc >= 0;
+#else
+    fprintf(stderr, "Client requested --l2checks but not supported on this platform\n");
+    return false;
+#endif
+}
 
 /* ------------------------------------------------------------------------
  * Do the equivalent of an accept() call for UDP sockets. This checks
@@ -834,24 +891,28 @@ bool Listener::L2_setup (thread_Settings *server, int sockfd) {
  * ------------------------------------------------------------------- ----*/
 int Listener::udp_accept (thread_Settings *server) {
     assert(server != NULL);
-    int rc;
+    int nread = 0;
     assert(ListenSocket > 0);
     // Preset the server socket to INVALID, hang recvfrom on the Listener's socket
     // The INVALID socket is used to keep the while loop going
     server->mSock = INVALID_SOCKET;
-    rc = recvfrom(ListenSocket, mBuf, mBufLen, MSG_PEEK, \
-		  reinterpret_cast<struct sockaddr*>(&server->peer), &server->size_peer);
+    nread = recvfrom(ListenSocket, server->mBuf, server->mBufLen, 0, \
+		     reinterpret_cast<struct sockaddr*>(&server->peer), &server->size_peer);
+    Timestamp now;
+    server->accept_time.tv_sec = now.getSecs();
+    server->accept_time.tv_usec = now.getUsecs();
+
 #if HAVE_THREAD_DEBUG
     {
 	char tmpaddr[200];
 	size_t len=200;
 	unsigned short port = SockAddr_getPort(&server->peer);
 	SockAddr_getHostAddress(&server->peer, tmpaddr, len);
-	thread_debug("rcvfrom peer: %s port %d len=%d", tmpaddr, port, rc);
+	thread_debug("rcvfrom peer: %s port %d len=%d", tmpaddr, port, nread);
     }
 #endif
-    FAIL_errno(rc == SOCKET_ERROR, "recvfrom", mSettings);
-    if (!(rc < 0) && !sInterupted) {
+    FAIL_errno(nread == SOCKET_ERROR, "recvfrom", mSettings);
+    if (!(nread < 0) && !sInterupted) {
 	// Handle connection for UDP sockets
 	int gid = Iperf_push_host_port_conditional(server);
 #if HAVE_THREAD_DEBUG
@@ -859,6 +920,7 @@ int Listener::udp_accept (thread_Settings *server) {
 	    thread_debug("rcvfrom peer: drop duplicate");
 #endif
 	if (gid > 0) {
+	    int rc;
 	    // We have a new UDP flow (based upon key of quintuple)
 	    // so let's hand off this socket
 	    // to the server and create a new listener socket
@@ -875,10 +937,41 @@ int Listener::udp_accept (thread_Settings *server) {
 	    server->size_local = sizeof(iperf_sockaddr);
 	    getsockname(server->mSock, reinterpret_cast<sockaddr*>(&server->local), &server->size_local);
 	    SockAddr_Ifrname(server);
+	    server->firstreadbytes = nread;
 	}
     }
     return server->mSock;
 }
+
+
+#if (((HAVE_TUNTAP_TUN) || (HAVE_TUNTAP_TAP)) && (AF_PACKET))
+int Listener::tuntap_accept(thread_Settings *server) {
+    int rc = recv(server->mSock, server->mBuf, (server->mBufLen + TAPBYTESSLOP + sizeof(struct iphdr) + sizeof(struct ether_header) + sizeof(struct udphdr)), 0);
+    if (rc <= 0)
+	return 0;
+//	rc = udpchecksum((void *)ip_hdr, (void *)udp_hdr, udplen, (isIPV6(mSettings) ? 1 : 0));
+    struct iphdr *l3hdr = (struct iphdr *)((char *)server->mBuf + sizeof(struct ether_header));
+    struct udphdr *l4hdr = (struct udphdr *)((char *)server->mBuf + sizeof(struct iphdr) + sizeof(struct ether_header));
+//    uint16_t ipver = (uint16_t) ntohs(mBuf + sizeof(struct ether_header));
+//    printf ("*** version = %d\n", ipver);
+    // Note: sockaddrs are stored in network bytes order
+    struct sockaddr_in *local = (struct sockaddr_in *) &server->local;
+    struct sockaddr_in *peer = (struct sockaddr_in *) &server->peer;
+    server->size_peer = sizeof(iperf_sockaddr);
+    server->size_local = sizeof(iperf_sockaddr);
+    peer->sin_family = AF_INET;
+    local->sin_family = AF_INET;
+    peer->sin_addr.s_addr = l3hdr->saddr;
+    local->sin_addr.s_addr = l3hdr->daddr;
+    peer->sin_port = l4hdr->source;
+    local->sin_port = l4hdr->dest;
+    server->l4offset = sizeof(struct iphdr) + sizeof(struct ether_header);
+    SockAddr_v4_Connect_TAP_BPF(server->mSock, local->sin_addr.s_addr, peer->sin_addr.s_addr, local->sin_port, peer->sin_port);
+    server->l4payloadoffset = sizeof(struct iphdr) + sizeof(struct ether_header) + sizeof(struct udphdr);
+    server->firstreadbytes = rc;
+    return server->mSock;
+}
+#endif
 /* -------------------------------------------------------------------
  * This is called by the Listener thread main loop, return a socket or error
  * ------------------------------------------------------------------- */
@@ -896,23 +989,28 @@ int Listener::my_accept (thread_Settings *server) {
     server->accept_time.tv_sec = 0;
     server->accept_time.tv_usec = 0;
     if (isUDP(server)) {
-	server->mSock = udp_accept(server);
+#if (((HAVE_TUNTAP_TUN) || (HAVE_TUNTAP_TAP)) && (AF_PACKET))
+	if (isTapDev(server) || isTunDev(server)) {
+	    server->mSock = tuntap_accept(server);
+	} else
+#endif
+	{
+	    server->mSock = udp_accept(server);
+	}
 	// note udp_accept will update the active host table
     } else {
 	// accept a TCP  connection
 	server->mSock = accept(ListenSocket, reinterpret_cast<sockaddr*>(&server->peer), &server->size_peer);
 	if (server->mSock > 0) {
+	    Timestamp now;
+	    server->accept_time.tv_sec = now.getSecs();
+	    server->accept_time.tv_usec = now.getUsecs();
 	    server->size_local = sizeof(iperf_sockaddr);
 	    getsockname(server->mSock, reinterpret_cast<sockaddr*>(&server->local), &server->size_local);
 	    SockAddr_Ifrname(server);
 	    Iperf_push_host(server);
 	}
     }
-    if (server->mSock > 0) {
-	Timestamp now;
-	server->accept_time.tv_sec = now.getSecs();
-	server->accept_time.tv_usec = now.getUsecs();
-    }
     return server->mSock;
 } // end my_accept
 
@@ -924,8 +1022,7 @@ int Listener::my_accept (thread_Settings *server) {
 // Description of bits and fields is in include/payloads.h
 bool Listener::apply_client_settings (thread_Settings *server) {
     assert(server != NULL);
-    assert(mBuf != NULL);
-    bool rc;
+    bool rc = false;
 
     // Set the receive timeout for the very first read
     int sorcvtimer = TESTEXCHANGETIMEOUT; // 4 sec in usecs
@@ -936,9 +1033,14 @@ bool Listener::apply_client_settings (thread_Settings *server) {
 
     if (isUDP(server)) {
 	rc = apply_client_settings_udp(server);
-    } else {
+    } else if (!isConnectOnly(server)) {
 	rc = apply_client_settings_tcp(server);
     }
+    if (isOverrideTOS(server)) {
+	SetSocketOptionsIPTos(server, server->mRTOS);
+    } else if (server->mTOS) {
+	SetSocketOptionsIPTos(server, server->mTOS);
+    }
     return rc;
 }
 
@@ -947,7 +1049,7 @@ inline bool Listener::test_permit_key(uint32_t flags, thread_Settings *server, i
 	server->mKeyCheck= false;
 	return false;
     }
-    struct permitKey *thiskey = reinterpret_cast<struct permitKey *>(mBuf + (keyoffset - sizeof(thiskey->length)));
+    struct permitKey *thiskey = reinterpret_cast<struct permitKey *>(server->mBuf + (keyoffset - sizeof(thiskey->length)));
     int keylen = ntohs(thiskey->length);
     if ((keylen < MIN_PERMITKEY_LEN) || (keylen > MAX_PERMITKEY_LEN)) {
 	server->mKeyCheck= false;
@@ -960,9 +1062,9 @@ inline bool Listener::test_permit_key(uint32_t flags, thread_Settings *server, i
 	return false;
     }
     if (!isUDP(server)) {
-	int n = recvn(server->mSock, mBuf, keyoffset + keylen, MSG_PEEK);
-	FAIL_errno((n < (keyoffset + keylen)), "read key", server);
-	server->skip = n;
+	int nread = 0;
+	nread = recvn(server->mSock, server->mBuf, keyoffset + keylen, 0);
+	FAIL_errno((nread < (keyoffset + keylen)), "read key", server);
     }
     strncpy(server->mPermitKey, thiskey->value, MAX_PERMITKEY_LEN + 1);
     if (strncmp(server->mPermitKey, mSettings->mPermitKey, keylen) != 0) {
@@ -975,15 +1077,21 @@ inline bool Listener::test_permit_key(uint32_t flags, thread_Settings *server, i
 }
 
 bool Listener::apply_client_settings_udp (thread_Settings *server) {
-    struct client_udp_testhdr *hdr = reinterpret_cast<struct client_udp_testhdr *>(mBuf);
+    struct client_udp_testhdr *hdr = reinterpret_cast<struct client_udp_testhdr *>(server->mBuf + server->l4payloadoffset);
     uint32_t flags = ntohl(hdr->base.flags);
     uint16_t upperflags = 0;
     if (flags & HEADER_SEQNO64B) {
 	setSeqNo64b(server);
     }
+#if HAVE_THREAD_DEBUG
+    thread_debug("UDP test flags = %X", flags);
+#endif
     if (flags & HEADER32_SMALL_TRIPTIMES) {
-	server->accept_time.tv_sec = ntohl(hdr->seqno_ts.tv_sec);
-	server->accept_time.tv_usec = ntohl(hdr->seqno_ts.tv_usec);
+#if HAVE_THREAD_DEBUG
+        thread_debug("UDP small header");
+#endif
+	server->sent_time.tv_sec = ntohl(hdr->seqno_ts.tv_sec);
+	server->sent_time.tv_usec = ntohl(hdr->seqno_ts.tv_usec);
 	uint32_t seqno = ntohl(hdr->seqno_ts.id);
 	if (seqno != 1) {
 	    fprintf(stderr, "WARN: first received packet (id=%d) was not first sent packet, report start time will be off\n", seqno);
@@ -1031,10 +1139,10 @@ bool Listener::apply_client_settings_udp (thread_Settings *server) {
 		}
 	    }
 	    if (upperflags & HEADER_TRIPTIME) {
-		server->accept_time.tv_sec = ntohl(hdr->start_fq.start_tv_sec);
-		server->accept_time.tv_usec = ntohl(hdr->start_fq.start_tv_usec);
+		server->sent_time.tv_sec = ntohl(hdr->start_fq.start_tv_sec);
+		server->sent_time.tv_usec = ntohl(hdr->start_fq.start_tv_usec);
 		Timestamp now;
-		if (!isTxStartTime(server) && ((abs(now.getSecs() - server->accept_time.tv_sec)) > (MAXDIFFTIMESTAMPSECS + 1))) {
+		if (!isTxStartTime(server) && ((abs(now.getSecs() - server->sent_time.tv_sec)) > (MAXDIFFTIMESTAMPSECS + 1))) {
 		    fprintf(stdout,"WARN: ignore --trip-times because client didn't provide valid start timestamp within %d seconds of now\n", MAXDIFFTIMESTAMPSECS);
 		} else {
 		    setTripTime(server);
@@ -1059,104 +1167,124 @@ bool Listener::apply_client_settings_udp (thread_Settings *server) {
     return true;
 }
 bool Listener::apply_client_settings_tcp (thread_Settings *server) {
-    bool rc;
-    int n = recvn(server->mSock, mBuf, sizeof(uint32_t), MSG_PEEK);
-    if (n == 0) {
+    bool rc = false;
+    int nread = recvn(server->mSock, server->mBuf, sizeof(uint32_t), 0);
+    char *readptr = server->mBuf;
+    if (nread == 0) {
 	//peer closed the socket, with no writes e.g. a connect-only test
-	rc = false;
+	WARN(1, "read tcp flags (peer close)");
+	goto DONE;
+    }
+    if (nread < (int) sizeof(uint32_t)) {
+	WARN(1, "read tcp flags (runt)");
+	goto DONE;
     } else {
-	FAIL_errno((n < (int) sizeof(uint32_t)), "read tcp flags", server);
 	rc = true;
-	struct client_tcp_testhdr *hdr = reinterpret_cast<struct client_tcp_testhdr *>(mBuf);
+	readptr += nread;
+	struct client_tcp_testhdr *hdr = reinterpret_cast<struct client_tcp_testhdr *>(server->mBuf);
 	uint32_t flags = ntohl(hdr->base.flags);
+	if (flags & HEADER_BOUNCEBACK) {
+	    setBounceBack(server);
+	}
 	uint16_t upperflags = 0;
-	int peeklen;
-	if ((flags & HEADER_VERSION1) || (flags & HEADER_VERSION2) || (flags & HEADER_EXTEND) || isPermitKey(mSettings)) {
-	    // figure out the length of the test header
-	    if ((peeklen = Settings_ClientHdrPeekLen(flags)) > 0) {
-		// read the test settings passed to the server by the client
-		int n = recvn(server->mSock, mBuf, peeklen, MSG_PEEK);
-		FAIL_errno((n < peeklen), "read tcp test info", server);
-		server->skip = n;
-		if (isPermitKey(mSettings)) {
-		    if (!test_permit_key(flags, server, peeklen)) {
-			rc = false;
-			goto DONE;
-		    }
-		} else if (flags & HEADER_KEYCHECK) {
+	int readlen;
+	// figure out the length of the test header
+	if ((readlen = Settings_ClientTestHdrLen(flags, server)) > 0) {
+	    // read the test settings passed to the server by the client
+	    nread += recvn(server->mSock, readptr, (readlen - (int) sizeof(uint32_t)), 0);
+	    FAIL_errno((nread < readlen), "read tcp test info", server);
+	    if (isPermitKey(mSettings)) {
+		if (!test_permit_key(flags, server, readlen)) {
 		    rc = false;
-		    server->mKeyCheck = false;
 		    goto DONE;
 		}
-		struct client_tcp_testhdr *hdr = reinterpret_cast<struct client_tcp_testhdr*>(mBuf);
-		if ((flags & HEADER_VERSION1) && !(flags & HEADER_VERSION2)) {
-		    if (flags & RUN_NOW)
-			server->mMode = kTest_DualTest;
-		    else
-			server->mMode = kTest_TradeOff;
+	    } else if (flags & HEADER_KEYCHECK) {
+		rc = false;
+		server->mKeyCheck = false;
+		goto DONE;
+	    }
+	    server->firstreadbytes = nread;
+	    struct client_tcp_testhdr *hdr = reinterpret_cast<struct client_tcp_testhdr*>(server->mBuf);
+	    if ((flags & HEADER_VERSION1) && !(flags & HEADER_VERSION2)) {
+		if (flags & RUN_NOW)
+		    server->mMode = kTest_DualTest;
+		else
+		    server->mMode = kTest_TradeOff;
+	    }
+	    if (flags & HEADER_EXTEND) {
+		upperflags = htons(hdr->extend.upperflags);
+		server->mTOS = ntohs(hdr->extend.tos);
+		server->peer_version_u = ntohl(hdr->extend.version_u);
+		server->peer_version_l = ntohl(hdr->extend.version_l);
+		if (upperflags & HEADER_ISOCH) {
+		    setIsochronous(server);
 		}
-		if (flags & HEADER_EXTEND) {
-		    upperflags = htons(hdr->extend.upperflags);
-		    server->mTOS = ntohs(hdr->extend.tos);
-		    server->peer_version_u = ntohl(hdr->extend.version_u);
-		    server->peer_version_l = ntohl(hdr->extend.version_l);
-		    if (upperflags & HEADER_ISOCH) {
-			setIsochronous(server);
+		if (upperflags & HEADER_EPOCH_START) {
+		    server->txstart_epoch.tv_sec = ntohl(hdr->start_fq.start_tv_sec);
+		    server->txstart_epoch.tv_usec = ntohl(hdr->start_fq.start_tv_usec);
+		    Timestamp now;
+		    if ((abs(now.getSecs() - server->txstart_epoch.tv_sec)) > (MAXDIFFTXSTART + 1)) {
+			fprintf(stdout,"WARN: ignore --txstart-time because client didn't provide valid start timestamp within %d seconds of now\n", MAXDIFFTXSTART);
+			unsetTxStartTime(server);
+		    } else {
+			setTxStartTime(server);
 		    }
-		    if (upperflags & HEADER_EPOCH_START) {
-			server->txstart_epoch.tv_sec = ntohl(hdr->start_fq.start_tv_sec);
-			server->txstart_epoch.tv_usec = ntohl(hdr->start_fq.start_tv_usec);
-			Timestamp now;
-			if ((abs(now.getSecs() - server->txstart_epoch.tv_sec)) > (MAXDIFFTXSTART + 1)) {
-			    fprintf(stdout,"WARN: ignore --txstart-time because client didn't provide valid start timestamp within %d seconds of now\n", MAXDIFFTXSTART);
-			    unsetTxStartTime(server);
-			} else {
-			    setTxStartTime(server);
-			}
-		    }
-		    if (upperflags & HEADER_TRIPTIME) {
-			Timestamp now;
-			if (!isTxStartTime(server) && ((abs(now.getSecs() - server->accept_time.tv_sec)) > (MAXDIFFTIMESTAMPSECS + 1))) {
-			    fprintf(stdout,"WARN: ignore --trip-times because client didn't provide valid start timestamp within %d seconds of now\n", MAXDIFFTIMESTAMPSECS);
-			} else {
-			    setTripTime(server);
-			    setEnhanced(server);
-			}
-		    }
-		    if (upperflags & HEADER_PERIODICBURST) {
+		}
+		if (upperflags & HEADER_TRIPTIME) {
+		    Timestamp now;
+		    server->sent_time.tv_sec = ntohl(hdr->start_fq.start_tv_sec);
+		    server->sent_time.tv_usec = ntohl(hdr->start_fq.start_tv_usec);
+		    if (!isTxStartTime(server) && ((abs(now.getSecs() - server->sent_time.tv_sec)) > (MAXDIFFTIMESTAMPSECS + 1))) {
+			fprintf(stdout,"WARN: ignore --trip-times because client didn't provide valid start timestamp within %d seconds of now\n", MAXDIFFTIMESTAMPSECS);
+		    } else {
+			setTripTime(server);
 			setEnhanced(server);
-			setFrameInterval(server);
-			setPeriodicBurst(server);
-			{
-			    struct client_tcp_testhdr *hdr = reinterpret_cast<struct client_tcp_testhdr *>(mBuf);
-			    server->mFPS = ntohl(hdr->isoch_settings.FPSl);
-			    server->mFPS += ntohl(hdr->isoch_settings.FPSu) / static_cast<double>(rMillion);
-			}
-			if (!server->mFPS) {
-			    server->mFPS = 1.0;
-			}
-		    }
-		    if (flags & HEADER_VERSION2) {
-			if (upperflags & HEADER_FULLDUPLEX) {
-			    setFullDuplex(server);
-			    setServerReverse(server);
-			}
-			if (upperflags & HEADER_REVERSE) {
-			    server->mThreadMode=kMode_Client;
-			    setServerReverse(server);
-			}
 		    }
 		}
+		if (upperflags & HEADER_PERIODICBURST) {
+		    setEnhanced(server);
+		    setFrameInterval(server);
+		    setPeriodicBurst(server);
+		    {
+			struct client_tcp_testhdr *hdr = reinterpret_cast<struct client_tcp_testhdr *>(server->mBuf);
+			server->mFPS = ntohl(hdr->isoch_settings.FPSl);
+			server->mFPS += ntohl(hdr->isoch_settings.FPSu) / static_cast<double>(rMillion);
+		    }
+		    if (!server->mFPS) {
+			server->mFPS = 1.0;
+		    }
+		}
+		if (flags & HEADER_VERSION2) {
+		    if (upperflags & HEADER_FULLDUPLEX) {
+			setFullDuplex(server);
+			setServerReverse(server);
+		    }
+		    if (upperflags & HEADER_REVERSE) {
+			server->mThreadMode=kMode_Client;
+			setServerReverse(server);
+		    }
+		}
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+		if ((isServerReverse(server) || isFullDuplex(server)) && (upperflags & HEADER_WRITEPREFETCH)) {
+		    server->mWritePrefetch = ntohl(hdr->extend.TCPWritePrefetch);
+		    if (server->mWritePrefetch > 0) {
+			setWritePrefetch(server);
+		    }
+		}
+#endif
+		if (upperflags & HEADER_BOUNCEBACK) {
+		    setBounceBack(server);
+		}
 	    }
-	    // Handle case that requires an ack back to the client
-	    // Signaled by not UDP (only supported by TCP)
-	    // and either 2.0.13 flags or the newer 2.0.14 flag of
-	    // V2PEERDETECT
-	    if (!isUDP(server) && !isCompat(mSettings) && \
-		((!(flags & HEADER_VERSION2) && (flags & HEADER_EXTEND)) || \
-		 (flags & HEADER_V2PEERDETECT))) {
-		client_test_ack(server);
-	    }
+	}
+	// Handle case that requires an ack back to the client
+	// Signaled by not UDP (only supported by TCP)
+	// and either 2.0.13 flags or the newer 2.0.14 flag of
+	// V2PEERDETECT
+	if (!isUDP(server) && !isCompat(mSettings) && \
+	    ((!(flags & HEADER_VERSION2) && (flags & HEADER_EXTEND)) || \
+	     (flags & HEADER_V2PEERDETECT))) {
+	    client_test_ack(server);
 	}
     }
   DONE:
@@ -1164,49 +1292,65 @@ bool Listener::apply_client_settings_tcp (thread_Settings *server) {
 }
 
 int Listener::client_test_ack(thread_Settings *server) {
+    if (isUDP(server))
+	return 1;
+
     client_hdr_ack ack;
     int sotimer = 0;
-    int optflag;
+    int size = sizeof(struct client_hdr_ack);
     ack.typelen.type  = htonl(CLIENTHDRACK);
-    ack.typelen.length = htonl(sizeof(client_hdr_ack));
+
     ack.flags = 0;
     ack.reserved1 = 0;
     ack.reserved2 = 0;
     ack.version_u = htonl(IPERF_VERSION_MAJORHEX);
     ack.version_l = htonl(IPERF_VERSION_MINORHEX);
+    if (isTripTime(server)) {
+	ack.ts.sent_tv_sec = htonl(server->sent_time.tv_sec);
+	ack.ts.sent_tv_usec = htonl(server->sent_time.tv_usec);
+	ack.ts.sentrx_tv_sec = htonl(server->accept_time.tv_sec);
+	ack.ts.sentrx_tv_usec = htonl(server->accept_time.tv_usec);
+	Timestamp now;
+	ack.ts.ack_tv_sec = htonl(now.getSecs());
+	ack.ts.ack_tv_usec = htonl(now.getUsecs());
+    } else {
+	size -= sizeof (struct client_hdr_ack_ts);
+    }
+    ack.typelen.length = htonl(size);
     int rc = 1;
     // This is a version 2.0.10 or greater client
     // write back to the client so it knows the server
     // version
-    if (!isUDP(server)) {
-	// sotimer units microseconds convert
-	if (server->mInterval) {
-	    sotimer = static_cast<int>((server->mInterval) / 4);
-	} else if (isModeTime(server)) {
-	    sotimer = static_cast<int>((server->mAmount * 10000) / 4);
-	}
-	if (sotimer > HDRXACKMAX) {
-	    sotimer = HDRXACKMAX;
-	} else if (sotimer < HDRXACKMIN) {
-	    sotimer = HDRXACKMIN;
-	}
-	SetSocketOptionsSendTimeout(server, sotimer);
-#ifdef TCP_NODELAY
-	optflag=1;
-	// Disable Nagle to reduce latency of this intial message
-	if ((rc = setsockopt(server->mSock, IPPROTO_TCP, TCP_NODELAY, reinterpret_cast<char *>(&optflag), sizeof(int))) < 0) {
-	    WARN_errno(rc < 0, "tcpnodelay");
-	}
-#endif
+
+    // sotimer units microseconds convert
+    if (server->mInterval) {
+	sotimer = static_cast<int>((server->mInterval) / 4);
+    } else if (isModeTime(server)) {
+	sotimer = static_cast<int>((server->mAmount * 10000) / 4);
     }
-    if ((rc = send(server->mSock, reinterpret_cast<const char*>(&ack), sizeof(client_hdr_ack),0)) < 0) {
+    if (sotimer > HDRXACKMAX) {
+	sotimer = HDRXACKMAX;
+    } else if (sotimer < HDRXACKMIN) {
+	sotimer = HDRXACKMIN;
+    }
+    SetSocketOptionsSendTimeout(server, sotimer);
+#if HAVE_DECL_TCP_NODELAY
+    int optflag = 1;
+    // Disable Nagle to reduce latency of this intial message
+    if ((rc = setsockopt(server->mSock, IPPROTO_TCP, TCP_NODELAY, reinterpret_cast<char *>(&optflag), sizeof(int))) < 0) {
+	WARN_errno(rc < 0, "tcpnodelay");
+    }
+#endif
+    if ((rc = send(server->mSock, reinterpret_cast<const char*>(&ack), size, 0)) < 0) {
 	WARN_errno(rc <= 0, "send_ack");
 	rc = 0;
     }
+#if HAVE_DECL_TCP_NODELAY
     // Re-nable Nagle
-    optflag=0;
+    optflag = isNoDelay(server) ? 1 : 0;
     if (!isUDP(server) && (rc = setsockopt(server->mSock, IPPROTO_TCP, TCP_NODELAY, reinterpret_cast<char *>(&optflag), sizeof(int))) < 0) {
 	WARN_errno(rc < 0, "tcpnodelay");
     }
+#endif
     return rc;
 }
