@@ -53,7 +53,7 @@
 
 // These static variables are not thread safe but ok to use becase only
 // the repoter thread usses them
-#define SNBUFFERSIZE 256
+#define SNBUFFERSIZE 512
 #define SNBUFFEREXTENDSIZE 512
 static char outbuffer[SNBUFFERSIZE]; // Buffer for printing
 static char outbufferext[SNBUFFEREXTENDSIZE]; // Buffer for printing
@@ -66,6 +66,7 @@ static int HEADING_FLAG(report_bw_jitter_loss) = 0;
 static int HEADING_FLAG(report_bw_read_enhanced) = 0;
 static int HEADING_FLAG(report_bw_read_enhanced_netpwr) = 0;
 static int HEADING_FLAG(report_bw_write_enhanced) = 0;
+static int HEADING_FLAG(report_write_enhanced_drain) = 0;
 static int HEADING_FLAG(report_bw_write_enhanced_netpwr) = 0;
 static int HEADING_FLAG(report_bw_pps_enhanced) = 0;
 static int HEADING_FLAG(report_bw_pps_enhanced_isoch) = 0;
@@ -95,6 +96,7 @@ void reporter_default_heading_flags (int flag) {
     HEADING_FLAG(report_bw_read_enhanced) = flag;
     HEADING_FLAG(report_bw_read_enhanced_netpwr) = flag;
     HEADING_FLAG(report_bw_write_enhanced) = flag;
+    HEADING_FLAG(report_write_enhanced_drain) = flag;
     HEADING_FLAG(report_write_enhanced_isoch) = flag;
     HEADING_FLAG(report_bw_write_enhanced_netpwr) = flag;
     HEADING_FLAG(report_bw_pps_enhanced) = flag;
@@ -172,7 +174,7 @@ static inline void set_netpowerbuf(double meantransit, struct TransferInfo *stat
   if (meantransit == 0.0) {
       strcpy(netpower_buf, "NAN");
   } else {
-      double netpwr = (NETPOWERCONSTANT * ((double) stats->cntBytes) / (stats->ts.iEnd - stats->ts.iStart) / meantransit);
+      double netpwr = NETPOWERCONSTANT * (((double) stats->cntBytes) / (stats->ts.iEnd - stats->ts.iStart) / meantransit);
       if (netpwr <  NETPWR_LOWERBOUNDS) {
 	  strcpy(netpower_buf, "OBL");
       } else {
@@ -239,8 +241,8 @@ void tcp_output_read_enhanced_triptime (struct TransferInfo *stats) {
 	       stats->common->transferIDStr, stats->ts.iStart, stats->ts.iEnd,
 	       outbuffer, outbufferext,
 	       (meantransit * 1e3),
-	       stats->transit.minTransit*1e3,
-	       stats->transit.maxTransit*1e3,
+	       (stats->transit.cntTransit < 2) ? 0 : stats->transit.minTransit*1e3,
+	       (stats->transit.cntTransit < 2) ? 0 : stats->transit.maxTransit*1e3,
 	       (stats->transit.cntTransit < 2) ? 0 : sqrt(stats->transit.m2Transit / (stats->transit.cntTransit - 1)) / 1e3,
 	       stats->transit.cntTransit,
 	       stats->transit.cntTransit ? (long) ((double)stats->cntBytes / (double) stats->transit.cntTransit) : 0,
@@ -353,13 +355,12 @@ void tcp_output_write (struct TransferInfo *stats) {
 void tcp_output_write_enhanced (struct TransferInfo *stats) {
     HEADING_PRINT_COND(report_bw_write_enhanced);
     _print_stats_common(stats);
-#ifndef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
+#if !(HAVE_TCP_STATS)
     printf(report_bw_write_enhanced_format,
 	   stats->common->transferIDStr, stats->ts.iStart, stats->ts.iEnd,
 	   outbuffer, outbufferext,
 	   stats->sock_callstats.write.WriteCnt,
-	   stats->sock_callstats.write.WriteErr,
-	   stats->sock_callstats.write.TCPretry);
+	   stats->sock_callstats.write.WriteErr);
 #else
     set_netpowerbuf(stats->sock_callstats.write.rtt * 1e-6, stats);
     if (stats->sock_callstats.write.cwnd > 0) {
@@ -371,6 +372,7 @@ void tcp_output_write_enhanced (struct TransferInfo *stats) {
 	       stats->sock_callstats.write.TCPretry,
 	       stats->sock_callstats.write.cwnd,
 	       stats->sock_callstats.write.rtt,
+	       stats->sock_callstats.write.rttvar,
 	       netpower_buf);
     } else {
 	printf(report_bw_write_enhanced_nocwnd_format,
@@ -382,14 +384,79 @@ void tcp_output_write_enhanced (struct TransferInfo *stats) {
 	       stats->sock_callstats.write.rtt,
 	       netpower_buf);
     }
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    if (stats->latency_histogram) {
+	histogram_print(stats->latency_histogram, stats->ts.iStart, stats->ts.iEnd);
+    }
 #endif
     fflush(stdout);
 }
 
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+void tcp_output_write_enhanced_drain (struct TransferInfo *stats) {
+    HEADING_PRINT_COND(report_write_enhanced_drain);
+    _print_stats_common(stats);
+#if !(HAVE_TCP_STATS)
+    printf(report_write_enhanced_drain_format,
+	   stats->common->transferIDStr, stats->ts.iStart, stats->ts.iEnd,
+	   outbuffer, outbufferext,
+	   stats->sock_callstats.write.WriteCnt,
+	   stats->sock_callstats.write.WriteErr,
+
+	   stats->drain_mmm.current.mean * 1e-3,
+	   stats->drain_mmm.current.min * 1e-3,
+	   stats->drain_mmm.current.max * 1e-3,
+	   (stats->drain_mmm.current.cnt < 2) ? 0 : (1e-3 * sqrt(stats->drain_mmm.current.m2 / (stats->drain_mmm.current.cnt - 1))),
+	   stats->drain_mmm.current.cnt);
+#else
+    set_netpowerbuf(stats->sock_callstats.write.rtt * 1e-6, stats);
+    if (stats->sock_callstats.write.cwnd > 0) {
+	printf(report_write_enhanced_drain_format,
+	       stats->common->transferIDStr, stats->ts.iStart, stats->ts.iEnd,
+	       outbuffer, outbufferext,
+	       stats->sock_callstats.write.WriteCnt,
+	       stats->sock_callstats.write.WriteErr,
+	       stats->sock_callstats.write.TCPretry,
+	       stats->sock_callstats.write.cwnd,
+	       stats->sock_callstats.write.rtt,
+	       netpower_buf,
+	       stats->drain_mmm.current.cnt,
+	       stats->drain_mmm.current.mean * 1e-3,
+	       stats->drain_mmm.current.min * 1e-3,
+	       stats->drain_mmm.current.max * 1e-3,
+	       (stats->drain_mmm.current.cnt < 2) ? 0 : (1e-3 * sqrt(stats->drain_mmm.current.m2 / (stats->drain_mmm.current.cnt - 1))),
+	       stats->drain_mmm.current.cnt);
+    } else {
+	printf(report_write_enhanced_nocwnd_drain_format,
+	       stats->common->transferIDStr, stats->ts.iStart, stats->ts.iEnd,
+	       outbuffer, outbufferext,
+	       stats->sock_callstats.write.WriteCnt,
+	       stats->sock_callstats.write.WriteErr,
+	       stats->sock_callstats.write.TCPretry,
+	       stats->sock_callstats.write.rtt,
+	       netpower_buf,
+	       stats->drain_mmm.current.cnt,
+	       stats->drain_mmm.current.mean * 1e-3,
+	       stats->drain_mmm.current.min * 1e-3,
+	       stats->drain_mmm.current.max * 1e-3,
+	       (stats->drain_mmm.current.cnt < 2) ? 0 : (1e-3 * sqrt(stats->drain_mmm.current.m2 / (stats->drain_mmm.current.cnt - 1))),
+	       stats->drain_mmm.current.cnt);
+    }
+#endif
+    if (stats->latency_histogram) {
+	histogram_print(stats->latency_histogram, stats->ts.iStart, stats->ts.iEnd);
+    }
+    if (stats->drain_histogram) {
+	histogram_print(stats->drain_histogram, stats->ts.iStart, stats->ts.iEnd);
+    }
+    fflush(stdout);
+}
+#endif
 void tcp_output_write_enhanced_isoch (struct TransferInfo *stats) {
     HEADING_PRINT_COND(report_write_enhanced_isoch);
     _print_stats_common(stats);
-#ifndef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
+#if !(HAVE_TCP_STATS)
     printf(report_write_enhanced_isoch_format,
 	   stats->common->transferIDStr, stats->ts.iStart, stats->ts.iEnd,
 	   outbuffer, outbufferext,
@@ -420,6 +487,11 @@ void tcp_output_write_enhanced_isoch (struct TransferInfo *stats) {
 	       stats->isochstats.cntFrames, stats->isochstats.cntFramesMissed, stats->isochstats.cntSlips,
 	       netpower_buf);
     }
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    if (stats->latency_histogram) {
+	histogram_print(stats->latency_histogram, stats->ts.iStart, stats->ts.iEnd);
+    }
 #endif
     fflush(stdout);
 }
@@ -833,8 +905,11 @@ void tcp_output_sum_write_enhanced (struct TransferInfo *stats) {
 	   stats->ts.iStart, stats->ts.iEnd,
 	   outbuffer, outbufferext,
 	   stats->sock_callstats.write.WriteCnt,
-	   stats->sock_callstats.write.WriteErr,
-	   stats->sock_callstats.write.TCPretry);
+	   stats->sock_callstats.write.WriteErr
+#if HAVE_TCP_STATS
+	   ,stats->sock_callstats.write.TCPretry
+#endif
+    );
     fflush(stdout);
 }
 void tcp_output_sumcnt_write_enhanced (struct TransferInfo *stats) {
@@ -844,8 +919,11 @@ void tcp_output_sumcnt_write_enhanced (struct TransferInfo *stats) {
 	   stats->ts.iStart, stats->ts.iEnd,
 	   outbuffer, outbufferext,
 	   stats->sock_callstats.write.WriteCnt,
-	   stats->sock_callstats.write.WriteErr,
-	   stats->sock_callstats.write.TCPretry);
+	   stats->sock_callstats.write.WriteErr
+#if HAVE_TCP_STATS
+	   ,stats->sock_callstats.write.TCPretry
+#endif
+    );
     fflush(stdout);
 }
 
@@ -869,20 +947,30 @@ static void format_ips_ports_string (struct TransferInfo *stats) {
     struct sockaddr *peer = (reverse ? (struct sockaddr*)&stats->common->local : (struct sockaddr*)&stats->common->peer);
     __ips_ports_string[0] = '\0';
     if (local->sa_family == AF_INET) {
-        inet_ntop(AF_INET, &((struct sockaddr_in*)local)->sin_addr,
+      if (isHideIPs(stats->common)) {
+	inet_ntop_hide(AF_INET, &((struct sockaddr_in*)local)->sin_addr,
+		       local_addr, REPORT_ADDRLEN);
+      } else {
+	inet_ntop(AF_INET, &((struct sockaddr_in*)local)->sin_addr,
 		  local_addr, REPORT_ADDRLEN);
+      }
     }
-#ifdef HAVE_IPV6
+#if HAVE_IPV6
     else {
         inet_ntop(AF_INET6, &((struct sockaddr_in6*)local)->sin6_addr,
 		  local_addr, REPORT_ADDRLEN);
     }
 #endif
     if (peer->sa_family == AF_INET) {
+      if (isHideIPs(stats->common)) {
+        inet_ntop_hide(AF_INET, &((struct sockaddr_in*)peer)->sin_addr,
+		  remote_addr, REPORT_ADDRLEN);
+      } else {
         inet_ntop(AF_INET, &((struct sockaddr_in*)peer)->sin_addr,
 		  remote_addr, REPORT_ADDRLEN);
+      }
     }
-#ifdef HAVE_IPV6
+#if HAVE_IPV6
     else {
         inet_ntop(AF_INET6, &((struct sockaddr_in6*)peer)->sin6_addr,
 		  remote_addr, REPORT_ADDRLEN);
@@ -891,14 +979,14 @@ static void format_ips_ports_string (struct TransferInfo *stats) {
     snprintf(__ips_ports_string, REPORT_ADDRLEN*2+10, reportCSV_peer,
 	     local_addr, (local->sa_family == AF_INET ?
 			  ntohs(((struct sockaddr_in*)local)->sin_port) :
-#ifdef HAVE_IPV6
+#if HAVE_IPV6
 			  ntohs(((struct sockaddr_in6*)local)->sin6_port)),
 #else
 	     0),
 #endif
 	remote_addr, (peer->sa_family == AF_INET ?
 		      ntohs(((struct sockaddr_in*)peer)->sin_port) :
-#ifdef HAVE_IPV6
+#if HAVE_IPV6
 		      ntohs(((struct sockaddr_in6*)peer)->sin6_port)));
 #else
                           0));
@@ -975,16 +1063,21 @@ static void reporter_output_listener_settings (struct ReportSettings *report) {
     }
     if (report->common->Localhost != NULL) {
 	if (isEnhanced(report->common) && !SockAddr_isMulticast(&report->local)) {
-	    if (report->common->Ifrname)
+	    if (report->common->Ifrname) {
 		printf(bind_address_iface, report->common->Localhost, report->common->Ifrname);
-	    else {
+	    } else {
 		char *host_ip = (char *) malloc(REPORT_ADDRLEN);
 		if (host_ip != NULL) {
 		    if (((struct sockaddr*)(&report->common->local))->sa_family == AF_INET) {
+		      if (isHideIPs(report->common)) {
+			inet_ntop_hide(AF_INET, &((struct sockaddr_in*)(&report->common->local))->sin_addr,
+				       host_ip, REPORT_ADDRLEN);
+		      } else {
 			inet_ntop(AF_INET, &((struct sockaddr_in*)(&report->common->local))->sin_addr,
 				  host_ip, REPORT_ADDRLEN);
+		      }
 		    }
-#ifdef HAVE_IPV6
+#if HAVE_IPV6
 		    else {
 			inet_ntop(AF_INET6, &((struct sockaddr_in6*)(&report->common->local))->sin6_addr,
 				  host_ip, REPORT_ADDRLEN);
@@ -1007,6 +1100,9 @@ static void reporter_output_listener_settings (struct ReportSettings *report) {
 		printf(join_multicast_sg_dev, report->common->SSMMulticastStr, report->common->Localhost, report->common->Ifrname);
         }
     }
+    if (isTunDev(report->common) || isTapDev(report->common)) {
+	printf(bind_address_iface_taptun, report->common->Ifrname);
+    }
     if (isEnhanced(report->common)) {
 	byte_snprintf(outbuffer, sizeof(outbuffer), report->common->BufLen, toupper((int)report->common->Format));
 	byte_snprintf(outbufferext, sizeof(outbufferext), report->common->BufLen / 8, 'A');
@@ -1017,6 +1113,12 @@ static void reporter_output_listener_settings (struct ReportSettings *report) {
     if (isCongestionControl(report->common) && report->common->Congestion) {
 	fprintf(stdout, "TCP congestion control set to %s\n", report->common->Congestion);
     }
+    if (isOverrideTOS(report->common)) {
+	fprintf(stdout, "Reflected TOS will be set to 0x%x\n", report->common->RTOS);
+    }
+    if (report->common->TOS) {
+	fprintf(stdout, "TOS will be set to 0x%x\n", report->common->TOS);
+    }
     if (isUDP(report->common)) {
 	if (isSingleClient(report->common)) {
 	    fprintf(stdout, "WARN: Suggested to use lower case -u instead of -U (to avoid serialize & bypass of reporter thread)\n");
@@ -1029,9 +1131,9 @@ static void reporter_output_listener_settings (struct ReportSettings *report) {
     if (isMulticast(report->common)) {
 	fprintf(stdout, "Server set to single client traffic mode (per multicast receive)\n");
     }
-    if (isRxHistogram(report->common)) {
-	fprintf(stdout, "Enabled rx-histograms bin-width=%0.3f ms, bins=%d (clients must use --trip-times)\n", \
-		((1e3 * report->common->RXbinsize) / pow(10,report->common->RXunits)), report->common->RXbins);
+    if (isHistogram(report->common)) {
+	fprintf(stdout, "Enabled receive histograms bin-width=%0.3f ms, bins=%d (clients should use --trip-times)\n", \
+		((1e3 * report->common->HistBinsize) / pow(10,report->common->HistUnits)), report->common->HistBins);
     }
     if (isFrameInterval(report->common)) {
 #if HAVE_FASTSAMPLING
@@ -1052,19 +1154,25 @@ static void reporter_output_listener_settings (struct ReportSettings *report) {
     fflush(stdout);
 }
 static void reporter_output_client_settings (struct ReportSettings *report) {
+    char *hoststr = (isHideIPs(report->common) ? report->common->HideHost \
+		      : report->common->Host);
     if (!report->common->Ifrnametx) {
-	printf(isEnhanced(report->common) ? client_pid_port : client_port, report->common->Host,
+	printf(isEnhanced(report->common) ? client_pid_port : client_port, hoststr,
 	       (isUDP(report->common) ? "UDP" : "TCP"), report->common->Port, report->pid, \
 	       (!report->common->threads ? 1 : report->common->threads));
     } else {
-	printf(client_pid_port_dev, report->common->Host,
+	printf(client_pid_port_dev, hoststr,
 	       (isUDP(report->common) ? "UDP" : "TCP"), report->common->Port, report->pid, \
 	       report->common->Ifrnametx, (!report->common->threads ? 1 : report->common->threads));
     }
     if ((isEnhanced(report->common) || isNearCongest(report->common)) && !isUDP(report->common)) {
 	byte_snprintf(outbuffer, sizeof(outbuffer), report->common->BufLen, 'B');
 	outbuffer[(sizeof(outbuffer)-1)] = '\0';
-	printf("%s: %s\n", client_write_size, outbuffer);
+	if (isTcpDrain(report->common)) {
+	    printf("%s: %s (drain-enabled)\n", client_write_size, outbuffer);
+	} else {
+	    printf("%s: %s\n", client_write_size, outbuffer);
+	}
     }
     if (isIsochronous(report->common)) {
 	char meanbuf[40];
@@ -1080,6 +1188,12 @@ static void reporter_output_client_settings (struct ReportSettings *report) {
 	tmpbuf[39]='\0';
 	printf(client_burstperiod, (1.0 / report->common->FPS), tmpbuf);
     }
+    if (isBounceBack(report->common)) {
+	char tmpbuf[40];
+	byte_snprintf(tmpbuf, sizeof(tmpbuf), report->common->BurstSize, 'A');
+	tmpbuf[39]='\0';
+	printf(client_bounceback, tmpbuf);
+    }
     if (isFQPacing(report->common)) {
 	byte_snprintf(outbuffer, sizeof(outbuffer), report->common->FQPacingRate, 'a');
 	outbuffer[(sizeof(outbuffer)-1)] = '\0';
@@ -1088,6 +1202,13 @@ static void reporter_output_client_settings (struct ReportSettings *report) {
     if (isCongestionControl(report->common) && report->common->Congestion) {
 	fprintf(stdout, "TCP congestion control set to %s\n", report->common->Congestion);
     }
+    if (isEnhanced(report->common)) {
+        if (isNoDelay(report->common)) {
+	    fprintf(stdout, "TOS set to 0x%x and nodelay (Nagle off)\n", report->common->TOS);
+	} else {
+	    fprintf(stdout, "TOS set to 0x%x (Nagle on)\n", report->common->TOS);
+	}
+    }
     if (isNearCongest(report->common)) {
 	if (report->common->rtt_weight == NEARCONGEST_DEFAULT) {
 	    fprintf(stdout, "TCP near-congestion delay weight set to %2.4f (use --near-congestion=<value> to change)\n", report->common->rtt_weight);
@@ -1112,6 +1233,15 @@ static void reporter_output_client_settings (struct ReportSettings *report) {
     } else {
 	output_window_size(report);
 	printf("\n");
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	if (isWritePrefetch(report->common)) {
+	    fprintf(stdout, "Event based writes (pending queue watermark at %d bytes)\n", report->common->WritePrefetch);
+	}
+	if (isHistogram(report->common)) {
+	    fprintf(stdout, "Enabled select histograms bin-width=%0.3f ms, bins=%d\n", \
+		((1e3 * report->common->HistBinsize) / pow(10,report->common->HistUnits)), report->common->HistBins);
+	}
+#endif
     }
     fflush(stdout);
 }
@@ -1131,179 +1261,225 @@ void reporter_connect_printf_tcp_final (struct ConnectionInfo * report) {
 
 void reporter_print_connection_report (struct ConnectionInfo *report) {
     assert(report->common);
-    if (!(report->connecttime < 0)) {
-	// copy the inet_ntop into temp buffers, to avoid overwriting
-	char local_addr[REPORT_ADDRLEN];
-	char remote_addr[REPORT_ADDRLEN];
-	struct sockaddr *local = ((struct sockaddr*)&report->common->local);
-	struct sockaddr *peer = ((struct sockaddr*)&report->common->peer);
-	outbuffer[0]='\0';
-	outbufferext[0]='\0';
-	outbufferext2[0]='\0';
-	char *b = &outbuffer[0];
-	if (!isUDP(report->common) && (report->common->socket > 0) && (isPrintMSS(report->common) || isEnhanced(report->common)))  {
-	    if (isPrintMSS(report->common) && (report->MSS <= 0)) {
-		printf(report_mss_unsupported, report->MSS);
-	    } else {
-		snprintf(b, SNBUFFERSIZE-strlen(b), " (%s%d)", "MSS=", report->MSS);
-		b += strlen(b);
-	    }
-	}
-	if (isIsochronous(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (isoch)");
+    // copy the inet_ntop into temp buffers, to avoid overwriting
+    char local_addr[REPORT_ADDRLEN];
+    char remote_addr[REPORT_ADDRLEN];
+    struct sockaddr *local = ((struct sockaddr*)&report->common->local);
+    struct sockaddr *peer = ((struct sockaddr*)&report->common->peer);
+    outbuffer[0]='\0';
+    outbufferext[0]='\0';
+    outbufferext2[0]='\0';
+    char *b = &outbuffer[0];
+    if (!isUDP(report->common) && (report->common->socket > 0) && (isPrintMSS(report->common) || isEnhanced(report->common)))  {
+	if (isPrintMSS(report->common) && (report->MSS <= 0)) {
+	    printf(report_mss_unsupported, report->MSS);
+	} else if (report->MSS != -1) {
+	    snprintf(b, SNBUFFERSIZE-strlen(b), " (%s%d)", "MSS=", report->MSS);
 	    b += strlen(b);
 	}
-	if (isPeriodicBurst(report->common) && (report->common->ThreadMode != kMode_Client) && !isServerReverse(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (burst-period=%0.2f)", (1.0 / report->common->FPS));
+    }
+#if HAVE_DECL_TCP_WINDOW_CLAMP
+    if (!isUDP(report->common) && isRxClamp(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (%s%d)", "clamp=", report->common->ClampSize);
+	b += strlen(b);
+    }
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    if (!isUDP(report->common) && (report->common->socket > 0) && isWritePrefetch(report->common))  {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (%s%d)", "prefetch=", report->common->WritePrefetch);
+	b += strlen(b);
+    }
+#endif
+    if (isIsochronous(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (isoch)");
+	b += strlen(b);
+    }
+    if (isPeriodicBurst(report->common) && (report->common->ThreadMode != kMode_Client) && !isServerReverse(report->common)) {
+#if HAVE_FASTSAMPLING
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (burst-period=%0.4fs)", (1.0 / report->common->FPS));
+#else
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (burst-period=%0.2fs)", (1.0 / report->common->FPS));
+#endif
+	b += strlen(b);
+    }
+    if (isFullDuplex(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (full-duplex)");
+	b += strlen(b);
+    } else if (isServerReverse(report->common) || isReverse(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (reverse)");
+	b += strlen(b);
+	if (isFQPacing(report->common)) {
+	    snprintf(b, SNBUFFERSIZE-strlen(b), " (fq)");
 	    b += strlen(b);
 	}
+    }
+    if (isTxStartTime(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (epoch-start)");
+	b += strlen(b);
+    }
+    if (isBounceBack(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (bounce-back)");
+	b += strlen(b);
+    }
+    if (isL2LengthCheck(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (l2mode)");
+	b += strlen(b);
+    }
+    if (isUDP(report->common) && isNoUDPfin(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (no-udp-fin)");
+	b += strlen(b);
+    }
+    if (isTripTime(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (trip-times)");
+	b += strlen(b);
+    }
+    if (isEnhanced(report->common)) {
+	snprintf(b, SNBUFFERSIZE-strlen(b), " (sock=%d)", report->common->socket);;
+	b += strlen(b);
+    }
+    if (isOverrideTOS(report->common)) {
 	if (isFullDuplex(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (full-duplex)");
-	    b += strlen(b);
-	} else if (isServerReverse(report->common) || isReverse(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (reverse)");
-	    b += strlen(b);
-	    if (isFQPacing(report->common)) {
-		snprintf(b, SNBUFFERSIZE-strlen(b), " (fq)");
-		b += strlen(b);
-	    }
+	    snprintf(b, SNBUFFERSIZE-strlen(b), " (tos rx/tx=0x%x/0x%x)", report->common->TOS, report->common->RTOS);
+	} else if (isReverse(report->common)) {
+	    snprintf(b, SNBUFFERSIZE-strlen(b), " (tos tx=0x%x)", report->common->TOS);
 	}
-	if (isTxStartTime(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (epoch-start)");
+	b += strlen(b);
+    } else if (report->common->TOS) {
+	if (isFullDuplex(report->common)) {
+	    snprintf(b, SNBUFFERSIZE-strlen(b), " (tos rx/tx=0x%x/0x%x)", report->common->TOS, report->common->TOS);
+	} else if (isReverse(report->common)) {
+	    snprintf(b, SNBUFFERSIZE-strlen(b), " (tos tx=0x%x)", report->common->TOS);
+	}
+	b += strlen(b);
+    }
+    if (isEnhanced(report->common) || isPeerVerDetect(report->common)) {
+	if (report->peerversion[0] != '\0') {
+	    snprintf(b, SNBUFFERSIZE-strlen(b), "%s", report->peerversion);
 	    b += strlen(b);
 	}
-	if (isL2LengthCheck(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (l2mode)");
-	    b += strlen(b);
-	}
-	if (isUDP(report->common) && isNoUDPfin(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (no-udp-fin)");
-	    b += strlen(b);
-	}
-	if (isTripTime(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (trip-times)");
-	    b += strlen(b);
-	}
-
-	if (isEnhanced(report->common)) {
-	    snprintf(b, SNBUFFERSIZE-strlen(b), " (sock=%d)", report->common->socket);;
-	    b += strlen(b);
-	}
-	if (isEnhanced(report->common) || isPeerVerDetect(report->common)) {
-	    if (report->peerversion[0] != '\0') {
-		snprintf(b, SNBUFFERSIZE-strlen(b), "%s", report->peerversion);
-		b += strlen(b);
-	    }
-	}
-	if (isEnhanced(report->common) || isConnectOnly(report->common)) {
-	    if (report->connect_timestamp.tv_sec > 0) {
-		struct tm ts;
-		ts = *localtime(&report->connect_timestamp.tv_sec);
-		char now_timebuf[80];
-		strftime(now_timebuf, sizeof(now_timebuf), "%Y-%m-%d %H:%M:%S (%Z)", &ts);
-		if (!isUDP(report->common) && (report->common->ThreadMode == kMode_Client)) {
-		    snprintf(b, SNBUFFERSIZE-strlen(b), " (ct=%4.2f ms) on %s", report->connecttime, now_timebuf);
-		} else {
-		    snprintf(b, SNBUFFERSIZE-strlen(b), " on %s", now_timebuf);
+    }
+    if (!isServerReverse(report->common) && (isEnhanced(report->common) || isConnectOnly(report->common))) {
+	if (report->connect_timestamp.tv_sec > 0) {
+	    struct tm ts;
+	    ts = *localtime(&report->connect_timestamp.tv_sec);
+	    char now_timebuf[80];
+	    strftime(now_timebuf, sizeof(now_timebuf), "%Y-%m-%d %H:%M:%S (%Z)", &ts);
+	    if (!isUDP(report->common) && (report->common->ThreadMode == kMode_Client)) {
+#if HAVE_TCP_STATS
+		if (report->init_cond.connecttime > 0.0) {
+		    snprintf(b, SNBUFFERSIZE-strlen(b), " (irtt/icwnd=%u/%u)", report->init_cond.rtt, report->init_cond.cwnd);
+		    b += strlen(b);
 		}
-		b += strlen(b);
+#endif
+		snprintf(b, SNBUFFERSIZE-strlen(b), " (ct=%4.2f ms) on %s", report->init_cond.connecttime, now_timebuf);
+	    } else {
+		snprintf(b, SNBUFFERSIZE-strlen(b), " on %s", now_timebuf);
 	    }
+	    b += strlen(b);
 	}
-	if (local->sa_family == AF_INET) {
+    }
+    if (local->sa_family == AF_INET) {
+	if (isHideIPs(report->common)) {
+	    inet_ntop_hide(AF_INET, &((struct sockaddr_in*)local)->sin_addr, local_addr, REPORT_ADDRLEN);
+	} else {
 	    inet_ntop(AF_INET, &((struct sockaddr_in*)local)->sin_addr, local_addr, REPORT_ADDRLEN);
 	}
-#ifdef HAVE_IPV6
-	else {
-	    inet_ntop(AF_INET6, &((struct sockaddr_in6*)local)->sin6_addr, local_addr, REPORT_ADDRLEN);
-	}
+    }
+#if HAVE_IPV6
+    else {
+	inet_ntop(AF_INET6, &((struct sockaddr_in6*)local)->sin6_addr, local_addr, REPORT_ADDRLEN);
+    }
 #endif
-	if (peer->sa_family == AF_INET) {
+    if (peer->sa_family == AF_INET) {
+	if (isHideIPs(report->common)) {
+	    inet_ntop_hide(AF_INET, &((struct sockaddr_in*)peer)->sin_addr, remote_addr, REPORT_ADDRLEN);
+	} else {
 	    inet_ntop(AF_INET, &((struct sockaddr_in*)peer)->sin_addr, remote_addr, REPORT_ADDRLEN);
 	}
-#ifdef HAVE_IPV6
-	else {
-	    inet_ntop(AF_INET6, &((struct sockaddr_in6*)peer)->sin6_addr, remote_addr, REPORT_ADDRLEN);
-	}
+    }
+#if HAVE_IPV6
+    else {
+	inet_ntop(AF_INET6, &((struct sockaddr_in6*)peer)->sin6_addr, remote_addr, REPORT_ADDRLEN);
+    }
 #endif
-#ifdef HAVE_IPV6
-	if (report->common->KeyCheck) {
-	    if (isEnhanced(report->common) && report->common->Ifrname && (strlen(report->common->Ifrname) < SNBUFFERSIZE-strlen(b))) {
-		printf(report_peer_dev, report->common->transferIDStr, local_addr, report->common->Ifrname, \
-		       (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : \
-			ntohs(((struct sockaddr_in6*)local)->sin6_port)), \
-		       remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) : \
-				     ntohs(((struct sockaddr_in6*)peer)->sin6_port)), outbuffer);
-	    } else {
-		printf(report_peer, report->common->transferIDStr, local_addr, \
-		       (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : \
-			ntohs(((struct sockaddr_in6*)local)->sin6_port)), \
-		       remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) : \
-				     ntohs(((struct sockaddr_in6*)peer)->sin6_port)), outbuffer);
-	    }
+#if HAVE_IPV6
+    if (report->common->KeyCheck) {
+	if (isEnhanced(report->common) && report->common->Ifrname && (strlen(report->common->Ifrname) < SNBUFFERSIZE-strlen(b))) {
+	    printf(report_peer_dev, report->common->transferIDStr, local_addr, report->common->Ifrname, \
+		   (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : \
+		    ntohs(((struct sockaddr_in6*)local)->sin6_port)), \
+		   remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) : \
+				 ntohs(((struct sockaddr_in6*)peer)->sin6_port)), outbuffer);
 	} else {
-	    printf(report_peer_fail, local_addr, \
+	    printf(report_peer, report->common->transferIDStr, local_addr, \
 		   (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : \
 		    ntohs(((struct sockaddr_in6*)local)->sin6_port)), \
 		   remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) : \
 				 ntohs(((struct sockaddr_in6*)peer)->sin6_port)), outbuffer);
 	}
+    } else {
+	printf(report_peer_fail, local_addr, \
+	       (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : \
+		ntohs(((struct sockaddr_in6*)local)->sin6_port)), \
+	       remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) : \
+			     ntohs(((struct sockaddr_in6*)peer)->sin6_port)), outbuffer);
+    }
 
 #else
-	if (report->common->KeyCheck) {
-	    if (isEnhanced(report->common) && report->common->Ifrname  && (strlen(report->common->Ifrname) < SNBUFFERSIZE-strlen(b))) {
-		printf(report_peer_dev, report->common->transferIDStr, local_addr, report->common->Ifrname, \
-		       local_addr, (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : 0), \
-		       remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) :  0), \
-		       outbuffer);
-	    } else {
-		printf(report_peer, report->common->transferIDStr, \
-		       local_addr, (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : 0), \
-		       remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) :  0), \
-		       outbuffer);
-	    }
+    if (report->common->KeyCheck) {
+	if (isEnhanced(report->common) && report->common->Ifrname  && (strlen(report->common->Ifrname) < SNBUFFERSIZE-strlen(b))) {
+	    printf(report_peer_dev, report->common->transferIDStr, local_addr, report->common->Ifrname, \
+		   local_addr, (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : 0), \
+		   remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) :  0), \
+		   outbuffer);
 	} else {
-	    printf(report_peer_fail, local_addr, (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : 0), \
+	    printf(report_peer, report->common->transferIDStr, \
+		   local_addr, (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : 0), \
 		   remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) :  0), \
 		   outbuffer);
 	}
+    } else {
+	printf(report_peer_fail, local_addr, (local->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)local)->sin_port) : 0), \
+	       remote_addr, (peer->sa_family == AF_INET ? ntohs(((struct sockaddr_in*)peer)->sin_port) :  0), \
+	       outbuffer);
+    }
 #endif
-	if ((report->common->ThreadMode == kMode_Client) && !isServerReverse(report->common)) {
-	    if (isTxHoldback(report->common) || isTxStartTime(report->common)) {
-		struct tm ts;
-		char start_timebuf[80];
-		struct timeval now;
-		struct timeval start;
+    if ((report->common->ThreadMode == kMode_Client) && !isServerReverse(report->common)) {
+	if (isTxHoldback(report->common) || isTxStartTime(report->common)) {
+	    struct tm ts;
+	    char start_timebuf[80];
+	    struct timeval now;
+	    struct timeval start;
 #ifdef HAVE_CLOCK_GETTIME
-		struct timespec t1;
-		clock_gettime(CLOCK_REALTIME, &t1);
-		now.tv_sec  = t1.tv_sec;
-		now.tv_usec = t1.tv_nsec / 1000;
+	    struct timespec t1;
+	    clock_gettime(CLOCK_REALTIME, &t1);
+	    now.tv_sec  = t1.tv_sec;
+	    now.tv_usec = t1.tv_nsec / 1000;
 #else
-		gettimeofday(&now, NULL);
+	    gettimeofday(&now, NULL);
 #endif
-		ts = *localtime(&now.tv_sec);
-		char now_timebuf[80];
-		strftime(now_timebuf, sizeof(now_timebuf), "%Y-%m-%d %H:%M:%S (%Z)", &ts);
-		// Format time, "ddd yyyy-mm-dd hh:mm:ss zzz"
-		int seconds_from_now;
-		if (isTxHoldback(report->common)) {
-		    seconds_from_now = report->txholdbacktime.tv_sec;
-		    if (report->txholdbacktime.tv_usec > 0)
-			seconds_from_now++;
-		    start.tv_sec = now.tv_sec + seconds_from_now;
-		    ts = *localtime(&start.tv_sec);
-		} else {
-		    ts = *localtime(&report->epochStartTime.tv_sec);
-		    seconds_from_now = ceil(TimeDifference(report->epochStartTime, now));
-		}
-		strftime(start_timebuf, sizeof(start_timebuf), "%Y-%m-%d %H:%M:%S", &ts);
-		if (seconds_from_now > 0) {
-		    printf(client_report_epoch_start_current, report->common->transferID, seconds_from_now, \
-			   start_timebuf, now_timebuf);
-		} else {
-		    printf(warn_start_before_now, report->common->transferID, report->epochStartTime.tv_sec, \
-			   report->epochStartTime.tv_usec, start_timebuf, now_timebuf);
-		}
+	    ts = *localtime(&now.tv_sec);
+	    char now_timebuf[80];
+	    strftime(now_timebuf, sizeof(now_timebuf), "%Y-%m-%d %H:%M:%S (%Z)", &ts);
+	    // Format time, "ddd yyyy-mm-dd hh:mm:ss zzz"
+	    int seconds_from_now;
+	    if (isTxHoldback(report->common)) {
+		seconds_from_now = report->txholdbacktime.tv_sec;
+		if (report->txholdbacktime.tv_usec > 0)
+		    seconds_from_now++;
+		start.tv_sec = now.tv_sec + seconds_from_now;
+		ts = *localtime(&start.tv_sec);
+	    } else {
+		ts = *localtime(&report->epochStartTime.tv_sec);
+		seconds_from_now = ceil(TimeDifference(report->epochStartTime, now));
+	    }
+	    strftime(start_timebuf, sizeof(start_timebuf), "%Y-%m-%d %H:%M:%S", &ts);
+	    if (seconds_from_now > 0) {
+		printf(client_report_epoch_start_current, report->common->transferID, seconds_from_now, \
+		       start_timebuf, now_timebuf);
+	    } else {
+		printf(warn_start_before_now, report->common->transferID, report->epochStartTime.tv_sec, \
+		       report->epochStartTime.tv_usec, start_timebuf, now_timebuf);
 	    }
 	}
     }
@@ -1338,13 +1514,19 @@ void reporter_peerversion (struct ConnectionInfo *report, uint32_t upper, uint32
 	    sprintf(report->peerversion + strlen(report->peerversion) - 1,"-dev)");
 	    break;
 	case 1:
-	    sprintf(report->peerversion + strlen(report->peerversion) - 1,"-rc1)");
+	    sprintf(report->peerversion + strlen(report->peerversion) - 1,"-rc)");
 	    break;
 	case 2:
 	    sprintf(report->peerversion + strlen(report->peerversion) - 1,"-rc2)");
 	    break;
 	case 3:
 	    break;
+	case 4:
+	    sprintf(report->peerversion + strlen(report->peerversion) - 1,"-private)");
+	    break;
+	case 5:
+	    sprintf(report->peerversion + strlen(report->peerversion) - 1,"-master)");
+	    break;
 	default:
 	    sprintf(report->peerversion + strlen(report->peerversion) - 1, "-unk)");
 	}
