@@ -63,6 +63,7 @@
 #include "delay.h"
 #include "packet_ring.h"
 #include "payloads.h"
+#include "gettcpinfo.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -75,9 +76,6 @@ extern "C" {
 struct ReportHeader *ReportRoot = NULL;
 struct ReportHeader *ReportPendingHead = NULL;
 struct ReportHeader *ReportPendingTail = NULL;
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-void gettcpistats(struct ReporterData *data, int final, struct tcp_info *tcp_stats);
-#endif
 
 // Reporter's reset of stats after a print occurs
 static void reporter_reset_transfer_stats_client_tcp(struct TransferInfo *stats);
@@ -85,6 +83,11 @@ static void reporter_reset_transfer_stats_client_udp(struct TransferInfo *stats)
 static void reporter_reset_transfer_stats_server_udp(struct TransferInfo *stats);
 static void reporter_reset_transfer_stats_server_tcp(struct TransferInfo *stats);
 
+static void reporter_mmm_update (struct MeanMinMaxStats *stats, double value);
+
+#if HAVE_TCP_STATS
+static inline void reporter_handle_packet_tcpistats(struct ReporterData *data, struct ReportStruct *packet);
+#endif
 static struct ConnectionInfo *myConnectionReport;
 
 void PostReport (struct ReportHeader *reporthdr) {
@@ -125,24 +128,30 @@ void PostReport (struct ReportHeader *reporthdr) {
  * will actually represent many packets). This needs to
  * be as simple and fast as possible as it gets called for
  * every "packet".
+ *
+ * Returns true when the tcpinfo was sampled, false ohterwise
  */
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-bool ReportPacket (struct ReporterData* data, struct ReportStruct *packet, struct tcp_info *tcp_stats) {
+bool ReportPacket (struct ReporterData* data, struct ReportStruct *packet) {
     assert(data != NULL);
-    struct TransferInfo *stats = &data->info;
+
     bool rc = false;
   #ifdef HAVE_THREAD_DEBUG
     if (packet->packetID < 0) {
 	thread_debug("Reporting last packet for %p  qdepth=%d sock=%d", (void *) data, packetring_getcount(data->packetring), data->info.common->socket);
     }
   #endif
-    // tcpi stats are only sampled on the report interval or per near-congestion
-    if ((isEnhanced(stats->common) && (stats->common->ThreadMode == kMode_Client) && \
-	(TimeDifference(stats->ts.nextTime, packet->packetTime) < 0)) \
-	|| isNearCongest(stats->common)) {
-	gettcpistats(data, 0, tcp_stats);
-	rc = true;
+#if HAVE_TCP_STATS
+    struct TransferInfo *stats = &data->info;
+    if (stats->isEnableTcpInfo) {
+	if (!TimeZero(stats->ts.nextTCPStampleTime) && (TimeDifference(stats->ts.nextTCPStampleTime, packet->packetTime) < 0)) {
+	    gettcpinfo(data->info.common->socket, packet);
+	    TimeAdd(stats->ts.nextTCPStampleTime, stats->ts.intervalTime);
+	} else {
+	    gettcpinfo(data->info.common->socket, packet);
+	}
     }
+#endif
+
     // Note for threaded operation all that needs
     // to be done is to enqueue the packet data
     // into the ring.
@@ -154,7 +163,7 @@ bool ReportPacket (struct ReporterData* data, struct ReportStruct *packet, struc
     // should be avoided.
   #ifdef HAVE_THREAD
     // bypass the reporter thread here for single UDP
-    if (isSingleUDP(stats->common))
+    if (isSingleUDP(data->info.common))
         reporter_process_transfer_report(data);
   #else
     /*
@@ -162,38 +171,9 @@ bool ReportPacket (struct ReporterData* data, struct ReportStruct *packet, struc
      */
     reporter_process_transfer_report(data);
   #endif
+
     return rc;
 }
-#else
-void ReportPacket (struct ReporterData* data, struct ReportStruct *packet) {
-    assert(data != NULL);
-    struct TransferInfo *stats = &data->info;
-  #ifdef HAVE_THREAD_DEBUG
-    if (packet->packetID < 0) {
-	thread_debug("Reporting last packet for %p  qdepth=%d sock=%d", (void *) data, packetring_getcount(data->packetring), data->info.common->socket);
-    }
-  #endif
-    // Note for threaded operation all that needs
-    // to be done is to enqueue the packet data
-    // into the ring.
-    packetring_enqueue(data->packetring, packet);
-    // The traffic thread calls the reporting process
-    // directly forr non-threaded operation
-    // These defeats the puropse of separating
-    // traffic i/o from user i/o and really
-    // should be avoided.
-  #ifdef HAVE_THREAD
-    // bypass the reporter thread here for single UDP
-    if (isSingleUDP(stats->common))
-        reporter_process_transfer_report(data);
-  #else
-    /*
-     * Process the report in this thread
-     */
-    reporter_process_transfer_report(data);
-  #endif
-}
-#endif
 
 /*
  * EndJob is called by a traffic thread to inform the reporter
@@ -205,9 +185,7 @@ int EndJob (struct ReportHeader *reporthdr, struct ReportStruct *finalpacket) {
     assert(finalpacket!=NULL);
     struct ReporterData *report = (struct ReporterData *) reporthdr->this_report;
     struct ReportStruct packet;
-#if defined(HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS) | defined(HAVE_THREAD_DEBUG)
-    struct TransferInfo *stats = &report->info;
-#endif
+
     memset(&packet, 0, sizeof(struct ReportStruct));
     int do_close = 1;
     /*
@@ -217,10 +195,11 @@ int EndJob (struct ReportHeader *reporthdr, struct ReportStruct *finalpacket) {
      * by the reporter thread as and end of traffic
      * event
      */
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
+#if HAVE_TCP_STATS
     // tcpi stats are sampled on a final packet
-    if (isEnhanced(stats->common) && (stats->common->ThreadMode == kMode_Client)) {
-	gettcpistats(report, 0, NULL);
+    struct TransferInfo *stats = &report->info;
+    if (stats->isEnableTcpInfo) {
+	gettcpinfo(report->info.common->socket, finalpacket);
     }
 #endif
     // clear the reporter done predicate
@@ -233,11 +212,7 @@ int EndJob (struct ReportHeader *reporthdr, struct ReportStruct *finalpacket) {
 	packetring_enqueue(report->packetring, &packet);
 	reporter_process_transfer_report(report);
     } else {
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-	ReportPacket(report, &packet, NULL);
-#else
 	ReportPacket(report, &packet);
-#endif
 #ifdef HAVE_THREAD_DEBUG
 	thread_debug( "Traffic thread awaiting reporter to be done with %p and cond %p", (void *)report, (void *) report->packetring->awake_producer);
 #endif
@@ -256,7 +231,7 @@ int EndJob (struct ReportHeader *reporthdr, struct ReportStruct *finalpacket) {
 	    struct Condition *tmp = &report->FullDuplexReport->fullduplex_barrier.await;
 	    Condition_Destroy(tmp);
 #if HAVE_THREAD_DEBUG
-	    thread_debug("Socket fullduplex close sock=%d", stats->common->socket);
+	    thread_debug("Socket fullduplex close sock=%d", report->FullDuplexReport->info.common->socket);
 #endif
 	    FreeSumReport(report->FullDuplexReport);
 	} else {
@@ -339,12 +314,13 @@ static inline struct ReportHeader *reporter_jobq_set_root (struct thread_Setting
 	    reset_consumption_detector();
 	    reporter_default_heading_flags((inSettings->mReportMode == kReport_CSV));
         }
-	if (!ReportPendingHead) {
+	// Only hang the timed wait if more than this thread is active
+	if (!ReportPendingHead && (thread_numuserthreads() > 1)) {
 	    Condition_TimedWait(&ReportCond, 1);
 #ifdef HAVE_THREAD_DEBUG
-	    thread_debug( "Jobq *WAIT* exit  %p/%p cond=%p threads=%d", \
+	    thread_debug( "Jobq *WAIT* exit  %p/%p cond=%p threads u/t=%d/%d", \
 			  (void *) ReportRoot, (void *) ReportPendingHead, \
-			  (void *) &ReportCond, thread_numtrafficthreads());
+			  (void *) &ReportCond, thread_numuserthreads(), thread_numtrafficthreads());
 #endif
 	}
     }
@@ -363,28 +339,57 @@ static inline struct ReportHeader *reporter_jobq_set_root (struct thread_Setting
     Condition_Unlock(ReportCond);
     return root;
 }
-
-static void reporter_update_connect_time (double connect_time) {
-    assert(myConnectionReport != NULL);
-    if (connect_time > 0.0) {
-	myConnectionReport->connect_times.sum += connect_time;
-	if ((myConnectionReport->connect_times.cnt++) == 1) {
-	    myConnectionReport->connect_times.vd = connect_time;
-	    myConnectionReport->connect_times.mean = connect_time;
-	    myConnectionReport->connect_times.m2 = connect_time * connect_time;
-	} else {
-	    myConnectionReport->connect_times.vd = connect_time - myConnectionReport->connect_times.mean;
-	    myConnectionReport->connect_times.mean = myConnectionReport->connect_times.mean + (myConnectionReport->connect_times.vd / myConnectionReport->connect_times.cnt);
-	    myConnectionReport->connect_times.m2 = myConnectionReport->connect_times.m2 + (myConnectionReport->connect_times.vd * (connect_time - myConnectionReport->connect_times.mean));
-	}
-	// mean min max tests
-	if (connect_time < myConnectionReport->connect_times.min)
-	    myConnectionReport->connect_times.min = connect_time;
-	if (connect_time > myConnectionReport->connect_times.max)
-	    myConnectionReport->connect_times.max = connect_time;
+/*
+ * Welford's online algorithm
+ *
+ * # For a new value newValue, compute the new count, new mean, the new M2.
+ * # mean accumulates the mean of the entire dataset
+ * # M2 aggregates the squared distance from the mean
+ * # count aggregates the number of samples seen so far
+ * def update(existingAggregate, newValue):
+ *   (count, mean, M2) = existingAggregate
+ *   count += 1
+ *   delta = newValue - mean
+ *   mean += delta / count
+ *   delta2 = newValue - mean
+ *   M2 += delta * delta2
+ *   return (count, mean, M2)
+ *
+ * # Retrieve the mean, variance and sample variance from an aggregate
+ * def finalize(existingAggregate):
+ *   (count, mean, M2) = existingAggregate
+ *   if count < 2:
+ *       return float("nan")
+ *   else:
+ *       (mean, variance, sampleVariance) = (mean, M2 / count, M2 / (count - 1))
+ *       return (mean, variance, sampleVariance)
+ *
+ */
+static void reporter_mmm_update (struct MeanMinMaxStats *stats, double value) {
+    assert(stats != NULL);
+    stats->cnt++;
+    if (stats->cnt == 1) {
+	// Very first entry
+	stats->min = value;
+	stats->max = value;
+	stats->sum = value;
+	stats->vd = value;
+	stats->mean = value;
+	stats->m2 = 0;
+	stats->sum = value;
     } else {
-	myConnectionReport->connect_times.err++;
+	stats->sum += value;
+	stats->vd = value - stats->mean;
+	stats->mean += (stats->vd / stats->cnt);
+	stats->m2 += stats->vd * (value - stats->mean);
+//	printf("*****m2=%f, mmm = %f/%f\n", stats->m2, stats->vd, (value - stats->mean));
+	// mean min max tests
+	if (value < stats->min)
+	    stats->min = value;
+	if (value > stats->max)
+	    stats->max = value;
     }
+//    printf("*****val=%f, mmm = %d/%f/%f/%f/%f/%f/%f\n", value, stats->cnt, stats->sum, stats->vd, stats->mean, stats->m2, stats->min, stats->max);
 }
 
 /*
@@ -482,7 +487,6 @@ void reporter_spawn (struct thread_Settings *thread) {
 // The Transfer or Data report is by far the most complicated report
 int reporter_process_transfer_report (struct ReporterData *this_ireport) {
     assert(this_ireport != NULL);
-    assert(this_ireport->packet_handler != NULL);
     struct TransferInfo *sumstats = (this_ireport->GroupSumReport ? &this_ireport->GroupSumReport->info : NULL);
     struct TransferInfo *fullduplexstats = (this_ireport->FullDuplexReport ? &this_ireport->FullDuplexReport->info : NULL);
     int need_free = 0;
@@ -509,20 +513,22 @@ int reporter_process_transfer_report (struct ReporterData *this_ireport) {
 	// thrashing,
 	consumption_detector.accounted_packets--;
 	// Check against a final packet event on this packet ring
+#if HAVE_TCP_STATS
+	if (this_ireport->info.isEnableTcpInfo && packet->tcpstats.isValid) {
+	    reporter_handle_packet_tcpistats(this_ireport, packet);
+	}
+#endif
 	if (!(packet->packetID < 0)) {
 	    // Check to output any interval reports,
             // bursts need to report the packet first
-	    if (this_ireport->burst_boundary) {
-		(*this_ireport->packet_handler)(this_ireport, packet);
-		if (this_ireport->transfer_interval_handler) {
-		    advance_jobq = (*this_ireport->transfer_interval_handler)(this_ireport, packet);
-		}
-	    } else {
-		// timing based events handle the reporting first
-		if (this_ireport->transfer_interval_handler) {
-		    advance_jobq = (*this_ireport->transfer_interval_handler)(this_ireport, packet);
-		}
-		(*this_ireport->packet_handler)(this_ireport, packet);
+	    if (this_ireport->packet_handler_pre_report) {
+		(*this_ireport->packet_handler_pre_report)(this_ireport, packet);
+	    }
+	    if (this_ireport->transfer_interval_handler) {
+		advance_jobq = (*this_ireport->transfer_interval_handler)(this_ireport, packet);
+	    }
+	    if (this_ireport->packet_handler_post_report) {
+		(*this_ireport->packet_handler_post_report)(this_ireport, packet);
 	    }
 	    // Sum reports update the report header's last
 	    // packet time after the handler. This means
@@ -538,7 +544,12 @@ int reporter_process_transfer_report (struct ReporterData *this_ireport) {
 	    // A last packet event was detected
 	    // printf("last packet event detected\n"); fflush(stdout);
 	    this_ireport->reporter_thread_suspends = consumption_detector.reporter_thread_suspends;
-	    (*this_ireport->packet_handler)(this_ireport, packet);
+	    if (this_ireport->packet_handler_pre_report) {
+		(*this_ireport->packet_handler_pre_report)(this_ireport, packet);
+	    }
+	    if (this_ireport->packet_handler_post_report) {
+		(*this_ireport->packet_handler_post_report)(this_ireport, packet);
+	    }
 	    this_ireport->info.ts.packetTime = packet->packetTime;
 	    assert(this_ireport->transfer_protocol_handler != NULL);
 	    (*this_ireport->transfer_protocol_handler)(this_ireport, 1);
@@ -598,9 +609,13 @@ inline int reporter_process_report (struct ReportHeader *reporthdr) {
     {
 	struct ConnectionInfo *creport = (struct ConnectionInfo *)reporthdr->this_report;
 	assert(creport!=NULL);
-	if (!isCompat(creport->common) && (creport->common->ThreadMode == kMode_Client)) {
+	if (!isCompat(creport->common) && (creport->common->ThreadMode == kMode_Client) && myConnectionReport) {
 	    // Clients' connect times will be inputs to the overall connect stats
-	    reporter_update_connect_time(creport->connecttime);
+	    if (creport->init_cond.connecttime > 0.0) {
+		reporter_mmm_update(&myConnectionReport->connect_times, creport->init_cond.connecttime);
+	    } else {
+		myConnectionReport->connect_times.err++;
+	    }
 	}
 	reporter_print_connection_report(creport);
 	fflush(stdout);
@@ -758,7 +773,7 @@ static inline void reporter_handle_burst_tcp_client_transit (struct ReporterData
     if (stats->burstid_transition && packet->frameID && packet->transit_ready) {
 	stats->burstid_transition = false;
 	// printf("***Burst id = %ld, transit = %f\n", packet->frameID, stats->transit.lastTransit);
-    } else if (!stats->burstid_transition) {
+    } else if (isIsochronous(stats->common) && !stats->burstid_transition) {
 	stats->burstid_transition = true;
 	if (packet->frameID && (packet->frameID != (stats->isochstats.frameID + 1))) {
 	    fprintf(stderr,"%sError: expected burst id %u but got %" PRIdMAX "\n", \
@@ -878,6 +893,19 @@ inline void reporter_handle_packet_server_udp (struct ReporterData *data, struct
     }
 }
 
+// This is done in reporter thread context
+#if HAVE_TCP_STATS
+static inline void reporter_handle_packet_tcpistats (struct ReporterData *data, struct ReportStruct *packet) {
+    assert(data!=NULL);
+    struct TransferInfo *stats = &data->info;
+    stats->sock_callstats.write.TCPretry += (packet->tcpstats.retry_tot - stats->sock_callstats.write.totTCPretry);
+    stats->sock_callstats.write.totTCPretry = packet->tcpstats.retry_tot;
+    stats->sock_callstats.write.cwnd = packet->tcpstats.cwnd;
+    stats->sock_callstats.write.rtt = packet->tcpstats.rtt;
+    stats->sock_callstats.write.rttvar = packet->tcpstats.rttvar;
+}
+#endif
+
 void reporter_handle_packet_client (struct ReporterData *data, struct ReportStruct *packet) {
     struct TransferInfo *stats = &data->info;
     stats->ts.packetTime = packet->packetTime;
@@ -888,13 +916,26 @@ void reporter_handle_packet_client (struct ReporterData *data, struct ReportStru
 	    stats->sock_callstats.write.totWriteErr++;
 	}
 	// These are valid packets that need standard iperf accounting
-	stats->sock_callstats.write.WriteCnt++;
-	stats->sock_callstats.write.totWriteCnt++;
+	stats->sock_callstats.write.WriteCnt += packet->writecnt;
+	stats->sock_callstats.write.totWriteCnt += packet->writecnt;
 	if (isIsochronous(stats->common)) {
 	    reporter_handle_packet_isochronous(data, packet);
 	} else if (isPeriodicBurst(stats->common)) {
 	    reporter_handle_burst_tcp_client_transit(data, packet);
 	}
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	if (stats->latency_histogram && (packet->select_delay > 0.0)) {
+	   histogram_insert(stats->latency_histogram, packet->select_delay, &packet->packetTime);
+       }
+       if (isTcpDrain(stats->common) && packet->transit_ready && (packet->drain_time)) {
+	   reporter_mmm_update(&stats->drain_mmm.current, (double) packet->drain_time);
+	   reporter_mmm_update(&stats->drain_mmm.total, (double) packet->drain_time);
+	   if (stats->drain_histogram ) {
+	       // convert drain time from microseconds to seconds prior to insert
+	       histogram_insert(stats->drain_histogram, (1e-6 * packet->drain_time), &packet->packetTime);
+	   }
+       }
+#endif
     }
     if (isUDP(stats->common)) {
 	stats->PacketID = packet->packetID;
@@ -902,56 +943,9 @@ void reporter_handle_packet_client (struct ReporterData *data, struct ReportStru
     }
 }
 
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-void gettcpistats (struct ReporterData *data, int final, struct tcp_info *tcp_stats) {
-    assert(data!=NULL);
-    struct TransferInfo *stats = &data->info;
-    struct TransferInfo *sumstats = (data->GroupSumReport != NULL) ? &data->GroupSumReport->info : NULL;
-    static int cnt = 0;
-    struct tcp_info tcp_info_buf;
-    struct tcp_info *this_tcp_stats = (tcp_stats != NULL) ? tcp_stats : &tcp_info_buf;
-    int retry = 0;
-    // Read the TCP retry stats for a client.  Do this
-    // on  a report interval period.
-    int rc = (stats->common->socket==INVALID_SOCKET) ? 0 : 1;
-    if (rc) {
-        socklen_t tcp_info_length = sizeof(struct tcp_info);
-        rc = (getsockopt(stats->common->socket, IPPROTO_TCP, TCP_INFO, this_tcp_stats, &tcp_info_length) < 0) ? 0 : 1;
-        if (!rc) {
-	    stats->common->socket = INVALID_SOCKET;
-	} else {
-	    // Mark stale now so next call at report interval will update
-	    stats->sock_callstats.write.up_to_date = 1;
-	}
-    }
-    if (!rc) {
-        stats->sock_callstats.write.TCPretry = 0;
-	stats->sock_callstats.write.cwnd = -1;
-	stats->sock_callstats.write.rtt = 0;
-    } else {
-        retry = this_tcp_stats->tcpi_total_retrans - stats->sock_callstats.write.lastTCPretry;
-	stats->sock_callstats.write.TCPretry = retry;
-	stats->sock_callstats.write.totTCPretry += retry;
-	stats->sock_callstats.write.cwnd = this_tcp_stats->tcpi_snd_cwnd * this_tcp_stats->tcpi_snd_mss / 1024;
-	stats->sock_callstats.write.rtt = this_tcp_stats->tcpi_rtt;
-	// New average = old average * (n-1)/n + new value/n
-	cnt++;
-	stats->sock_callstats.write.meanrtt = (stats->sock_callstats.write.meanrtt * ((double) (cnt - 1) / (double) cnt)) + ((double) (this_tcp_stats->tcpi_rtt) / (double) cnt);
-	stats->sock_callstats.write.rtt = this_tcp_stats->tcpi_rtt;
-	if (sumstats) {
-	    sumstats->sock_callstats.write.TCPretry += retry;
-	    sumstats->sock_callstats.write.totTCPretry += retry;
-	}
-    }
-    if (final) {
-        stats->sock_callstats.write.rtt = stats->sock_callstats.write.meanrtt;
-    }
-}
-#endif
 /*
  * Report printing routines below
  */
-
 static inline void reporter_set_timestamps_time (struct ReportTimeStamps *times, enum TimeStampType tstype) {
     // There is a corner case when the first packet is also the last where the start time (which comes
     // from app level syscall) is greater than the packetTime (which come for kernel level SO_TIMESTAMP)
@@ -997,7 +991,7 @@ static inline void reporter_transfer_protocol_missed_reports (struct TransferInf
 	emptystats.ts.iStart = stats->ts.iStart;
 	emptystats.ts.iEnd = stats->ts.iEnd;
 	emptystats.common = stats->common;
-	if ((stats->output_handler) && !(stats->filter_this_sample_output))
+	if ((stats->output_handler) && !(stats->isMaskOutput))
 	    (*stats->output_handler)(&emptystats);
     }
 }
@@ -1009,10 +1003,19 @@ static inline void reporter_reset_transfer_stats_client_tcp (struct TransferInfo
     stats->isochstats.framecnt.prev = stats->isochstats.framecnt.current;
     stats->isochstats.framelostcnt.prev = stats->isochstats.framelostcnt.current;
     stats->isochstats.slipcnt.prev = stats->isochstats.slipcnt.current;
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-    stats->sock_callstats.write.lastTCPretry = stats->sock_callstats.write.TCPretry;
+#if HAVE_TCP_STATS
     stats->sock_callstats.write.TCPretry = 0;
-    stats->sock_callstats.write.up_to_date = 0;
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    if (isTcpDrain(stats->common)) {
+	stats->drain_mmm.current.cnt = 0;
+	stats->drain_mmm.current.min = FLT_MAX;
+	stats->drain_mmm.current.max = FLT_MIN;
+	stats->drain_mmm.current.sum = 0;
+	stats->drain_mmm.current.vd = 0;
+	stats->drain_mmm.current.mean = 0;
+	stats->drain_mmm.current.m2 = 0;
+    }
 #endif
 }
 
@@ -1094,10 +1097,18 @@ void reporter_transfer_protocol_server_udp (struct ReporterData *data, int final
 	stats->cntError = 0;
     stats->cntDatagrams = stats->PacketID - stats->total.Datagrams.prev;
     stats->cntIPG = stats->total.IPG.current - stats->total.IPG.prev;
+    if (stats->latency_histogram) {
+        stats->latency_histogram->final = final;
+    }
+
     if (isIsochronous(stats->common)) {
 	stats->isochstats.cntFrames = stats->isochstats.framecnt.current - stats->isochstats.framecnt.prev;
 	stats->isochstats.cntFramesMissed = stats->isochstats.framelostcnt.current - stats->isochstats.framelostcnt.prev;
 	stats->isochstats.cntSlips = stats->isochstats.slipcnt.current - stats->isochstats.slipcnt.prev;
+	if (stats->framelatency_histogram) {
+	    stats->framelatency_histogram->final = final;
+	}
+
     }
     if (stats->total.Datagrams.current == 1)
 	stats->jitter = 0;
@@ -1130,7 +1141,7 @@ void reporter_transfer_protocol_server_udp (struct ReporterData *data, int final
 	    if (stats->cntError < 0)
 		stats->cntError = 0;
 	    stats->cntDatagrams = stats->PacketID - stats->total.Datagrams.prev;
-	    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 		reporter_set_timestamps_time(&stats->ts, FINALPARTIAL);
 		if ((stats->ts.iEnd - stats->ts.iStart) > stats->ts.significant_partial)
 		    (*stats->output_handler)(stats);
@@ -1173,7 +1184,7 @@ void reporter_transfer_protocol_server_udp (struct ReporterData *data, int final
 	    stats->framelatency_histogram->final = 1;
 	}
     }
-    if ((stats->output_handler) && !(stats->filter_this_sample_output))
+    if ((stats->output_handler) && !(stats->isMaskOutput))
 	(*stats->output_handler)(stats);
     if (!final)
 	reporter_reset_transfer_stats_server_udp(stats);
@@ -1205,7 +1216,7 @@ void reporter_transfer_protocol_sum_server_udp (struct TransferInfo *stats, int
 	stats->cntBytes = stats->total.Bytes.current - stats->total.Bytes.prev;
 	stats->cntIPG = stats->total.IPG.current - stats->total.IPG.prev;
     }
-    if ((stats->output_handler) && !(stats->filter_this_sample_output))
+    if ((stats->output_handler) && !(stats->isMaskOutput))
 	(*stats->output_handler)(stats);
     if (!final) {
 	stats->threadcnt = 0;
@@ -1219,7 +1230,6 @@ void reporter_transfer_protocol_sum_client_udp (struct TransferInfo *stats, int
 	reporter_set_timestamps_time(&stats->ts, TOTAL);
 	stats->sock_callstats.write.WriteErr = stats->sock_callstats.write.totWriteErr;
 	stats->sock_callstats.write.WriteCnt = stats->sock_callstats.write.totWriteCnt;
-	stats->sock_callstats.write.TCPretry = stats->sock_callstats.write.totTCPretry;
 	stats->cntDatagrams = stats->total.Datagrams.current;
 	stats->cntBytes = stats->total.Bytes.current;
 	stats->IPGsum = TimeDifference(stats->ts.packetTime, stats->ts.startTime);
@@ -1229,13 +1239,13 @@ void reporter_transfer_protocol_sum_client_udp (struct TransferInfo *stats, int
 	stats->cntIPG = stats->total.IPG.current - stats->total.IPG.prev;
 	stats->cntDatagrams = stats->total.Datagrams.current - stats->total.Datagrams.prev;
     }
-    if ((stats->output_handler) && !(stats->filter_this_sample_output))
+    if ((stats->output_handler) && !(stats->isMaskOutput))
 	(*stats->output_handler)(stats);
 
     if (!final) {
 	stats->threadcnt = 0;
 	reporter_reset_transfer_stats_client_udp(stats);
-    } else if ((stats->common->ReportMode != kReport_CSV) && !(stats->filter_this_sample_output)) {
+    } else if ((stats->common->ReportMode != kReport_CSV) && !(stats->isMaskOutput)) {
 	printf(report_sumcnt_datagrams, stats->threadcnt, stats->total.Datagrams.current);
 	fflush(stdout);
     }
@@ -1292,7 +1302,7 @@ void reporter_transfer_protocol_client_udp (struct ReporterData *data, int final
 	    stats->cntIPG = 0;
 	}
     }
-    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 	(*stats->output_handler)(stats);
 	if (final && (stats->common->ReportMode != kReport_CSV)) {
 	    printf(report_datagrams, stats->common->transferID, stats->total.Datagrams.current);
@@ -1308,6 +1318,9 @@ void reporter_transfer_protocol_server_tcp (struct ReporterData *data, int final
     struct TransferInfo *fullduplexstats = (data->FullDuplexReport != NULL) ? &data->FullDuplexReport->info : NULL;
     stats->cntBytes = stats->total.Bytes.current - stats->total.Bytes.prev;
     int ix;
+    if (stats->framelatency_histogram) {
+        stats->framelatency_histogram->final = 0;
+    }
     if (sumstats) {
 	sumstats->threadcnt++;
 	sumstats->total.Bytes.current += stats->cntBytes;
@@ -1324,7 +1337,7 @@ void reporter_transfer_protocol_server_tcp (struct ReporterData *data, int final
     if (final) {
 	if ((stats->cntBytes > 0) && stats->output_handler && !TimeZero(stats->ts.intervalTime)) {
 	    // print a partial interval report if enable and this a final
-	    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 		if (isIsochronous(stats->common)) {
 		    stats->isochstats.cntFrames = stats->isochstats.framecnt.current - stats->isochstats.framecnt.prev;
 		    stats->isochstats.cntFramesMissed = stats->isochstats.framelostcnt.current - stats->isochstats.framelostcnt.prev;
@@ -1336,6 +1349,9 @@ void reporter_transfer_protocol_server_tcp (struct ReporterData *data, int final
 		reporter_reset_transfer_stats_server_tcp(stats);
 	    }
         }
+        if (stats->framelatency_histogram) {
+	    stats->framelatency_histogram->final = 1;
+	}
 	stats->final = true;
 	reporter_set_timestamps_time(&stats->ts, TOTAL);
         stats->cntBytes = stats->total.Bytes.current;
@@ -1362,7 +1378,7 @@ void reporter_transfer_protocol_server_tcp (struct ReporterData *data, int final
 	stats->isochstats.cntFramesMissed = stats->isochstats.framelostcnt.current - stats->isochstats.framelostcnt.prev;
 	stats->isochstats.cntSlips = stats->isochstats.slipcnt.current - stats->isochstats.slipcnt.prev;
     }
-    if ((stats->output_handler) && !stats->filter_this_sample_output) {
+    if ((stats->output_handler) && !stats->isMaskOutput) {
 	(*stats->output_handler)(stats);
 	if (isFrameInterval(stats->common) && stats->framelatency_histogram) {
 	    histogram_print(stats->framelatency_histogram, stats->ts.iStart, stats->ts.iEnd);
@@ -1377,6 +1393,14 @@ void reporter_transfer_protocol_client_tcp (struct ReporterData *data, int final
     struct TransferInfo *sumstats = (data->GroupSumReport != NULL) ? &data->GroupSumReport->info : NULL;
     struct TransferInfo *fullduplexstats = (data->FullDuplexReport != NULL) ? &data->FullDuplexReport->info : NULL;
     stats->cntBytes = stats->total.Bytes.current - stats->total.Bytes.prev;
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    if (stats->latency_histogram) {
+        stats->latency_histogram->final = final;
+    }
+    if (stats->drain_histogram) {
+        stats->drain_histogram->final = final;
+    }
+#endif
     if (isIsochronous(stats->common)) {
 	if (final) {
 	    stats->isochstats.cntFrames = stats->isochstats.framecnt.current;
@@ -1392,19 +1416,29 @@ void reporter_transfer_protocol_client_tcp (struct ReporterData *data, int final
 	sumstats->total.Bytes.current += stats->cntBytes;
 	sumstats->sock_callstats.write.WriteErr += stats->sock_callstats.write.WriteErr;
 	sumstats->sock_callstats.write.WriteCnt += stats->sock_callstats.write.WriteCnt;
-	sumstats->sock_callstats.write.TCPretry += stats->sock_callstats.write.TCPretry;
 	sumstats->sock_callstats.write.totWriteErr += stats->sock_callstats.write.WriteErr;
 	sumstats->sock_callstats.write.totWriteCnt += stats->sock_callstats.write.WriteCnt;
-	sumstats->sock_callstats.write.totTCPretry += stats->sock_callstats.write.TCPretry;
 	sumstats->threadcnt++;
+#if HAVE_TCP_STATS
+	sumstats->sock_callstats.write.TCPretry += stats->sock_callstats.write.TCPretry;
+	sumstats->sock_callstats.write.totTCPretry += stats->sock_callstats.write.TCPretry;
+#endif
     }
     if (fullduplexstats) {
 	fullduplexstats->total.Bytes.current += stats->cntBytes;
     }
     if (final) {
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	if (stats->latency_histogram) {
+	    stats->latency_histogram->final = 1;
+	}
+	if (stats->drain_histogram) {
+	    stats->drain_histogram->final = 1;
+	}
+#endif
 	if ((stats->cntBytes > 0) && stats->output_handler && !TimeZero(stats->ts.intervalTime)) {
 	    // print a partial interval report if enable and this a final
-	    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 		if (isIsochronous(stats->common)) {
 		    stats->isochstats.cntFrames = stats->isochstats.framecnt.current - stats->isochstats.framecnt.prev;
 		    stats->isochstats.cntFramesMissed = stats->isochstats.framelostcnt.current - stats->isochstats.framelostcnt.prev;
@@ -1423,16 +1457,23 @@ void reporter_transfer_protocol_client_tcp (struct ReporterData *data, int final
 	}
 	stats->sock_callstats.write.WriteErr = stats->sock_callstats.write.totWriteErr;
 	stats->sock_callstats.write.WriteCnt = stats->sock_callstats.write.totWriteCnt;
+#if HAVE_TCP_STATS
 	stats->sock_callstats.write.TCPretry = stats->sock_callstats.write.totTCPretry;
+#endif
+	if (stats->framelatency_histogram) {
+	    stats->framelatency_histogram->final = 1;
+	}
 	stats->cntBytes = stats->total.Bytes.current;
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	stats->drain_mmm.current = stats->drain_mmm.total;
+#endif
 	reporter_set_timestamps_time(&stats->ts, TOTAL);
     } else if (isIsochronous(stats->common)) {
 	stats->isochstats.cntFrames = stats->isochstats.framecnt.current - stats->isochstats.framecnt.prev;
 	stats->isochstats.cntFramesMissed = stats->isochstats.framelostcnt.current - stats->isochstats.framelostcnt.prev;
 	stats->isochstats.cntSlips = stats->isochstats.slipcnt.current - stats->isochstats.slipcnt.prev;
     }
-
-    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 	(*stats->output_handler)(stats);
     }
     if (!final)
@@ -1446,13 +1487,13 @@ void reporter_transfer_protocol_sum_client_tcp (struct TransferInfo *stats, int
     if (!final || (final && (stats->cntBytes > 0) && !TimeZero(stats->ts.intervalTime))) {
 	stats->cntBytes = stats->total.Bytes.current - stats->total.Bytes.prev;
 	if (final) {
-	    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 		reporter_set_timestamps_time(&stats->ts, FINALPARTIAL);
 		if ((stats->ts.iEnd - stats->ts.iStart) > stats->ts.significant_partial)
 		    (*stats->output_handler)(stats);
 		reporter_reset_transfer_stats_client_tcp(stats);
 	    }
-	} else if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	} else if ((stats->output_handler) && !(stats->isMaskOutput)) {
 	    (*stats->output_handler)(stats);
 	    stats->threadcnt = 0;
 	}
@@ -1461,10 +1502,12 @@ void reporter_transfer_protocol_sum_client_tcp (struct TransferInfo *stats, int
     if (final) {
 	stats->sock_callstats.write.WriteErr = stats->sock_callstats.write.totWriteErr;
 	stats->sock_callstats.write.WriteCnt = stats->sock_callstats.write.totWriteCnt;
+#if HAVE_TCP_STATS
 	stats->sock_callstats.write.TCPretry = stats->sock_callstats.write.totTCPretry;
+#endif
 	stats->cntBytes = stats->total.Bytes.current;
 	reporter_set_timestamps_time(&stats->ts, TOTAL);
-	if ((stats->output_handler) && !(stats->filter_this_sample_output))
+	if ((stats->output_handler) && !(stats->isMaskOutput))
 	    (*stats->output_handler)(stats);
     }
 }
@@ -1473,12 +1516,12 @@ void reporter_transfer_protocol_sum_server_tcp (struct TransferInfo *stats, int
     if (!final || (final && (stats->cntBytes > 0) && !TimeZero(stats->ts.intervalTime))) {
 	stats->cntBytes = stats->total.Bytes.current - stats->total.Bytes.prev;
 	if (final) {
-	    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 		reporter_set_timestamps_time(&stats->ts, FINALPARTIAL);
 		if ((stats->ts.iEnd - stats->ts.iStart) > stats->ts.significant_partial)
 		    (*stats->output_handler)(stats);
 	    }
-	} else if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	} else if ((stats->output_handler) && !(stats->isMaskOutput)) {
 	    (*stats->output_handler)(stats);
 	    stats->threadcnt = 0;
 	}
@@ -1493,7 +1536,7 @@ void reporter_transfer_protocol_sum_server_tcp (struct TransferInfo *stats, int
 	}
 	stats->cntBytes = stats->total.Bytes.current;
 	reporter_set_timestamps_time(&stats->ts, TOTAL);
-	if ((stats->output_handler) && !(stats->filter_this_sample_output))
+	if ((stats->output_handler) && !(stats->isMaskOutput))
 	    (*stats->output_handler)(stats);
     }
 }
@@ -1501,7 +1544,7 @@ void reporter_transfer_protocol_fullduplex_tcp (struct TransferInfo *stats, int
     if (!final || (final && (stats->cntBytes > 0) && !TimeZero(stats->ts.intervalTime))) {
 	stats->cntBytes = stats->total.Bytes.current - stats->total.Bytes.prev;
 	if (final) {
-	    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 		reporter_set_timestamps_time(&stats->ts, FINALPARTIAL);
 		if ((stats->ts.iEnd - stats->ts.iStart) > stats->ts.significant_partial)
 		    (*stats->output_handler)(stats);
@@ -1515,7 +1558,7 @@ void reporter_transfer_protocol_fullduplex_tcp (struct TransferInfo *stats, int
     } else {
 	reporter_set_timestamps_time(&stats->ts, INTERVAL);
     }
-    if ((stats->output_handler) && !(stats->filter_this_sample_output))
+    if ((stats->output_handler) && !(stats->isMaskOutput))
 	(*stats->output_handler)(stats);
 }
 
@@ -1525,7 +1568,7 @@ void reporter_transfer_protocol_fullduplex_udp (struct TransferInfo *stats, int
 	stats->cntDatagrams = stats->total.Datagrams.current - stats->total.Datagrams.prev;
 	stats->cntIPG = stats->total.IPG.current - stats->total.IPG.prev;
 	if (final) {
-	    if ((stats->output_handler) && !(stats->filter_this_sample_output)) {
+	    if ((stats->output_handler) && !(stats->isMaskOutput)) {
 		reporter_set_timestamps_time(&stats->ts, FINALPARTIAL);
 		if ((stats->ts.iEnd - stats->ts.iStart) > stats->ts.significant_partial)
 		    (*stats->output_handler)(stats);
@@ -1534,8 +1577,6 @@ void reporter_transfer_protocol_fullduplex_udp (struct TransferInfo *stats, int
 	stats->total.Bytes.prev = stats->total.Bytes.current;
 	stats->total.IPG.prev = stats->total.IPG.current;
 	stats->total.Datagrams.prev = stats->total.Datagrams.current;
-	if (stats->cntDatagrams)
-	    stats->IPGsum = 0.0;
     }
     if (final) {
 	stats->cntBytes = stats->total.Bytes.current;
@@ -1547,8 +1588,10 @@ void reporter_transfer_protocol_fullduplex_udp (struct TransferInfo *stats, int
     } else {
 	reporter_set_timestamps_time(&stats->ts, INTERVAL);
     }
-    if ((stats->output_handler) && !(stats->filter_this_sample_output))
+    if ((stats->output_handler) && !(stats->isMaskOutput))
 	(*stats->output_handler)(stats);
+    if (stats->cntDatagrams)
+        stats->IPGsum = 0.0;
 }
 
 // Conditional print based on time
@@ -1579,11 +1622,11 @@ int reporter_condprint_time_interval_report (struct ReporterData *data, struct R
 	if (sumstats) {
 	    if ((++data->GroupSumReport->threads) == data->GroupSumReport->reference.count)   {
 		data->GroupSumReport->threads = 0;
-		if ((data->GroupSumReport->reference.count > 1) || \
+		if ((data->GroupSumReport->reference.count > (fullduplexstats ? 2 : 1)) || \
 		    isSumOnly(data->info.common)) {
-		    sumstats->filter_this_sample_output = 0;
+		    sumstats->isMaskOutput = false;
 		} else {
-		    sumstats->filter_this_sample_output = 1;
+		    sumstats->isMaskOutput = true;
 		}
 		reporter_set_timestamps_time(&sumstats->ts, INTERVAL);
 		assert(data->GroupSumReport->transfer_protocol_sum_handler != NULL);
@@ -1592,7 +1635,7 @@ int reporter_condprint_time_interval_report (struct ReporterData *data, struct R
 	}
         // In the (hopefully unlikely event) the reporter fell behind
         // output the missed reports to catch up
-	if ((stats->output_handler) && !(stats->filter_this_sample_output))
+	if ((stats->output_handler) && !(stats->isMaskOutput))
 	    reporter_transfer_protocol_missed_reports(stats, packet);
     }
     return advance_jobq;
@@ -1621,7 +1664,7 @@ int reporter_condprint_frame_interval_report_server_udp (struct ReporterData *da
 	if (stats->cntError < 0)
 	    stats->cntError = 0;
 	stats->cntDatagrams = stats->PacketID - stats->total.Datagrams.prev;
-	if ((stats->output_handler) && !(stats->filter_this_sample_output))
+	if ((stats->output_handler) && !(stats->isMaskOutput))
 	    (*stats->output_handler)(stats);
 	reporter_reset_transfer_stats_server_udp(stats);
 	advance_jobq = 1;
@@ -1651,7 +1694,7 @@ int reporter_condprint_burst_interval_report_server_tcp (struct ReporterData *da
 	stats->ts.packetTime = packet->packetTime;
 	reporter_set_timestamps_time(&stats->ts, FRAME);
 	stats->cntBytes = stats->total.Bytes.current - stats->total.Bytes.prev;
-	if ((stats->output_handler) && !(stats->filter_this_sample_output))
+	if ((stats->output_handler) && !(stats->isMaskOutput))
 	    (*stats->output_handler)(stats);
 	reporter_reset_transfer_stats_server_tcp(stats);
 	advance_jobq = 1;
