@@ -84,6 +84,7 @@ static void common_copy (struct ReportCommon **common, struct thread_Settings *i
     *common = (struct ReportCommon *) calloc(1, sizeof(struct ReportCommon));
     int bytecnt = 0;
     bytecnt = my_str_copy(&(*common)->Host, inSettings->mHost);
+    bytecnt += my_str_copy(&(*common)->HideHost, inSettings->mHideHost);
     bytecnt += my_str_copy(&(*common)->Localhost, inSettings->mLocalhost);
     bytecnt += my_str_copy(&(*common)->Ifrname, inSettings->mIfrname);
     bytecnt += my_str_copy(&(*common)->Ifrnametx, inSettings->mIfrnametx);
@@ -91,6 +92,7 @@ static void common_copy (struct ReportCommon **common, struct thread_Settings *i
     bytecnt += my_str_copy(&(*common)->Congestion, inSettings->mCongestion);
     bytecnt += my_str_copy(&(*common)->transferIDStr, inSettings->mTransferIDStr);
     bytecnt += my_str_copy(&(*common)->PermitKey, inSettings->mPermitKey);
+
     // copy some relevant settings
     (*common)->flags = inSettings->flags;
     (*common)->flags_extend = inSettings->flags_extend;
@@ -123,13 +125,21 @@ static void common_copy (struct ReportCommon **common, struct thread_Settings *i
     (*common)->size_peer = inSettings->size_peer;
     (*common)->local = inSettings->local;
     (*common)->size_local = inSettings->size_local;
-    (*common)->RXbins =inSettings->mRXbins;
-    (*common)->RXbinsize =inSettings->mRXbinsize;
-    (*common)->RXunits =inSettings->mRXunits;
+    (*common)->HistBins =inSettings->mHistBins;
+    (*common)->HistBinsize =inSettings->mHistBinsize;
+    (*common)->HistUnits =inSettings->mHistUnits;
     (*common)->pktIPG =inSettings->mBurstIPG;
-    (*common)->rtt_weight =inSettings->rtt_nearcongest_divider;
+    (*common)->rtt_weight = inSettings->rtt_nearcongest_weight_factor;
     (*common)->ListenerTimeout =inSettings->mListenerTimeout;
     (*common)->FPS = inSettings->mFPS;
+    (*common)->TOS = inSettings->mTOS;
+    (*common)->RTOS = inSettings->mRTOS;
+#if HAVE_DECL_TCP_WINDOW_CLAMP
+    (*common)->ClampSize = inSettings->mClampSize;
+#endif
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    (*common)->WritePrefetch = inSettings->mWritePrefetch;
+#endif
 #ifdef HAVE_THREAD_DEBUG
     thread_debug("Alloc common rpt/com/size/strsz %p/%p/%d/%d", (void *) common, (void *)(*common), sizeof(struct ReportCommon), bytecnt);
 #endif
@@ -143,6 +153,8 @@ static void free_common_copy (struct ReportCommon *common) {
     // Free deep copies
     if (common->Host)
 	free(common->Host);
+    if (common->HideHost)
+	free(common->HideHost);
     if (common->Localhost)
 	free(common->Localhost);
     if (common->Ifrname)
@@ -283,6 +295,7 @@ struct SumReport* InitSumReport(struct thread_Settings *inSettings, int inID, in
     common_copy(&sumreport->info.common, inSettings);
     sumreport->info.groupID = inID;
     sumreport->info.threadcnt = 0;
+    sumreport->info.isMaskOutput = false;
     // Only initialize the interval time here
     // The startTime and nextTime for summing reports will be set by
     // the reporter thread in realtime
@@ -480,9 +493,9 @@ struct ReportHeader* InitIndividualReport (struct thread_Settings *inSettings) {
     }
     // Copy common settings into the transfer report section
     common_copy(&ireport->info.common, inSettings);
-    ireport->burst_boundary = false;
     ireport->info.final = false;
     ireport->info.burstid_transition = false;
+    ireport->info.isEnableTcpInfo = false;
     // Create a new packet ring which is used to communicate
     // packet stats from the traffic thread to the reporter
     // thread.  The reporter thread does all packet accounting
@@ -510,10 +523,12 @@ struct ReportHeader* InitIndividualReport (struct thread_Settings *inSettings) {
 	ireport->transfer_interval_handler = reporter_condprint_time_interval_report;
 	ireport->info.ts.significant_partial = (double) inSettings->mInterval * PARTIALPERCENT / rMillion ;
     }
+    ireport->packet_handler_pre_report = NULL;
+    ireport->packet_handler_post_report = NULL;
     switch (inSettings->mThreadMode) {
     case kMode_Server :
 	if (isUDP(inSettings)) {
-	    ireport->packet_handler = reporter_handle_packet_server_udp;
+	    ireport->packet_handler_post_report = reporter_handle_packet_server_udp;
 	    ireport->transfer_protocol_handler = reporter_transfer_protocol_server_udp;
 	    if ((inSettings->mIntervalMode == kInterval_Frames) && isIsochronous(inSettings)) {
 		ireport->transfer_interval_handler = reporter_condprint_frame_interval_report_server_udp;
@@ -538,16 +553,18 @@ struct ReportHeader* InitIndividualReport (struct thread_Settings *inSettings) {
 		}
 	    }
 	} else {  // TCP case
-	    ireport->packet_handler = reporter_handle_packet_server_tcp;
+	    ireport->packet_handler_post_report = reporter_handle_packet_server_tcp;
 	    ireport->transfer_protocol_handler = reporter_transfer_protocol_server_tcp;
 	    if (isPeriodicBurst(inSettings)) {
 		ireport->transfer_interval_handler = reporter_condprint_burst_interval_report_server_tcp;
 		ireport->info.output_handler = tcp_output_burst_read;
-		ireport->burst_boundary = true;
+		ireport->packet_handler_pre_report = reporter_handle_packet_server_tcp;
+		ireport->packet_handler_post_report = NULL;
 	    } else if ((inSettings->mIntervalMode == kInterval_Frames) && isIsochronous(inSettings)) {
 		ireport->transfer_interval_handler = reporter_condprint_frame_interval_report_server_tcp;
 		ireport->info.output_handler = tcp_output_frame_read_triptime;
-		ireport->burst_boundary = true;
+		ireport->packet_handler_pre_report = reporter_handle_packet_server_tcp;
+		ireport->packet_handler_post_report = NULL;
 	    } else if (inSettings->mReportMode == kReport_CSV) {
 		ireport->info.output_handler = tcp_output_basic_csv;
 	    } else if (isSumOnly(inSettings)) {
@@ -564,7 +581,7 @@ struct ReportHeader* InitIndividualReport (struct thread_Settings *inSettings) {
 	}
 	break;
     case kMode_Client :
-	ireport->packet_handler = reporter_handle_packet_client;
+	ireport->packet_handler_post_report = reporter_handle_packet_client;
 	if (isUDP(inSettings)) {
 	    ireport->transfer_protocol_handler = reporter_transfer_protocol_client_udp;
             if (inSettings->mReportMode == kReport_CSV) {
@@ -586,6 +603,10 @@ struct ReportHeader* InitIndividualReport (struct thread_Settings *inSettings) {
 		ireport->info.output_handler = tcp_output_basic_csv;
 	    } else if (isSumOnly(inSettings)) {
 		ireport->info.output_handler = NULL;
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	    } else if (isTcpDrain(inSettings)) {
+		ireport->info.output_handler = tcp_output_write_enhanced_drain;
+#endif
 	    } else if (isIsochronous(inSettings)) {
 		ireport->info.output_handler = tcp_output_write_enhanced_isoch;
 	    } else if (isEnhanced(inSettings)) {
@@ -607,20 +628,40 @@ struct ReportHeader* InitIndividualReport (struct thread_Settings *inSettings) {
 
     if (inSettings->mThreadMode == kMode_Server) {
 	ireport->info.sock_callstats.read.binsize = inSettings->mBufLen / 8;
-	if (isRxHistogram(inSettings) && isUDP(inSettings) && isTripTime(inSettings)) {
+	if (isHistogram(inSettings) && isUDP(inSettings) && isTripTime(inSettings)) {
 	    char name[] = "T8";
-	    ireport->info.latency_histogram =  histogram_init(inSettings->mRXbins,inSettings->mRXbinsize,0,\
-							      pow(10,inSettings->mRXunits), \
-							      inSettings->mRXci_lower, inSettings->mRXci_upper, ireport->info.common->transferID, name);
+	    ireport->info.latency_histogram =  histogram_init(inSettings->mHistBins,inSettings->mHistBinsize,0,\
+							      pow(10,inSettings->mHistUnits), \
+							      inSettings->mHistci_lower, inSettings->mHistci_upper, ireport->info.common->transferID, name);
 	}
-	if (isRxHistogram(inSettings) && (isIsochronous(inSettings) || (!isUDP(inSettings) && isTripTime(inSettings)))) {
+	if (isHistogram(inSettings) && (isIsochronous(inSettings) || (!isUDP(inSettings) && isTripTime(inSettings)))) {
 	    char name[] = "F8";
 	    // make sure frame bin size min is 100 microsecond
-	    ireport->info.framelatency_histogram =  histogram_init(inSettings->mRXbins,inSettings->mRXbinsize,0, \
-								   pow(10,inSettings->mRXunits), inSettings->mRXci_lower, \
-								   inSettings->mRXci_upper, ireport->info.common->transferID, name);
+	    ireport->info.framelatency_histogram =  histogram_init(inSettings->mHistBins,inSettings->mHistBinsize,0, \
+								   pow(10,inSettings->mHistUnits), inSettings->mHistci_lower, \
+								   inSettings->mHistci_upper, ireport->info.common->transferID, name);
 	}
     }
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+    if ((inSettings->mThreadMode == kMode_Client) && isWritePrefetch(inSettings) && !isUDP(inSettings) && isHistogram(inSettings)) {
+	char name[] = "S8";
+	ireport->info.latency_histogram =  histogram_init(inSettings->mHistBins,inSettings->mHistBinsize,0,\
+							  pow(10,inSettings->mHistUnits), \
+							  inSettings->mHistci_lower, inSettings->mHistci_upper, ireport->info.common->transferID, name);
+    } else if ((inSettings->mThreadMode == kMode_Client) && isTcpDrain(inSettings) && isHistogram(inSettings) && !isUDP(inSettings)) {
+	char name[] = "D8";
+	inSettings->mHistBins = 100000; // 10 seconds wide
+	inSettings->mHistBinsize = 100; // 100 usec bins
+	inSettings->mHistUnits = 6;  // usecs 10 pow(x)
+	inSettings->mHistci_lower = 5;
+	inSettings->mHistci_upper = 95;
+#if HAVE_DECL_TCP_NOTSENT_LOWAT
+	ireport->info.drain_histogram =  histogram_init(inSettings->mHistBins,inSettings->mHistBinsize,0,	\
+							pow(10,inSettings->mHistUnits), \
+							inSettings->mHistci_lower, inSettings->mHistci_upper, ireport->info.common->transferID, name);
+#endif
+    }
+#endif
     return reporthdr;
 }
 
@@ -644,7 +685,7 @@ struct ReportHeader* InitIndividualReport (struct thread_Settings *inSettings) {
  * to achieve this.  Such code will be easier to maintain
  * and to extend.
  */
-struct ReportHeader* InitConnectionReport (struct thread_Settings *inSettings, double ct) {
+struct ReportHeader* InitConnectionReport (struct thread_Settings *inSettings, struct tcp_init_conditions *init_cond) {
     assert(inSettings != NULL);
     struct ReportHeader *reporthdr = (struct ReportHeader *) calloc(1, sizeof(struct ReportHeader));
     if (reporthdr == NULL) {
@@ -660,14 +701,23 @@ struct ReportHeader* InitConnectionReport (struct thread_Settings *inSettings, d
     struct ConnectionInfo * creport = (struct ConnectionInfo *)(reporthdr->this_report);
     common_copy(&creport->common, inSettings);
     if (!isUDP(inSettings) && (inSettings->mSock > 0) && !isDontRoute(inSettings) && \
-	!(ct <= 0.0 && (inSettings->mThreadMode == kMode_Client))) {
+	(inSettings->mThreadMode == kMode_Client) && \
+	(init_cond && (init_cond->connecttime > 0.0))) {
 	creport->MSS = getsock_tcp_mss(inSettings->mSock);
     } else {
 	creport->MSS = -1;
     }
     // Fill out known fields for the connection report
     reporter_peerversion(creport, inSettings->peer_version_u, inSettings->peer_version_l);
-    creport->connecttime = ct;
+    if (init_cond) {
+        creport->init_cond.connecttime = init_cond->connecttime;
+        creport->init_cond.rtt = init_cond->rtt;
+        creport->init_cond.cwnd = init_cond->cwnd;
+    } else {
+        creport->init_cond.connecttime = -1;
+        creport->init_cond.rtt = -1;
+        creport->init_cond.cwnd = -1;
+    }
     if (isEnhanced(inSettings) && isTxStartTime(inSettings)) {
 	creport->epochStartTime.tv_sec = inSettings->txstart_epoch.tv_sec;
 	creport->epochStartTime.tv_usec = inSettings->txstart_epoch.tv_usec;
@@ -684,6 +734,11 @@ struct ReportHeader* InitConnectionReport (struct thread_Settings *inSettings, d
     if (inSettings->mSock > 0) {
 	creport->winsize = getsock_tcp_windowsize(inSettings->mSock,	\
                   (inSettings->mThreadMode != kMode_Client ? 0 : 1) );
+#if HAVE_DECL_TCP_WINDOW_CLAMP
+	if (isRxClamp(inSettings)) {
+	    getsock_tcp_windowclamp(inSettings->mSock);
+	}
+#endif
     }
     creport->common->winsize_requested = inSettings->mTCPWin;
     creport->txholdbacktime = inSettings->txholdback_timer;
@@ -822,12 +877,17 @@ struct ReportHeader* InitServerRelayUDPReport(struct thread_Settings *inSettings
  * termination datagrams, so re-transmit our AckFIN.
  * Sent by server to client
  * ------------------------------------------------------------------- */
-void write_UDP_AckFIN (struct TransferInfo *stats) {
+void write_UDP_AckFIN (struct TransferInfo *stats, int len) {
     assert(stats!= NULL);
     int ackpacket_length = (int) (sizeof(struct UDP_datagram) + sizeof(struct server_hdr));
-    char *ackPacket = (char *) calloc(1, ackpacket_length);
+    char *ackPacket = (char *) calloc(1, len);
     int success = 0;
     assert(ackPacket);
+    fd_set readSet;
+    int rc = 1;
+    struct timeval timeout;
+    int readlen = ((ackpacket_length * 2) > len * 2) ? (ackpacket_length * 2) : (len * 2);
+
     if (ackPacket) {
 	struct UDP_datagram *UDP_Hdr = (struct UDP_datagram *)ackPacket;
 	struct server_hdr *hdr = (struct server_hdr *)(UDP_Hdr+1);
@@ -878,8 +938,6 @@ void write_UDP_AckFIN (struct TransferInfo *stats) {
 #define TRYCOUNT 10
 	int count = TRYCOUNT;
 	while (--count) {
-	    int rc;
-	    struct timeval timeout;
 	    // write data
 #if defined(HAVE_LINUX_FILTER_H) && defined(HAVE_AF_PACKET)
 	    // If in l2mode, use the AF_INET socket to write this packet
@@ -893,11 +951,11 @@ void write_UDP_AckFIN (struct TransferInfo *stats) {
 #endif
 	    WARN_errno(rc < 0, "write-ackfin");
 	    // wait here is for silence, no more packets from the client
-	    fd_set readSet;
+
 	    FD_ZERO(&readSet);
 	    FD_SET(stats->common->socket, &readSet);
-	    timeout.tv_sec  = 1;
-	    timeout.tv_usec = 0;
+	    timeout.tv_sec  = 0;
+	    timeout.tv_usec = 250000;
 	    rc = select(stats->common->socket+1, &readSet, NULL, NULL, &timeout);
 	    if (rc == 0) {
 #ifdef HAVE_THREAD_DEBUG
@@ -906,15 +964,16 @@ void write_UDP_AckFIN (struct TransferInfo *stats) {
 		success = 1;
 		break;
 	    }
-	    rc = read(stats->common->socket, ackPacket, ackpacket_length);
-	    WARN_errno(rc < 0, "read");
-	    if (rc < 0)
+	    rc = read(stats->common->socket, ackPacket, readlen);
+	    // WARN_errno(rc < 0, "ack await silence");
+	    if ((rc < 0) && FATALUDPREADERR(errno)) {
 		break;
-	    if (rc > 0) {
-#ifdef HAVE_THREAD_DEBUG
-		thread_debug("UDP server thinks server stats packet lost, will retransmit and try again", rc);
-#endif
 	    }
+#ifdef HAVE_THREAD_DEBUG
+	    if (rc > 0) {
+		thread_debug("UDP server thinks server stats packet maybe lost, will retransmit and try again", rc);
+	    }
+#endif
 	}
 	free(ackPacket);
     }
