@@ -80,7 +80,6 @@ Server::Server (thread_Settings *inSettings) {
     thread_debug("Server constructor with thread=%p sum=%p (sock=%d)", (void *) inSettings, (void *)inSettings->mSumReport, inSettings->mSock);
 #endif
     mSettings = inSettings;
-    mBuf = NULL;
     myJob = NULL;
     reportstruct = &scratchpad;
     memset(&scratchpad, 0, sizeof(struct ReportStruct));
@@ -95,13 +94,6 @@ Server::Server (thread_Settings *inSettings) {
 	}
     }
 #endif
-    mBufLen = (mSettings->mBufLen > MINMBUFALLOCSIZE) ? mSettings->mBufLen : MINMBUFALLOCSIZE;
-    mBuf = new char[mBufLen];
-    FAIL_errno(mBuf == NULL, "No memory for buffer\n", mSettings);
-    if (mSettings->mBufLen < static_cast<int>(sizeof(UDP_datagram))) {
-	fprintf(stderr, warn_buffer_too_small, mSettings->mBufLen);
-    }
-
     // Enable kernel level timestamping if available
     InitKernelTimeStamping();
     int sorcvtimer = 0;
@@ -110,7 +102,7 @@ Server::Server (thread_Settings *inSettings) {
     // mAmount integer, units 10 milliseconds
     // divide by two so timeout is 1/2 the interval
     if (mSettings->mInterval && (mSettings->mIntervalMode == kInterval_Time)) {
-	sorcvtimer = static_cast<int>(round(1000000.0 * mSettings->mInterval / 2.0));
+	sorcvtimer = static_cast<int>(round(mSettings->mInterval / 2.0));
     } else if (isServerModeTime(mSettings)) {
 	sorcvtimer = static_cast<int>(round(mSettings->mAmount * 10000) / 2);
     }
@@ -137,7 +129,6 @@ Server::~Server () {
 	myDropSocket = INVALID_SOCKET;
     }
 #endif
-    DELETE_ARRAY(mBuf);
 }
 
 inline bool Server::InProgress () {
@@ -166,7 +157,6 @@ void Server::RunTCP () {
 
     burst_info.send_tt.write_tv_sec = 0;
     burst_info.send_tt.write_tv_usec = 0;
-
     now.setnow();
     reportstruct->packetTime.tv_sec = now.getSecs();
     reportstruct->packetTime.tv_usec = now.getUsecs();
@@ -217,16 +207,19 @@ void Server::RunTCP () {
 		    WARN(burst_nleft <= 0, "invalid burst read req size");
 		    // thread_debug("***read burst header size %d id=%d", burst_info.burst_size, burst_info.burst_id);
 		} else {
-		    if (n > 0)
-			WARN(1, "partial readn");
+		    if (n > 0) {
+		        WARN(1, "partial readn");
 #ifdef HAVE_THREAD_DEBUG
-		    thread_debug("TCP burst partial read of %d wanted %d", n, sizeof(struct TCP_burst_payload));
+		        thread_debug("TCP burst partial read of %d wanted %d", n, sizeof(struct TCP_burst_payload));
+		    } else {
+		        thread_debug("Detected peer close");
 #endif
+		    }
 		    goto Done;
 		}
 	    }
 	    if (!reportstruct->transit_ready) {
-		n = recv(mSettings->mSock, mBuf, readLen, 0);
+		n = recv(mSettings->mSock, mSettings->mBuf, readLen, 0);
 		if (n > 0) {
 		    reportstruct->emptyreport = 0;
 		    if (isburst) {
@@ -241,8 +234,9 @@ void Server::RunTCP () {
 #ifdef HAVE_THREAD_DEBUG
 		    thread_debug("Server thread detected EOF on socket %d", mSettings->mSock);
 #endif
-		} else if ((n < 0) && (!NONFATALTCPREADERR(errno))) {
-		    // WARN_errno(1, "recv");
+		} else if ((n < 0) && (FATALTCPREADERR(errno))) {
+		    WARN_errno(1, "recv");
+		    peerclose = true;
 		    n = 0;
 		}
 		currLen += n;
@@ -255,11 +249,7 @@ void Server::RunTCP () {
 		tokens -= currLen;
 
 	    reportstruct->packetLen = currLen;
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-	    ReportPacket(myReport, reportstruct, NULL);
-#else
 	    ReportPacket(myReport, reportstruct);
-#endif
 	    // Check for reverse and amount where
 	    // the server stops after receiving
 	    // the expected byte count
@@ -289,9 +279,12 @@ void Server::RunTCP () {
     FreeReport(myJob);
 }
 
+void Server::RunTcpBounceBack () {
+}
+
 void Server::InitKernelTimeStamping () {
 #if HAVE_DECL_SO_TIMESTAMP
-    iov[0].iov_base=mBuf;
+    iov[0].iov_base=mSettings->mBuf;
     iov[0].iov_len=mSettings->mBufLen;
 
     message.msg_iov=iov;
@@ -327,9 +320,14 @@ inline void Server::SetFullDuplexReportStartTime () {
     thread_debug("Server fullduplex report start=%ld.%ld next=%ld.%ld", fullduplexstats->ts.startTime.tv_sec, fullduplexstats->ts.startTime.tv_usec, fullduplexstats->ts.nextTime.tv_sec, fullduplexstats->ts.nextTime.tv_usec);
 #endif
 }
+
 inline void Server::SetReportStartTime () {
     if (TimeZero(myReport->info.ts.startTime)) {
-	if (!TimeZero(mSettings->accept_time) && !isTxStartTime(mSettings)) {
+	if (!TimeZero(mSettings->sent_time) && !isTxStartTime(mSettings)) {
+	    // Servers that aren't full duplex use the accept timestamp for start
+	    myReport->info.ts.startTime.tv_sec = mSettings->sent_time.tv_sec;
+	    myReport->info.ts.startTime.tv_usec = mSettings->sent_time.tv_usec;
+	} else if (!TimeZero(mSettings->accept_time) && !isTxStartTime(mSettings)) {
 	    // Servers that aren't full duplex use the accept timestamp for start
 	    myReport->info.ts.startTime.tv_sec = mSettings->accept_time.tv_sec;
 	    myReport->info.ts.startTime.tv_usec = mSettings->accept_time.tv_usec;
@@ -362,7 +360,78 @@ inline void Server::SetReportStartTime () {
 #endif
 }
 
-bool Server::InitTrafficLoop () {
+void Server::ClientReverseFirstRead (void) {
+    // Handle the case when the client spawns a server (no listener) and need the initial header
+    // Case of --trip-times and --reverse or --fullduplex, listener handles normal case
+    // Handle the case when the client spawns a server (no listener) and need the initial header
+    // Case of --trip-times and --reverse or --fullduplex, listener handles normal case
+    if (isReverse(mSettings) && (isTripTime(mSettings) || isPeriodicBurst(mSettings) || isIsochronous(mSettings))) {
+        int nread = 0;
+	uint32_t flags = 0;
+	int readlen = 0;
+	if (isUDP(mSettings)) {
+	    nread = recvn(mSettings->mSock, mSettings->mBuf, mSettings->mBufLen, 0);
+	    switch (nread) {
+	    case 0:
+		//peer closed the socket, with no writes e.g. a connect-only test
+		peerclose = true;
+		break;
+	    case -1 :
+		FAIL_errno(1, "recvn-reverse", mSettings);
+		break;
+	    default :
+		struct client_udp_testhdr *udp_pkt = reinterpret_cast<struct client_udp_testhdr *>(mSettings->mBuf);
+		flags = ntohl(udp_pkt->base.flags);
+		if (isTripTime(mSettings)) {
+		    mSettings->sent_time.tv_sec = ntohl(udp_pkt->start_fq.start_tv_sec);
+		    mSettings->sent_time.tv_usec = ntohl(udp_pkt->start_fq.start_tv_usec);
+		} else {
+		    now.setnow();
+		    mSettings->sent_time.tv_sec = now.getSecs();
+		    mSettings->sent_time.tv_usec = now.getUsecs();
+		}
+		reportstruct->packetLen = nread;
+		reportstruct->packetID = 1;
+		break;
+	    }
+	} else {
+	    nread = recvn(mSettings->mSock, mSettings->mBuf, sizeof(uint32_t), 0);
+	    if (nread == 0) {
+		fprintf(stderr, "WARN: zero read on header flags\n");
+		//peer closed the socket, with no writes e.g. a connect-only test
+		peerclose = true;
+	    }
+	    FAIL_errno((nread < (int) sizeof(uint32_t)), "client read tcp flags", mSettings);
+	    reportstruct->packetID = 1;
+	    struct client_tcp_testhdr *tcp_pkt = reinterpret_cast<struct client_tcp_testhdr *>(mSettings->mBuf);
+	    flags = ntohl(tcp_pkt->base.flags);
+	    // figure out the length of the test header
+	    if ((readlen = Settings_ClientTestHdrLen(flags, mSettings)) > 0) {
+		// read the test settings passed to the mSettings by the client
+	        int adj = (readlen - sizeof(uint32_t));
+	        nread = recvn(mSettings->mSock, (mSettings->mBuf + sizeof(uint32_t)), adj, 0);
+		if (nread == 0) {
+		    peerclose = true;
+		}
+		FAIL_errno((nread < adj), "client read tcp test info", mSettings);
+		if (nread > 0) {
+		    if (isTripTime(mSettings)) {
+			struct client_tcp_testhdr *tcp_pkt = reinterpret_cast<struct client_tcp_testhdr *>(mSettings->mBuf);
+			mSettings->sent_time.tv_sec = ntohl(tcp_pkt->start_fq.start_tv_sec);
+			mSettings->sent_time.tv_usec = ntohl(tcp_pkt->start_fq.start_tv_usec);
+		    } else {
+			now.setnow();
+			mSettings->sent_time.tv_sec = now.getSecs();
+			mSettings->sent_time.tv_sec = now.getUsecs();
+		    }
+		}
+		mSettings->firstreadbytes = readlen;
+	    }
+	}
+    }
+}
+
+bool Server::InitTrafficLoop (void) {
     myJob = InitIndividualReport(mSettings);
     myReport = static_cast<struct ReporterData *>(myJob->this_report);
     assert(myJob != NULL);
@@ -384,64 +453,23 @@ bool Server::InitTrafficLoop () {
 	if ((setfullduplexflag = fullduplex_start_barrier(&mSettings->mFullDuplexReport->fullduplex_barrier)) < 0)
 	    exit(-1);
     }
-
-    // Handle the case when the client spawns a server (no listener) and need the initial header
-    // Case of --trip-times and --reverse or --fullduplex, listener handles normal case
-    if (isReverse(mSettings) && (isTripTime(mSettings) || isPeriodicBurst(mSettings))) {
-	int n = 0;
-	uint32_t flags = 0;
-	int peeklen = 0;
-	if (isUDP(mSettings)) {
-	    n = recvn(mSettings->mSock, mBuf, mBufLen, MSG_PEEK);
-	    if (n == 0) {
-		//peer closed the socket, with no writes e.g. a connect-only test
-		return false;
-	    }
-	    FAIL_errno(n < 0, "recvn-reverse", mSettings);
-	    struct client_udp_testhdr *udp_pkt = reinterpret_cast<struct client_udp_testhdr *>(mBuf);
-	    flags = ntohl(udp_pkt->base.flags);
-	    mSettings->accept_time.tv_sec = ntohl(udp_pkt->start_fq.start_tv_sec);
-	    mSettings->accept_time.tv_usec = ntohl(udp_pkt->start_fq.start_tv_usec);
-	    reportstruct->packetLen = n;
-	} else {
-	    n = recvn(mSettings->mSock, mBuf, sizeof(uint32_t), MSG_PEEK);
-	    if (n == 0) {
-		fprintf(stderr, "WARN: zero read on header flags\n");
-		//peer closed the socket, with no writes e.g. a connect-only test
-		return false;
-	    }
-	    FAIL_errno((n < (int) sizeof(uint32_t)), "read tcp flags", mSettings);
-	    struct client_tcp_testhdr *tcp_pkt = reinterpret_cast<struct client_tcp_testhdr *>(mBuf);
-	    flags = ntohl(tcp_pkt->base.flags);
-	    // figure out the length of the test header
-	    if ((peeklen = Settings_ClientHdrPeekLen(flags)) > 0) {
-		// read the test settings passed to the mSettings by the client
-		n = recvn(mSettings->mSock, mBuf, peeklen, 0);
-		FAIL_errno((n < peeklen), "read tcp test info", mSettings);
-		struct client_tcp_testhdr *tcp_pkt = reinterpret_cast<struct client_tcp_testhdr *>(mBuf);
-		mSettings->accept_time.tv_sec = ntohl(tcp_pkt->start_fq.start_tv_sec);
-		mSettings->accept_time.tv_usec = ntohl(tcp_pkt->start_fq.start_tv_usec);
-		reportstruct->packetLen = n;
-		mSettings->skip	= n;
-		if (n == 0)
-		    return false;
-		reportstruct->packetID = (n > 0) ? 1 : 0;
-	    }
-	}
+    Timestamp now;
+    if (isReverse(mSettings)) {
+	mSettings->accept_time.tv_sec = now.getSecs();
+	mSettings->accept_time.tv_usec = now.getUsecs();
+	ClientReverseFirstRead();
     }
     if (isTripTime(mSettings)) {
-	Timestamp now;
-	if ((abs(now.getSecs() - mSettings->accept_time.tv_sec)) > MAXDIFFTIMESTAMPSECS) {
+	if ((abs(now.getSecs() - mSettings->sent_time.tv_sec)) > MAXDIFFTIMESTAMPSECS) {
 	    unsetTripTime(mSettings);
 	    fprintf(stdout,"WARN: ignore --trip-times because client didn't provide valid start timestamp within %d seconds of now\n", MAXDIFFTIMESTAMPSECS);
+	    mSettings->accept_time.tv_sec = now.getSecs();
+	    mSettings->accept_time.tv_usec = now.getUsecs();
 	}
     }
-    // skip the test exchange header to get to the first burst
-    // The test exchange header was read in listener context
-    if (mSettings->skip && (isTripTime(mSettings) || isPeriodicBurst(mSettings))) {
-	reportstruct->packetLen = recvn(mSettings->mSock, mBuf, mSettings->skip, 0);
-    }
     SetReportStartTime();
+    reportstruct->prevPacketTime = myReport->info.ts.startTime;
+
     if (setfullduplexflag)
 	SetFullDuplexReportStartTime();
 
@@ -465,17 +493,18 @@ bool Server::InitTrafficLoop () {
 	PostReport(myJob);
     // The first payload is different for TCP so read it and report it
     // before entering the main loop
-    if (reportstruct->packetLen > 0) {
-	// printf("**** burst size = %d id = %d\n", burst_info.burst_size, burst_info.burst_id);
+    if (mSettings->firstreadbytes > 0) {
 	reportstruct->frameID = 0;
-	reportstruct->sentTime.tv_sec = myReport->info.ts.startTime.tv_sec;
-	reportstruct->sentTime.tv_usec = myReport->info.ts.startTime.tv_usec;
-	reportstruct->packetTime = reportstruct->sentTime;
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-	ReportPacket(myReport, reportstruct, NULL);
-#else
+	reportstruct->packetLen = mSettings->firstreadbytes;
+	if (isUDP(mSettings)) {
+	    ReadPacketID();
+	    reportstruct->packetTime = mSettings->accept_time;
+	} else {
+	    reportstruct->sentTime.tv_sec = myReport->info.ts.startTime.tv_sec;
+	    reportstruct->sentTime.tv_usec = myReport->info.ts.startTime.tv_usec;
+	    reportstruct->packetTime = reportstruct->sentTime;
+	}
 	ReportPacket(myReport, reportstruct);
-#endif
     }
     return true;
 }
@@ -496,23 +525,18 @@ inline int Server::ReadWithRxTimestamp () {
 	}
     }
 #else
-    currLen = recv(mSettings->mSock, mBuf, mSettings->mBufLen, mSettings->recvflags);
+    currLen = recv(mSettings->mSock, mSettings->mBuf, mSettings->mBufLen, mSettings->recvflags);
 #endif
     if (currLen <=0) {
 	// Socket read timeout or read error
 	reportstruct->emptyreport=1;
 	if (currLen == 0) {
 	    peerclose = true;
-	} else if
-#ifdef WIN32
-		(WSAGetLastError() != WSAEWOULDBLOCK)
-#else
-	    ((errno != EAGAIN) && (errno != EWOULDBLOCK) && (errno != ECONNREFUSED))
-#endif
-    {
-	WARN_errno(currLen, "recvmsg");
-	currLen= 0;
-    }
+	} else if (FATALUDPREADERR(errno)) {
+	    WARN_errno(1, "recvmsg");
+	    currLen = 0;
+	    peerclose = true;
+	}
     } else if (TimeZero(myReport->info.ts.prevpacketTime)) {
 	myReport->info.ts.prevpacketTime = reportstruct->packetTime;
     }
@@ -527,7 +551,7 @@ inline int Server::ReadWithRxTimestamp () {
 // Returns true if the client has indicated this is the final packet
 inline bool Server::ReadPacketID () {
     bool terminate = false;
-    struct UDP_datagram* mBuf_UDP  = reinterpret_cast<struct UDP_datagram*>(mBuf + mSettings->l4payloadoffset);
+    struct UDP_datagram* mBuf_UDP  = reinterpret_cast<struct UDP_datagram*>(mSettings->mBuf + mSettings->l4payloadoffset);
 
     // terminate when datagram begins with negative index
     // the datagram ID should be correct, just negated
@@ -559,11 +583,11 @@ inline bool Server::ReadPacketID () {
 }
 
 void Server::L2_processing () {
-#if defined(HAVE_LINUX_FILTER_H) && defined(HAVE_AF_PACKET)
-    eth_hdr = reinterpret_cast<struct ether_header *>(mBuf);
-    ip_hdr = reinterpret_cast<struct iphdr *>(mBuf + sizeof(struct ether_header));
+#if (HAVE_LINUX_FILTER_H) && (HAVE_AF_PACKET)
+    eth_hdr = reinterpret_cast<struct ether_header *>(mSettings->mBuf);
+    ip_hdr = reinterpret_cast<struct iphdr *>(mSettings->mBuf + sizeof(struct ether_header));
     // L4 offest is set by the listener and depends upon IPv4 or IPv6
-    udp_hdr = reinterpret_cast<struct udphdr *>(mBuf + mSettings->l4offset);
+    udp_hdr = reinterpret_cast<struct udphdr *>(mSettings->mBuf + mSettings->l4offset);
     // Read the packet to get the UDP length
     int udplen = ntohs(udp_hdr->len);
     //
@@ -615,7 +639,7 @@ int Server::L2_quintuple_filter () {
     }
 
     // check the L2 ethertype
-    struct ether_header *l2hdr = reinterpret_cast<struct ether_header *>(mBuf);
+    struct ether_header *l2hdr = reinterpret_cast<struct ether_header *>(mSettings->mBuf);
 
     if (!isIPV6(mSettings)) {
 	if (ntohs(l2hdr->ether_type) != ETHERTYPE_IP)
@@ -626,11 +650,11 @@ int Server::L2_quintuple_filter () {
     }
     // check the ip src/dst
     const uint32_t *data;
-    udp_hdr = reinterpret_cast<struct udphdr *>(mBuf + mSettings->l4offset);
+    udp_hdr = reinterpret_cast<struct udphdr *>(mSettings->mBuf + mSettings->l4offset);
 
     // Check plain old v4 using v4 addr structs
     if (l->sa_family == AF_INET) {
-	data = reinterpret_cast<const uint32_t *>(mBuf + sizeof(struct ether_header) + IPV4SRCOFFSET);
+	data = reinterpret_cast<const uint32_t *>(mSettings->mBuf + sizeof(struct ether_header) + IPV4SRCOFFSET);
 	if ((reinterpret_cast<struct sockaddr_in *>(p))->sin_addr.s_addr != *data++)
 	    return -1;
 	if ((reinterpret_cast<struct sockaddr_in *>(l))->sin_addr.s_addr != *data)
@@ -641,12 +665,12 @@ int Server::L2_quintuple_filter () {
 	    return -1;
     } else {
 	// Using the v6 addr structures
-#  ifdef HAVE_IPV6
+#  if HAVE_IPV6
 	struct in6_addr *v6peer = SockAddr_get_in6_addr(&mSettings->peer);
 	struct in6_addr *v6local = SockAddr_get_in6_addr(&mSettings->local);
 	if (isIPV6(mSettings)) {
 	    int i;
-	    data = reinterpret_cast<const uint32_t *>(mBuf + sizeof(struct ether_header) + IPV6SRCOFFSET);
+	    data = reinterpret_cast<const uint32_t *>(mSettings->mBuf + sizeof(struct ether_header) + IPV6SRCOFFSET);
 	    // check for v6 src/dst address match
 	    for (i = 0; i < 4; i++) {
 		if (v6peer->s6_addr32[i] != *data++)
@@ -657,7 +681,7 @@ int Server::L2_quintuple_filter () {
 		    return -1;
 	    }
 	} else { // v4 addr in v6 family struct
-	    data = reinterpret_cast<const uint32_t *>(mBuf + sizeof(struct ether_header) + IPV4SRCOFFSET);
+	    data = reinterpret_cast<const uint32_t *>(mSettings->mBuf + sizeof(struct ether_header) + IPV4SRCOFFSET);
 	    if (v6peer->s6_addr32[3] != *data++)
 		return -1;
 	    if (v6peer->s6_addr32[3] != *data)
@@ -682,7 +706,7 @@ inline void Server::udp_isoch_processing (int rxlen) {
 	reportstruct->remaining = 0;
 	reportstruct->frameID = 0;
     } else {
-	struct client_udp_testhdr *udp_pkt = reinterpret_cast<struct client_udp_testhdr *>(mBuf);
+	struct client_udp_testhdr *udp_pkt = reinterpret_cast<struct client_udp_testhdr *>(mSettings->mBuf);
 	reportstruct->isochStartTime.tv_sec = ntohl(udp_pkt->isoch.start_tv_sec);
 	reportstruct->isochStartTime.tv_usec = ntohl(udp_pkt->isoch.start_tv_usec);
 	reportstruct->frameID = ntohl(udp_pkt->isoch.frameid);
@@ -703,7 +727,6 @@ inline void Server::udp_isoch_processing (int rxlen) {
  * ------------------------------------------------------------------- */
 void Server::RunUDP () {
     int rxlen;
-    int readerr = 0;
     bool lastpacket = false;
 
     if (!InitTrafficLoop())
@@ -713,7 +736,7 @@ void Server::RunUDP () {
     // 1) Fatal read error
     // 2) Last packet of traffic flow sent by client
     // 3) -t timer expires
-    while (InProgress() && !readerr && !lastpacket) {
+    while (InProgress() && !lastpacket) {
 	// The emptyreport flag can be set
 	// by any of the packet processing routines
 	// If it's set the iperf reporter won't do
@@ -749,11 +772,7 @@ void Server::RunUDP () {
 		}
 	    }
 	}
-#ifdef HAVE_STRUCT_TCP_INFO_TCPI_TOTAL_RETRANS
-	ReportPacket(myReport, reportstruct, NULL);
-#else
 	ReportPacket(myReport, reportstruct);
-#endif
     }
     disarm_itimer();
     int do_close = EndJob(myJob, reportstruct);
@@ -762,7 +781,7 @@ void Server::RunUDP () {
 	// 1) we're NOT receiving multicast
 	// 2) the user requested no final exchange
 	// 3) this is a full duplex test
-	write_UDP_AckFIN(&myReport->info);
+	write_UDP_AckFIN(&myReport->info, mSettings->mBufLen);
     }
     if (do_close) {
 #if HAVE_THREAD_DEBUG
@@ -774,4 +793,5 @@ void Server::RunUDP () {
     Iperf_remove_host(mSettings);
     FreeReport(myJob);
 }
+
 // end Recv
