@@ -70,6 +70,7 @@
 #include "active_hosts.h"
 #include "util.h"
 #include "Reporter.h"
+#include "payloads.h"
 
 #ifdef WIN32
 #include "service.h"
@@ -224,6 +225,12 @@ int main(int argc, char **argv) {
 
     }
 
+    int mbuflen = (ext_gSettings->mBufLen > MINMBUFALLOCSIZE) ? ext_gSettings->mBufLen : MINMBUFALLOCSIZE;
+#if (((HAVE_TUNTAP_TUN) || (HAVE_TUNTAP_TAP)) && (AF_PACKET))
+    mbuflen += TAPBYTESSLOP;
+#endif
+    ext_gSettings->mBuf = new char[mbuflen];
+
     unsetReport(ext_gSettings);
     switch (ext_gSettings->mThreadMode) {
     case kMode_Client :
@@ -283,8 +290,10 @@ int main(int argc, char **argv) {
 	into->runNow = ext_gSettings;
 	// Start all the threads that are ready to go
 	thread_start_all(into);
+	Condition_Lock(threads_start.await);
 	threads_start.ready = 1;
         Condition_Signal(&threads_start.await);
+	Condition_Unlock(threads_start.await);
     }
 #else
     // No need to make a reporter thread because we don't have threads
